#!/usr/bin/perl

use strict;

my $degree = 7;
my $c_offset = 129;

#
# On each molecule 'a' denotes an O atom while 'b' and 'c'
# label the two H atoms. The two molecules are referred
# to as 1 and 2.
#

# O-O, O-H, H-H
my @dist = qw(a1a2
              a1b2 a1c2 a2b1 a2c1
              b1b2 b1c2 b2c1 c1c2);

my $ndist = scalar @dist;

my %dist2idx;

# map distances <-> integers
for (my $n = 0; $n < $ndist; ++$n) {
    $dist[$n] = &reorder_label($dist[$n]);
    $dist2idx{$dist[$n]} = $n;
}

die unless scalar (keys %dist2idx) == $ndist; # unique distances

# generate all permutations that must leave the monomials invariant;
# the permutations are: (a) permutations of hydrogens on each molecule
# and (b) permutations of the molecules

my @permutations;

# permute molecules for hydrogens' each permutation

my @p2 = (0, 1,
          1, 0);

foreach my $h1 (0, 1) {
    foreach my $h2 (0, 1) {
        my @h_mask = ($h1, $h2);
        for (my $i = 0; $i < scalar @p2; $i += 2) {
            my @m_perm = @p2[$i, $i + 1];
            my @curr;
            for (my $n = 0; $n < $ndist; ++$n) {
                my $label = &permute_molecules
                    (&permute_hydrogens($dist[$n], \@h_mask), \@m_perm);
                die "unknown distance '$label'" unless exists $dist2idx{$label};
                push @curr, $dist2idx{$label};
            }
            push @permutations, \@curr;
#            for (my $n = 0; $n < $ndist; ++$n) {
#                printf "%3d", $curr[$n];
#            }
#            print "\n";
        }
    }
}

#
# having the permutations, loop over $degree variables
#

my %monomials;

my @ii;
for (my $n = 0; $n < $degree; ++$n) {
    $ii[$n] = 0;
}

&process_product(\@ii);

while (1) {
    my $i = $degree - 1;

    for (; $i >= 0; --$i) {
        $ii[$i] += 1;
        my $upper = ($i == 0 ? ($ndist - 1) : $ii[$i - 1]);
        if ($ii[$i] <= $upper) {
            goto looping;
        }
        $ii[$i] = 0;
    }

    last;
    looping: &process_product(\@ii);
}

#&print_c_dist();

my $n = $c_offset;
# print the monomials
foreach (keys %monomials) {
#    print "$_\n";
    print "    m[$n] = ", &monomial_to_c($_), ";\n";
    $n++;
}

print STDERR scalar (keys %monomials), "\n";

################################################################################

sub process_product {

    my $ii = shift;

    die unless $degree == scalar @{$ii};

    # construct a monomial by applying all permutations to $ii
    my %terms; # must be unique and ordered, to be comparable

    foreach (@permutations) {
        $terms{&make_term(&permute($ii, $_))} = undef;
    }

    $monomials{join('+', sort {&compare_terms($a, $b)} (keys %terms))} = undef;
}

################################################################################

sub permute {

    my $ii = shift;
    my $pp = shift;

    my $result = [];
    for (my $n = 0; $n < scalar @{$ii}; ++$n) {
        $result->[$n] = $pp->[$ii->[$n]];
    }

    return $result;
}

################################################################################

sub make_term {

    my $ii = shift;
    return join ('*', sort {$a <=> $b} @{$ii})
}

################################################################################

sub compare_terms {

    my ($t1, $t2) = @_;

    my @n1 = split /\*/, $t1;
    my @n2 = split /\*/, $t2;

    my $len1 = scalar @n1;
    my $len2 = scalar @n2;

    my $l = $len1 < $len2 ? $len1 : $len2;

    for (my $i = 0; $i < $l; ++$i) {
        if ($n1[$i] < $n2[$i]) {
            return -1;
        } elsif ($n1[$i] > $n2[$i]) {
            return 1;
        } else {
            next;
        }
    }

    return 0;
}

################################################################################

sub reorder_label {

    my $label = shift;

    die unless length($label) == 4;

    my $a1 = substr($label, 0, 1);
    my $m1 = substr($label, 1, 1);
    my $a2 = substr($label, 2, 1);
    my $m2 = substr($label, 3, 1);

    die unless $m1 != $m2;

    return ($m1 < $m2 ? "$a1$m1$a2$m2" : "$a2$m2$a1$m1");
}

################################################################################

sub permute_molecules {

    my $label = shift;
    my $perm2 = shift;

    die unless length($label) == 4;

    my $a1 = substr($label, 0, 1);
    my $m1 = substr($label, 1, 1);
    my $a2 = substr($label, 2, 1);
    my $m2 = substr($label, 3, 1);

    die unless $m1 < $m2;

    my $pm1 = 1 + $perm2->[$m1 - 1];
    my $pm2 = 1 + $perm2->[$m2 - 1];

    return ($pm1 < $pm2 ? "$a1$pm1$a2$pm2" : "$a2$pm2$a1$pm1");
}

################################################################################

sub permute_hydrogens {

    my $label = shift;
    my $mask = shift;

    die unless length($label) == 4;

    my $a1 = substr($label, 0, 1);
    my $m1 = substr($label, 1, 1);
    my $a2 = substr($label, 2, 1);
    my $m2 = substr($label, 3, 1);

    die unless $m1 < $m2;

    if ($mask->[$m1 - 1] == 1) {
        if ($a1 eq 'b') {
            $a1 = 'c';
        } elsif ($a1 eq 'c') {
            $a1 = 'b';
        }
    }

    if ($mask->[$m2 - 1] == 1) {
        if ($a2 eq 'b') {
            $a2 = 'c';
        } elsif ($a2 eq 'c') {
            $a2 = 'b';
        }
    }

    return "$a1$m1$a2$m2";
}

################################################################################

sub monomial_to_c {

    my $txt = shift;
    my @toks = split /\+/, $txt;

    my $result = '';
    for (my $n = 0; $n < scalar @toks; ++$n) {
        $result .= " + " if $n > 0;
        my @d = split /\*/, $toks[$n];
        my $term = '';
        for (my $k = 0; $k < scalar @d; ++$k) {
            $term .= "x\[$d[$k]\]";
            $term .= '*' unless $k + 1 == scalar @d;
        }
        $result .= $term;
    }

    return $result;
}

################################################################################

sub print_c_dist {

    print "    const double x[$ndist] = \{\n";
    for (my $n = 0; $n < $ndist; ++$n) {
        my $label = $dist[$n];
        my $a1 = substr($label, 0, 2);
        my $a2 = substr($label, 2, 2);

        print "        morse($a1, $a2)";
        print ",\n" unless $n == $ndist - 1;
    }
    print "\n    \};\n";
}

################################################################################
