#!/usr/bin/perl

use strict;
use warnings;

#
# monomial is a list of length L <= N*(N-1)/2 listing the powers of the
# variables; e.g., (1, 2, 0, ..., 0) <=> x1*x2*x2; total degree is just
# the sum of the elements
#
# permutation initially acts on the atoms and so is represented by a list
# of length N : (0, 1, 2, 3, ..., N) is "identity", and so on
#
# for (H2O)x2 the atoms are: Oa Ha1 Ha2 Ob Hb1 Hb2
# and possible permutations permute Hydrogens withing the molecules
# and molecules themselves
#

my $i0 = 0;
my $nmolec = 2;
my @atoms = qw(Oa Ha1 Ha2 Ob Hb1 Hb2);
my %atom_idx;
for (my $n = 0; $n < scalar @atoms; ++$n) {
    $atom_idx{$atoms[$n]} = $n;
}

my @permutations;

my @p3 = (0, 1,
          1, 0);

foreach my $h1 (0, 1) {
    foreach my $h2 (0, 1) {
	my @h_mask = ($h1, $h2);
    	for (my $i = 0; $i < scalar @p3; $i += $nmolec) {
    	    my @m_perm = @p3[$i, $i + 1];
	    push @permutations, &permute_molecules
	        (&permute_hydrogens(\@h_mask), \@m_perm);
        }
    }
}

print STDERR scalar @permutations, " permutations:\n";
foreach (@permutations) {
    print STDERR join(' ', @{$_}), "\n";
    for (my $n = 0; $n < scalar @atoms; ++$n) {
        $_->[$n] = $atom_idx{$_->[$n]};
    }
}
exit;

#
# now find how the permutations act on the monomials
# composed out of (some) distances
#

my @variables = (#['Ha1', 'Ha2'],
                 ['Ha1', 'Hb1'],
                 ['Ha1', 'Hb2'],
                 ['Ha2', 'Hb1'],
                 ['Ha2', 'Hb2'],
                 #['Hb1', 'Hb2'],
                 ['Oa', 'Ha1'],
                 ['Oa', 'Ha2'],
                 ['Oa', 'Hb1'],
                 ['Oa', 'Hb2'],
                 ['Ob', 'Ha1'],
                 ['Ob', 'Ha2'],
                 ['Ob', 'Hb1'],
                 ['Ob', 'Hb2'],
                 ['Oa', 'Ob']);

print STDERR "\n";
my $nvar = scalar @variables;
my %atom_pair_to_var;
for (my $v = 0; $v < $nvar; ++$v) {
    my $key = join(':', sort @{$variables[$v]});
    print STDERR $key, "\n";
    $atom_pair_to_var{$key} = $v;
}
print STDERR "\n";

#
# now loop over 2nd degree monomials [we want 2B-terms only]
#

my %mono_all;
for (my $k1 = 0; $k1 < $nvar; ++$k1) {
    for (my $k2 = 0; $k2 < $nvar; ++$k2) {
        my @mono = (0) x $nvar;
        $mono[$k1] += 1;
        $mono[$k2] += 1;
        next unless is_2B(\@mono);
#        print STDERR join(':', @mono), "\n";
        $mono_all{join(':', @mono)} = undef;
    }
}

#print STDERR "\n";
print STDERR scalar keys %mono_all, " possible 2B, 2nd degree monomials\n";
my $mono2_orb = find_orbits(\%mono_all);

print STDERR scalar @{$mono2_orb}, " 2nd degree terms\n";

#
# now loop over 3rd degree monomials [keeping 2B-terms only]
#

for (my $k1 = 0; $k1 < $nvar; ++$k1) {
    for (my $k2 = 0; $k2 < $nvar; ++$k2) {
        for (my $k3 = 0; $k3 < $nvar; ++$k3) {
            my @mono = (0) x $nvar;
            $mono[$k1] += 1;
            $mono[$k2] += 1;
            $mono[$k3] += 1;
            next unless is_2B(\@mono);
            $mono_all{join(':', @mono)} = undef;
        }
    }
}

print STDERR scalar keys %mono_all, " of possible 2B, 3rd degree monomials\n";
my $mono3_orb = find_orbits(\%mono_all);

print STDERR scalar @{$mono3_orb}, " 3rd degree terms\n";

print "\n";
for (my $v = 0; $v < $nvar; ++$v) {
    print "    x[$v] = morse($variables[$v]->[0], $variables[$v]->[1]);\n"
}

print "\n\n";

&print_orbits($i0, $mono2_orb);

print "\n\n";

$i0 += scalar @{$mono2_orb};
&print_orbits($i0, $mono3_orb);

print "\n\n";

################################################################################

sub permute_hydrogens {
    my $mask = shift;
    my @dimer = qw(Oa Ha1 Ha2 Ob Hb1 Hb2);

    for (my $m = 0; $m < $nmolec; ++$m) {
        if ($mask->[$m] == 1) {
            my $tt = $dimer[3*$m + 1];
            $dimer[3*$m + 1] = $dimer[3*$m + 2];
            $dimer[3*$m + 2] = $tt;
        }
    }

    return \@dimer;
}

################################################################################

sub permute_molecules {
    my ($dimer, $mp) = @_;
    my $permuted = [];

    for (my $m = 0; $m < $nmolec; ++$m) {
        my $src = $mp->[$m];
        for (my $k = 0; $k < 3; ++$k) {
            $permuted->[3*$m + $k] = $dimer->[3*$src + $k];
        }
    }

    return $permuted;
}

################################################################################

sub is_2B {
    my $mono = shift;
    my %mols;

    die unless $nvar == scalar @{$mono};
    for (my $v = 0; $v < $nvar; ++$v) {
        next unless $mono->[$v] > 0;
        my $var = $variables[$v];
        my $m1 = substr ($var->[0], 1, 1);
        my $m2 = substr ($var->[1], 1, 1);
        $mols{$m1} = undef;
        $mols{$m2} = undef;
    }

    my $nmol = scalar keys %mols;

    return ($nmol == 2 ? 1 : 0);
}

################################################################################

sub find_orbits {
    my $all = shift;

    my @orbits;

    while (1) {
        my @still_there = keys %{$all};
        last if scalar @still_there == 0;
        my $curr = shift @still_there;
        push @orbits, $curr;
        for my $p (@permutations) {
            my $permuted = permute_monomial($curr, $p);
#	    print $count, " >> ", $curr, " | ", @{$p}, " | ", $permuted, "\n";
            delete ${$all}{$permuted};
        }
#	++$count;
    }

    return \@orbits;
}

################################################################################

sub permute_monomial {
    my ($src_str, $p) = @_;

    my @src = split(':', $src_str);
    my $nvar = scalar @src;
    my @dst = (0) x $nvar;

#    print $src_str, "\n", $p, "\n", @{$p}, "\n";
#    exit;
    # p is array ref for array of atoms numbers, e.g., 012345
    # src_str is distance combination (variables): 0:0:0:0:0:0:1:0:1:0:0:0:0:0:0

    for (my $v = 0; $v < $nvar; ++$v) {
        next unless $src[$v] > 0;
        my $atms = $variables[$v];
        my ($a1, $a2) = ($atom_idx{$atms->[0]}, $atom_idx{$atms->[1]});
	# $a1 and $a2 are atom numbers, for Oa-Ob could be 0,3
	# if p = (3, x,x,0,x,x); where x is arbitrary, then 
	# which means @permuted = ("Ob", "Oa");
        my @permuted = ($atoms[$p->[$a1]], $atoms[$p->[$a2]]);
        my $var = join(':', sort @permuted);
        die " ** something is wrong **\n"
            unless exists($atom_pair_to_var{$var});
        $dst[$atom_pair_to_var{$var}] = $src[$v];
    }
    return join(':', @dst);
}

################################################################################

sub print_one {
    my ($i, $m) = @_;
    my %terms;
    for my $p (@permutations) {
        my $xx = permute_monomial($m, $p);
        my @yy = split(':', $xx);
        my @pp;
        for (my $n = 0; $n < scalar @yy; ++$n) {
            next unless $yy[$n] > 0;
            for (my $p = 0; $p < $yy[$n]; ++$p) {
#                push @pp, (sprintf "x%2.2d", $n + 1);
                push @pp, (sprintf "x[%2.2d]", $n );
            }
        }
        my $curr = join('*', sort @pp);
        $terms{$curr} = undef;
    }
    print "    p[$i] = ", join(' + ', keys %terms), ";\n";
}

################################################################################

sub print_orbits {
    my ($i0, $orbs) = @_;

    foreach (@{$orbs}) {
        print_one($i0++, $_);
    }
}

################################################################################
