#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <stdexcept>

#include "read-xyz.h"
#include "orient.h"

#include "ttm4-es.h"
#include "ps.h"

#include "dip_poly2b5.h"

namespace {

static ttm::ttm4_es ind_model;

double Rswitch(7.5);

static std::vector<tools::xyz_frame> molec;

//----------------------------------------------------------------------------//

double dot_diff(double*a, double*b)
{
    double d0 = a[0] - b[0];
    double d1 = a[1] - b[1];
    double d2 = a[2] - b[2];

    return std::sqrt(d0*d0 + d1*d1 + d2*d2);
}

//----------------------------------------------------------------------------//

const double r2i = 5.5;
const double r2f = 7.5;

double f_switch2(const double& r)
{
    if (r > r2f)
	return 0.0;
    else if (r > r2i) {
	const double x = (r - r2i)/(r2f - r2i);
	return 1 + x*x*(2*x - 3);
    } else
	return 1.0;
}

//----------------------------------------------------------------------------//

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{

    try {
	std::ifstream ifs(argv[1]);

	if (!ifs)
	    throw std::runtime_error("could not open the XYZ file");

	std::string comment;
	tools::read_xyz(argv[1], molec, true);
    } catch (const std::exception& e) {
	std::cerr << " ** Error ** : " << e.what() << std::endl;
	return 1;
    }

    std::ofstream out_ind;
    out_ind.open("dip.ind");

    std::ofstream out_corr;
    out_corr.open("dip.corr");

    std::ofstream out_ref;
    out_ref.open("dip.ref");

    for(size_t n = 0; n < molec.size(); ++n){
	const size_t nw = molec[n].nw;
	double* xyz = molec[n].xyz;

	// Compute the Partridge-Schwenke 1B dipole

	double dip_ps1B[3] = {0.0, 0.0, 0.0};
	for(size_t i = 0; i < nw; ++i){
	    double* mon_xyz = molec[n].xyz + i*9;
	    double q3[3];

//	    std::cout << "Molecule " << i << std::endl;

	    ps::dms_nasa(0.0, 0.0, 0.0,
			 mon_xyz, q3, 0, false);
	    for(size_t a = 0; a < 3; ++a){ // o,h,h
		const size_t a3 = 3*a;
		for(size_t b = 0; b < 3; ++b){ // x,y,z
//		    std::cout << mon_xyz[a3+b]*q3[a] << "  ";
		    dip_ps1B[b] += mon_xyz[a3+b]*q3[a];
		}
//		std::cout << std::endl;
	    }
	}

	               // Compute TTM4-x RMS //
	
    	double Eelec, Eind;
    	ind_model(nw, xyz, Eelec, 0, Eind, 0);

	double dipind[3];
    	ind_model.dipind(nw, dipind);

	double err_ind = dot_diff(molec[n].dip, dipind);

	               // Compute model error //
	 
	double corrdip[3] = {dipind[0], dipind[1], dipind[2]};
	//double corrdip[3] = {0.0, 0.0, 0.0};
	// Add in the 2B corrections
	for(size_t i = 0; i < nw; ++i){
	    const size_t i9 = 9*i;
	    for(size_t j = i+1; j < nw; ++j){
		const size_t j9 = 9*j;
		double dip_corr2b[3];
		h2o::dip_poly2b5::eval(xyz + i9, xyz + j9, dip_corr2b);

		double s = f_switch2(tools::distance(xyz + i9, xyz + j9));

#ifdef JUST_POLY
		double dimer[18];
		std::copy(xyz + i9, xyz + i9 + 9, dimer);
		std::copy(xyz + j9, xyz + j9 + 9, dimer + 9);
		ind_model(2, dimer, Eelec, 0, Eind, 0);

		double dip_ind2b[3];
		ind_model.dipind(2, dip_ind2b);

		for(size_t k = 0; k < 3; ++k)
		    corrdip[k] += s*(dip_corr2b[k] - dip_ind2b[k]);
		    //corrdip[k] += s*(dip_corr2b[k] + dip_ind2b[k]);
#else
		for(size_t k = 0; k < 3; ++k)
		    corrdip[k] += s*dip_corr2b[k];
#endif		

	    }
	}

	double err_corr = dot_diff(molec[n].dip, corrdip);

	// Now print results	
	std::cerr << err_ind << ' ' 
	          << err_corr << std::endl;

	out_ind  << dipind[0] << ' ' << dipind[1] << ' '
	         << dipind[2] << std::endl;

	out_corr <<corrdip[0] << ' ' << corrdip[1] << ' '
	         << corrdip[2] << std::endl;

	out_ref  << molec[n].dip[0] << ' ' << molec[n].dip[1] << ' '
	         << molec[n].dip[2] << std::endl;


    }

    return 0;
}
