#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <vector>

#include "read-xyz.h"
#include "orient.h"

#include "alpha-ttm4.h"

#include "alpha-el-poly2b3.h"

namespace {

double Rswitch(7.5);

static std::vector<tools::xyz_frame> molec;

//----------------------------------------------------------------------------//

double dot_diff(size_t n, double*a, double*b)
{
    double sum(0);
    for(size_t i = 0; i < n; ++i){
	double d0 = a[i] - b[i];
	sum += d0*d0;
    }

    return std::sqrt(sum);
}

//----------------------------------------------------------------------------//

double dot_diff_ref_pol(double* ref, // 6 element upper triangular
			double* model) // 9 elements
{
    std::vector<double> vals;

    vals.push_back(ref[0] - model[0]); // xx
    vals.push_back(ref[1] - model[1]); // xy
    vals.push_back(ref[2] - model[2]); // xz
    vals.push_back(ref[1] - model[3]); // yx
    vals.push_back(ref[3] - model[4]); // yy
    vals.push_back(ref[4] - model[5]); // yz
    vals.push_back(ref[2] - model[6]); // zx
    vals.push_back(ref[4] - model[7]); // zy
    vals.push_back(ref[5] - model[8]); // zz

    double sum(0);
    for(size_t i = 0; i < vals.size(); ++i){
	double d0 = vals[i];
	sum += d0*d0;
    }

    return std::sqrt(sum);
}

//----------------------------------------------------------------------------//

const double r2i = 5.5;
const double r2f = 7.5;

double f_switch2(const double& r)
{
    if (r > r2f)
	return 0.0;
    else if (r > r2i) {
	const double x = (r - r2i)/(r2f - r2i);
	return 1 + x*x*(2*x - 3);
    } else
	return 1.0;
}

//----------------------------------------------------------------------------//

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{

    try {
	std::ifstream ifs(argv[1]);

	if (!ifs)
	    throw std::runtime_error("could not open the XYZ file");

	std::string comment;
	tools::read_xyz(argv[1], molec, false, true);
    } catch (const std::exception& e) {
	std::cerr << " ** Error ** : " << e.what() << std::endl;
	return 1;
    }

    std::ofstream out_ind;
    out_ind.open("pol.ind");

    std::ofstream out_corr;
    out_corr.open("pol.corr");

    std::ofstream out_ref;
    out_ref.open("pol.ref");

    std::cerr.setf( std::ios::fixed, std:: ios::floatfield );
    std::cerr.precision(6);

    for(size_t n = 0; n < molec.size(); ++n){
	const size_t nw = molec[n].nw;
	double* xyz = molec[n].xyz;

	               // Compute TTM4-x RMS //
	
	double polind[9];
	ttm::alpha_no_1b(nw, xyz, polind);

	double err_ind = dot_diff_ref_pol(molec[n].pol, polind);

	               // Compute model error //
	 
	double corrpol[9] = {0.0};
	std::copy(polind, polind + 9, corrpol);
	
	// Add in the 2B corrections
	for(size_t i = 0; i < nw; ++i){
	    const size_t i9 = 9*i;
	    for(size_t j = i+1; j < nw; ++j){
		const size_t j9 = 9*j;
		double pol_corr2b[9];
		std::fill(pol_corr2b, pol_corr2b + 9, 0.0);
		for(size_t k = 0; k < 9; ++k)
		    std::cout << pol_corr2b[k] << ' ';
		std::cout << std::endl;
		h2o::alpha_el_poly2b3::eval(xyz + i9, xyz + j9, pol_corr2b);

		double s = f_switch2(tools::distance(xyz + i9, xyz + j9));

		for(size_t k = 0; k < 9; ++k)
		    corrpol[k] += s*pol_corr2b[k];

	    }
	}

	double err_corr = dot_diff_ref_pol(molec[n].pol, corrpol);

	// Now print results	
	std::cerr << err_ind << ' ' 
	          << err_corr << std::endl;

	out_ind  << polind[0] << ' ' << polind[1] << ' '
	         << polind[2] << ' ' << polind[4] << ' ' 
	         << polind[5] << ' ' << polind[8] << ' ' 
		 << std::endl;

	out_corr << corrpol[0] << ' ' << corrpol[1] << ' '
	         << corrpol[2] << ' ' << corrpol[4] << ' ' 
	         << corrpol[5] << ' ' << corrpol[8] << ' ' 
		 << std::endl;

	out_ref  << molec[n].pol[0] << ' ' << molec[n].pol[1] << ' '
	         << molec[n].pol[2] << ' ' << molec[n].pol[3] << ' '
	         << molec[n].pol[4] << ' ' << molec[n].pol[5] << ' '
		 << std::endl;

    }

    return 0;
}
