#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstring>

#include "read-hist.h"
#include "pbc-tools.h"
#include "struct-oh-vect.h"

#define SHIFT_HALF_CELL yes

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

std::vector<tools::pair_frame> pair_frames;

// placeholders
double int_time = 20;

//----------------------------------------------------------------------------//

void print_all(const double time, const size_t N, const double* drij)
{
    std::cout << std::setw(8) << N/3
	      << std::setw(15) << time << std::endl;

    for(size_t i = 0; i < N/3; ++i){
	for(size_t j = 0; j < 3; ++j){
	    std::cout << std::setw(15) << drij[3*i + j];
	}
	std::cout << std::endl;
    }

}

//----------------------------------------------------------------------------//

size_t do_collect_oh(size_t lineno, std::ifstream& ifs,
		     const double active_min, const double active_max)
{

    size_t frameno(0);

    while(!ifs.eof()){

	opt::hist_frame frame;

	opt::read_cmd_frame(lineno, ifs, frame);

	if(frame.natm > 0){

	    tools::center_slab_at_origin(frame.natm, frame.cell, frame.mass,
		       			 frame.xyz,
#ifdef SHIFT_HALF_CELL
		       			 true
#else
       					 false
#endif
       					);
//	    dump_frame(frame.cell, frame.natm, frame.xyz);
//	    if(frameno > 0)
//		exit(1);

	    const int num_molec = frame.natm / 3;
	    const int num_pairs = num_molec * 2;

	    // initialize/allocate pair_frame
	    tools::pair_frame pair_frame(num_pairs);
	    pair_frame.time = frame.time;

	    for(size_t i = 0; i < frame.natm; i += 3){ // O index, assume OHH

		for(size_t ij = 1; ij < 3; ++ij){ // two H per O

		    const size_t j = i + ij; // H index
		    const size_t pair_id = i/3*2 + ij - 1;

		    const double* qi = frame.xyz + 3*i;
		    const double* qj = frame.xyz + 3*j;

		    double drij[3];
		    const double rij = tools::distance3(frame.imcon, 
					       		frame.cell, 
					       		frame.rcell,
					       		qi, qj, drij);

		    pair_frame.d_rij[3*pair_id + 0] = drij[0]/rij;
    		    pair_frame.d_rij[3*pair_id + 1] = drij[1]/rij;
    		    pair_frame.d_rij[3*pair_id + 2] = drij[2]/rij;

		    const double z_abs = std::fabs(qi[2]); // z-component of O
		    if(z_abs < active_max && z_abs > active_min){
			pair_frame.region[pair_id] = 1; // accum c2(t)
		    }else{
			pair_frame.region[pair_id] = 0; // ignore
		    }
		}
	    }

	    pair_frames.push_back(pair_frame);
//	    print_all(pair_frame.time, pair_frame.npairs*3, pair_frame.d_rij);

	    ++frameno;
	}

    }

    return frameno;
}

//----------------------------------------------------------------------------//

void compute_correlation_function()
{

    const double dt = pair_frames[1].time - pair_frames[0].time;
    const int int_time_ind = (int) (int_time/dt);

    double corr_p2[int_time_ind];
    double ntime[int_time_ind];

    std::fill(corr_p2, corr_p2 + int_time_ind, 0.0);
    std::fill(  ntime,   ntime + int_time_ind, 0.0);

    // loop over each pair
    for(size_t n = 0; n < pair_frames[0].npairs; ++n){

	// loop over t0
	for(int i = 0; i < (int) pair_frames.size(); ++i){

	  // molecule is in region of interest at time t0
	  if(pair_frames[i].region[n] == 1){

	    // loop over t0 + \delta t
	    for(int j = i;
		       j < std::min((int) pair_frames.size(), i + int_time_ind);
		       ++j){
		const int time = (j-i);

		const double cth = 
		    pair_frames[i].d_rij[3*n + 0]*pair_frames[j].d_rij[3*n + 0]
		  + pair_frames[i].d_rij[3*n + 1]*pair_frames[j].d_rij[3*n + 1]
		  + pair_frames[i].d_rij[3*n + 2]*pair_frames[j].d_rij[3*n + 2];

		corr_p2[time] += (3.0*cth*cth - 1.0);
		++ntime[time];
	    }

	  } // restriction for bulk or interface
	}
    }

    for(int i = 0; i < int_time_ind; ++i)
	std::cout << std::setw(20) << i*dt
	          << std::setw(20) << 0.5*corr_p2[i]/ntime[i]
	          << std::setw(20) << 0.5*corr_p2[i]
	          << std::setw(20) << ntime[i]
		  << std::endl;

}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 5) {
        std::cerr << "usage: collect_oh_vectors int_time active_min active_max POSITION_CMD ... " 
	          << std::endl;
        return EXIT_FAILURE;
    }

    {
	std::istringstream iss(argv[1]);
	iss >> int_time;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[1]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }

    double active_min(0);
    {
	std::istringstream iss(argv[2]);
	iss >> active_min;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[2]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }

    double active_max(0);
    {
	std::istringstream iss(argv[3]);
	iss >> active_max;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[3]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }
    std::cout.setf(std::ios::fixed, std::ios::floatfield);
    std::cout << std::setprecision(8);

    for(int i = 4; i < argc; ++i){

	std::ifstream ifs(argv[i]);
	std::string line;
	size_t lineno(0);

	const size_t frameno = do_collect_oh(lineno, ifs,
					     active_min, active_max);

	std::cerr << argv[i] << " : " << frameno << std::endl;

    }

    compute_correlation_function();

}
