#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include "read-hist.h"

#include "constants.h"
#include "ps.h"

#include "ltp2011.h"

#include "dip_poly2b5.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static std::vector<opt::hist_frame> frames;

static h2o::ltp2011 dip_model;

const double r2f = 7.5;

const double dr = 0.01;
const double dmu1b = 0.01;
const double dmu2b = 0.01;

const double    r_max = 7.5;
const double mu1b_max = 5.0;
const double mu2b_max = 5.0;

int pairs_per_frame = 0;
double invvolm = 0;

//----------------------------------------------------------------------------//

inline double norm(const int nentry, const double* arr)
{
    double val(0);
    for(int i = 0; i < nentry; ++i)
	val += arr[i]*arr[i];

    return std::sqrt(val);
}

//----------------------------------------------------------------------------//

inline double distance(int imcon, const double* cell, 
		       const double* r1, const double* r2)
{
    assert(imcon == 0 || imcon == 1);

    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    dx -= cell[0]*std::floor(dx/cell[0] + 0.5);
    dy -= cell[4]*std::floor(dy/cell[4] + 0.5);
    dz -= cell[8]*std::floor(dz/cell[8] + 0.5);

    return std::sqrt(dx*dx + dy*dy + dz*dz);
}

//----------------------------------------------------------------------------//

inline void image(const double* cell, const double* r1, const double* r2, 
		  double* crd_out)
{
    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    double facx = std::floor(dx/cell[0] + 0.5);
    double facy = std::floor(dy/cell[4] + 0.5);
    double facz = std::floor(dz/cell[8] + 0.5);

    crd_out[0] = r2[0] + cell[0]*facx;
    crd_out[1] = r2[1] + cell[4]*facy;
    crd_out[2] = r2[2] + cell[8]*facz;

}

//----------------------------------------------------------------------------//

inline void image_monomer(const double* cell, const double* r1, double* monomer)
{
    std::fill(monomer, monomer+9, 0.0);

    // take r1 oxygen as central atom, image its two hydrogens
     
    for(int i = 0; i < 3; ++i)
	monomer[i] = r1[i];
    
    for(int a = 1; a < 3; ++a)
	image(cell, r1, r1 + 3*a, monomer + 3*a);

}

//----------------------------------------------------------------------------//

inline void image_dimer(const double* cell, 
      			const double* r1, const double* r2,
		      	double* dimer)
{
    std::fill(dimer, dimer+18, 0.0);

    // take r1 oxygen as central atom, image its two hydrogens
     
    for(int i = 0; i < 3; ++i)
	dimer[i] = r1[i];
    
    for(int a = 1; a < 3; ++a)
	image(cell, r1, r1 + 3*a, dimer + 3*a);

    // Now image the oxygen of the second molecule
     
    image(cell, r1, r2, dimer + 9);

    // Finally, reattach hydrogens of molec two to oxygen of molec 2
    for(int a = 1; a < 3; ++a)
	image(cell, dimer+9, r2 + 3*a, dimer + 9 + 3*a);
}

//----------------------------------------------------------------------------//

void dump_1d_distrib(const int n_frame, const char* filename, const int* bin, 
		     const double dr, const int nbins)
{
    std::ofstream ofs;
    ofs.open(filename);
    ofs.precision(9);

    ofs << std::scientific;

    int total(0);
    for(int i = 0; i < nbins; ++i)
	total += bin[i];

    ofs << "# " << n_frame << std::endl;
    for(int i = 0; i < nbins; ++i)
	ofs << std::setw(20) << dr*(i+0.5)
	    << std::setw(20) << bin[i]/((double) total) << std::endl;

    ofs.close();
}

//----------------------------------------------------------------------------//

void dump_2d_radial_distrib(const int n_frame, const char* filename, 
			    const int* bin, 
			    const double dr, const double dmu,
			    const int nbin_r, const int nbin_mu)
{
    std::ofstream ofs;
    ofs.open(filename);
    ofs.precision(9);

    ofs << std::scientific;

    const double norm_fac = 1.0/invvolm/pairs_per_frame;

    ofs << "# " << std::setw(10) << nbin_r 
                << std::setw(10) << nbin_mu
                << std::setw(10) << n_frame
		<< std::endl;
    for(int i = 0; i < nbin_r; ++i){
	const double r = dr*(i+0.5);

	const double rm = r - dr/2;
	const double rp = r + dr/2;

	const double dvol = 4.0/3.0*M_PI*(std::pow(rp, 3) - std::pow(rm, 3));
	for(int j = 0; j < nbin_mu; ++j){
	    ofs << std::setw(20) << r
		<< std::setw(20) << dmu*(j+0.5)
		<< std::setw(20) << norm_fac*bin[i*nbin_mu + j]/dvol 
		<< std::endl;
	}
	ofs << std::endl;
    }

    ofs.close();
}

//----------------------------------------------------------------------------//

void dump_2d_distrib(const int n_frame, const char* filename, const int* bin, 
 		     const double di, const double dj,
	 	     const int nbin_i, const int nbin_j)
{
    std::ofstream ofs;
    ofs.open(filename);
    ofs.precision(9);

    ofs << std::scientific;

    int total(0);
    for(int i = 0; i < nbin_i*nbin_j; ++i)
	total += bin[i];

    ofs << "# " << std::setw(10) << nbin_i 
                << std::setw(10) << nbin_j
                << std::setw(10) << n_frame
		<< std::endl;
    for(int i = 0; i < nbin_i; ++i){
	for(int j = 0; j < nbin_j; ++j){
	    ofs << std::setw(20) << di*(i+0.5)
		<< std::setw(20) << dj*(j+0.5)
		<< std::setw(20) << bin[i*nbin_j + j]/((double) total) 
		<< std::endl;
	}
	ofs << std::endl;
    }

    ofs.close();
}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc != 1) {
        std::cerr << "usage: bulk_dipole ..." << std::endl;
        return EXIT_FAILURE;
    }

    try {

    	//opt::read_cmd(frames, true);
    	opt::read_cmd(frames, false);

	std::cerr <<" CMD : " << frames.size() << std::endl;

    } catch (const std::exception& e) {
	std::cerr << e.what() << std::endl;
	return EXIT_FAILURE;
    }

    const int nbin_r    = (int)(r_max/dr);
    const int nbin_mu1b = (int)(mu1b_max/dmu1b);
    const int nbin_mu2b = (int)(mu2b_max/dmu2b);

    int bin_1b[nbin_mu1b];
    int bin_2b[nbin_mu2b];
    int bin_jointRoo1b[nbin_r*nbin_mu1b];
    int bin_jointRoo2b[nbin_r*nbin_mu2b];
    int bin_joint1b2b[nbin_mu1b*nbin_mu2b];

    // zero bins

    std::fill(        bin_1b,         bin_1b + nbin_mu1b          , 0.0);
    std::fill(        bin_2b,         bin_2b + nbin_mu2b          , 0.0);
    std::fill(bin_jointRoo1b, bin_jointRoo1b + nbin_r*nbin_mu1b   , 0.0);
    std::fill(bin_jointRoo2b, bin_jointRoo2b + nbin_r*nbin_mu2b   , 0.0);
    std::fill( bin_joint1b2b,  bin_joint1b2b + nbin_mu1b*nbin_mu2b, 0.0);

    // Define models
    
    for(int n = 0; n < (int) frames.size(); ++n){
	const int nmolec = frames[n].natm/3;

	double dip1b[3*nmolec];
	std::fill(dip1b, dip1b + 3*nmolec, 0.0);

	std::vector<int> excluded;

	for(int i = 0; i < nmolec; ++i){

	    double monomer[9];
	    image_monomer(frames[n].cell, frames[n].xyz + i*9, monomer);

#ifdef PS
	    double q3[3];
	    ps::dms_nasa(0.0, 0.0, 0.0,
			 monomer, q3, 0, false);
	    for(size_t a = 0; a < 3; ++a){ // o,h,h
		const size_t a3 = 3*a;
		for(size_t b = 0; b < 3; ++b){ // x,y,z
		    dip1b[3*i + b] += monomer[a3+b]*q3[a];
		}
	    }
	    for(size_t b = 0; b < 3; ++b)
		dip1b[3*i + b] *= constants::DEBYE;
#else
	    dip_model.cartesian(1, monomer, dip1b + 3*i);
#endif

	    double norm_mu1b = norm(3, dip1b + 3*i);

	    std::cout << norm_mu1b << std::endl;
	    if(norm_mu1b < mu1b_max){
    		++bin_1b[(int)(norm_mu1b/dmu1b)];
	    }else{
		std::cerr << "mu1b too big for this monomer! " 
		          << norm_mu1b << std::endl;
		for(size_t a = 0; a < 3; ++a){
    		    if(a == 0)
    			std::cerr << std::setw(8) << "O"; 
	    	    else
    			std::cerr << std::setw(8) << "H";
	    	    for(size_t b = 0; b < 3; ++b){
			std::cerr << std::setw(20) << monomer[3*a+b];
		    }
		    std::cerr << std::endl;
		}
		std::cerr << "Frame " << n << "; time " << frames[n].time
		          << std::endl;
		excluded.push_back(i);
	    }
	}

	int pairs_this_frame(0);
#if 0
	if(excluded.size() == 0){
	for(int i = 0; i < nmolec; ++i){
	    for(int j = i+1; j < nmolec; ++j){
		const int i9 = 9*i;
		const int j9 = 9*j;

		double rOO = distance(frames[n].imcon, frames[n].cell, 
				      frames[n].xyz + i9, frames[n].xyz + j9);

		if(rOO < r2f){

		    double dimer[18];
		    image_dimer(frames[n].cell, 
				frames[n].xyz + i9, frames[n].xyz + j9, dimer);

		    // Get corr two-body dipole
		    
	    	    double dip2b[3];
	    	    corr_dip(dimer, dimer + 9, dip2b);

    		    double norm_mu2b = norm(3, dip2b);

		    // Sum the 1B dipoles
		    double tot_dip1b[3];
		    for(size_t a = 0; a < 3; ++a)
			tot_dip1b[a] = dip1b[3*i + a] + dip1b[3*j + a];
		    double norm_mu1b = norm(3, tot_dip1b);

		    if(norm_mu2b < mu2b_max){
			const int ind_Roo = (int)(rOO/dr);
			const int ind_2b = (int)(norm_mu2b/dmu2b);

			++bin_2b[ind_2b];
		    	++bin_jointRoo2b[ind_Roo*nbin_mu2b + ind_2b];

			const int ind_1b = (int)(norm_mu1b/dmu1b);

			++bin_jointRoo1b[ind_Roo*nbin_mu1b + ind_1b];
			if(rOO < 3.2)
			    ++bin_joint1b2b[ind_1b*nbin_mu2b + ind_2b];

		    }else{
			std::cerr << "mu2b too big! " << norm_mu2b << std::endl;
			exit(0);
		    }
		}
		++pairs_this_frame;
	    }
	}
	}//excluded.size()
#endif

	if(pairs_per_frame == 0)
	    pairs_per_frame = pairs_this_frame;
	else if(pairs_this_frame != pairs_per_frame){
	    std::cerr << "Dying. Not a constant number of atm per frame"
		<< std::endl;
	    exit(1);
	}

	invvolm += 1.0/(frames[n].vol);

	if(n > 0 && (n%1000 == 0)){

	    dump_1d_distrib(n, "distrib_1b.dat", bin_1b, dmu1b, nbin_mu1b);
	    dump_1d_distrib(n, "distrib_2b.dat", bin_2b, dmu2b, nbin_mu2b);
	    dump_2d_radial_distrib(n, "distrib_joint_roo1b.dat", bin_jointRoo1b,
				   dr, dmu1b, nbin_r, nbin_mu1b);
	    dump_2d_radial_distrib(n, "distrib_joint_roo2b.dat", bin_jointRoo2b,
				   dr, dmu2b, nbin_r, nbin_mu2b);
	    dump_2d_distrib(n, "distrib_joint_1b2b.dat", bin_joint1b2b,
	 		    dmu1b, dmu2b, nbin_mu1b, nbin_mu2b);
	    std::cerr << "printing step " << n << std::endl;

	}

    }

    dump_1d_distrib(frames.size(), "distrib_1b.dat", bin_1b, dmu1b, nbin_mu1b);
    dump_1d_distrib(frames.size(), "distrib_2b.dat", bin_2b, dmu2b, nbin_mu2b);
    dump_2d_radial_distrib(frames.size(), "distrib_joint_roo1b.dat", bin_jointRoo1b,
			   dr, dmu1b, nbin_r, nbin_mu1b);
    dump_2d_radial_distrib(frames.size(), "distrib_joint_roo2b.dat", bin_jointRoo2b,
			   dr, dmu2b, nbin_r, nbin_mu2b);
    dump_2d_distrib(frames.size(), "distrib_joint_1b2b.dat", bin_joint1b2b,
		    dmu1b, dmu2b, nbin_mu1b, nbin_mu2b);

    std::cerr << "printed final step" << std::endl;

    return 0;
}
