#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include "read-hist.h"
#include "read-xyz.h"
#include "constants.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static std::vector<opt::hist_frame> frames;

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////
double INNER_SHELL_RADIUS;
double OUTER_SHELL_RADIUS;
double BOX_WIDTH;

double switching_function(double radius){

    //look up formula from Adaptive QM/MM paper 2012 

    if (radius <= INNER_SHELL_RADIUS) return 1.0;
    else if (radius >= OUTER_SHELL_RADIUS) return 0.0;

    double range = OUTER_SHELL_RADIUS - INNER_SHELL_RADIUS;
    double first_term_numerator = (radius - INNER_SHELL_RADIUS);
    first_term_numerator *= first_term_numerator;

    double second_term_numerator = (3*OUTER_SHELL_RADIUS - INNER_SHELL_RADIUS
				   - 2*radius);

    double result = first_term_numerator * second_term_numerator;
    result /= (range*range*range);

    result = 1 - result;

    return result;

}

int main(int argc, char** argv)
{
    if (argc != 3) {
       std::cerr << "usage: bulk_dipole INNER_RADIUS BOX_WIDTH > STDOUT" << std::endl;
        return EXIT_FAILURE;
    }

    std::istringstream iss(argv[1]);
    iss >> INNER_SHELL_RADIUS;

    std::istringstream is(argv[2]);
    is >> BOX_WIDTH;

    OUTER_SHELL_RADIUS = INNER_SHELL_RADIUS + 1.0;

    try {

	std::vector<opt::hist_frame> this_file;
    	opt::read_cmd(this_file, true);
    	std::copy(this_file.begin(), this_file.end(),
    		  std::back_inserter(frames));

	std::cerr <<" CMD : " << this_file.size() << std::endl;

    } catch (const std::exception& e) {
	std::cerr << e.what() << std::endl;
	return EXIT_FAILURE;
    }




//    Open files for Writing

    std::cout.setf(std::ios_base::showpoint);
    std::cout.precision(9);

    for(size_t n = 0; n < frames.size(); ++n){

	const size_t nmolec = frames[n].natm/3;

	double orig_dip[3];
	std::fill(orig_dip, orig_dip + 3, 0.0);

	for(size_t i = 0; i < nmolec; ++i){
	    size_t i3 = 3*i;
	    for(size_t j = 0; j < 3; ++j){
		orig_dip[j]  += frames[n].dipmol[i3 + j] 
  	                      + frames[n].dipind[i3 + j];
	    }
	}

	const double orig_dip_mag = std::sqrt(orig_dip[0]*orig_dip[0]
				            + orig_dip[1]*orig_dip[1]
					    + orig_dip[2]*orig_dip[2]);

	std::cout << frames[n].time << ' ' 
	          << orig_dip[0]    << ' ' 
	          << orig_dip[1]    << ' ' 
		  << orig_dip[2]    << ' '
		  << orig_dip_mag   << std::endl;

    }

// write cluster dipole moments

//change for-loop limit from frames... -> 1
//to create an ion centered cluster defined by 
//radial cut-off macros

    for(size_t i = 0; i < frames[0].natm; i+=3){

	std::ofstream cluster_file;
	std::string base = "dip_1B+NB_cluster_";
	std::string tail = ".dat";

	std::ostringstream oss;
	oss << (i/3);
	std::string file_num;
	file_num = oss.str();

	std::string filename = base + file_num + tail;

	cluster_file.open(filename.c_str());
	
	//cycle through the trajectory to print out dynamic cluster information

	for(size_t k = 0; k < frames.size(); ++k){

	    size_t radial_cutoff_index = 0;
            
	    double orig_dip_distance[3];
            std::fill(orig_dip_distance, orig_dip_distance + 3, 0.0);

	    //need an ordered array of water molecules by distance from cluster center

	    double ordered_distances[frames[k].natm/3];
	    std::fill(ordered_distances, ordered_distances + frames[k].natm/3, 0.0);

	    size_t ordered_keys[frames[k].natm/3];
	    std::fill(ordered_keys, ordered_keys + frames[k].natm/3, 0);
	    
	    for(size_t m = 0; m < frames[k].natm; m+=3){

		size_t m3 = 3*m; //index of molecular oxygen wrt cluster center
		size_t i3 = 3*i; //index of cluster center oxygen

		double d_xyz[3];

		for(size_t o = 0; o < 3; ++o){

		    d_xyz[o] = frames[k].xyz[m3+o] - frames[k].xyz[i3+o];

		    if (d_xyz[o] >   BOX_WIDTH/2.0) d_xyz[o] -= BOX_WIDTH;		
		    if (d_xyz[o] <= -BOX_WIDTH/2.0) d_xyz[o] += BOX_WIDTH;		

		}


		double dist_from_cluster_center = std::sqrt((d_xyz[0] * d_xyz[0])
				+ (d_xyz[1] * d_xyz[1]) + (d_xyz[2] * d_xyz[2]));

		ordered_distances[(m/3)] = dist_from_cluster_center;
		ordered_keys[(m/3)] = (m/3);

	    }

	    //insertion sort-orders the arrays for both the distances
	    //and the keys

	    for(size_t m = 1; m < frames[k].natm/3; ++m){
	   
		for (size_t x = m; x>0 && ordered_distances[x] < ordered_distances[x-1]; --x){

		    double temp_dist = ordered_distances[x];
		    size_t temp_index = ordered_keys[x];

		    ordered_distances[x] = ordered_distances[x-1];
		    ordered_distances[x-1] = temp_dist;
		    
		    ordered_keys[x] = ordered_keys[x-1];
		    ordered_keys[x-1] = temp_index;

		}
	    } 
	
	    while(ordered_distances[radial_cutoff_index] < OUTER_SHELL_RADIUS) ++radial_cutoff_index;

	    //build up the cluster dipoles using the sorted keys

	    for(size_t p = 0; p < radial_cutoff_index; ++p){

		size_t relevant_index = ordered_keys[p];
		size_t rel_ind_3 = 3*relevant_index;

//		if (relevant_index == 0) print_cluster = false; //if HOD enters the picture, discard
		
		for(size_t c = 0; c < 3; ++c){
			
			double c_component = frames[k].dipmol[rel_ind_3 + c]
					   + frames[k].dipind[rel_ind_3 + c];

			double weighted_dipole_contribution = 
			    switching_function(ordered_distances[p]) * c_component;

			orig_dip_distance[c] += weighted_dipole_contribution;

		}

	    }
	
            const double orig_dip_distance_mag = std::sqrt(orig_dip_distance[0]*orig_dip_distance[0]
                                                         + orig_dip_distance[1]*orig_dip_distance[1]
                                                         + orig_dip_distance[2]*orig_dip_distance[2]);

	    cluster_file << frames[k].time          << ' '
                         << orig_dip_distance[0]    << ' '
                         << orig_dip_distance[1]    << ' '
                         << orig_dip_distance[2]    << ' '
                         << orig_dip_distance_mag   << std::endl;

	    }
	

	cluster_file.close();

    }


// write molecule zero dipole (specific for HOD-h2o sims)    

/*

    for(size_t n = 0; n < (frames[0].natm/3); ++n){

        std::ofstream molecular_file;

        std::string base = "dip1b_1B_mol_";
        std::string tail = ".dat";

	std::ostringstream oss;
	oss << n;
	std::string file_num;
	file_num = oss.str();

	std::string filename = base + file_num + tail;
	molecular_file.open(filename.c_str());

	for(size_t k = 0; k < frames.size(); ++k){

            const size_t nmolec = frames[k].natm/3;

            double orig_dip[3];
            std::fill(orig_dip, orig_dip + 3, 0.0);

            for(size_t i = n; i < (n+1); ++i){
                size_t i3 = 3*i;
                for(size_t j = 0; j < 3; ++j){
                    orig_dip[j]  += frames[k].dipmol[i3 + j]
                                    + frames[k].dipind[i3 + j];
                }
            }

            const double orig_dip_mag = std::sqrt(orig_dip[0]*orig_dip[0]
                                                + orig_dip[1]*orig_dip[1]
                                                + orig_dip[2]*orig_dip[2]);

            molecular_file << frames[k].time << ' '
                           << orig_dip[0]    << ' '
                           << orig_dip[1]    << ' '
                           << orig_dip[2]    << ' '
                           << orig_dip_mag   << std::endl;

	}

 

        molecular_file.close();

    }
*/
}

