#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include "read-hist.h"
#include "read-xyz.h"
#include "constants.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static std::vector<opt::hist_frame> frames;

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////
double INNER_SHELL_RADIUS;
double OUTER_SHELL_RADIUS;
double BOX_WIDTH;
double INTEGRATION_TIME;
double TIME_STEP;

double switching_function(double radius){

    //look up formula from Adaptive QM/MM paper 2012 

    if (radius <= INNER_SHELL_RADIUS) return 1.0;
    else if (radius >= OUTER_SHELL_RADIUS) return 0.0;

    double range = OUTER_SHELL_RADIUS - INNER_SHELL_RADIUS;
    double first_term_numerator = (radius - INNER_SHELL_RADIUS);
    first_term_numerator *= first_term_numerator;

    double second_term_numerator = (3*OUTER_SHELL_RADIUS - INNER_SHELL_RADIUS
				   - 2*radius);

    double result = first_term_numerator * second_term_numerator;
    result /= (range*range*range);

    result = 1 - result;

    return result;

}

int main(int argc, char** argv)
{
    if (argc != 5) {
       std::cerr << "usage: bulk_dipole INNER_RADIUS BOX_WIDTH INT_TIME(ps) "
		 << "TIME_STEP(ps) > STDOUT" << std::endl;
        return EXIT_FAILURE;
    }

    std::istringstream iss(argv[1]);
    iss >> INNER_SHELL_RADIUS;

    std::istringstream is(argv[2]);
    is >> BOX_WIDTH;

    std::istringstream iis(argv[3]);
    iis >> INTEGRATION_TIME;

    std::istringstream isi(argv[4]);
    isi >> TIME_STEP;

    int num_data = int(INTEGRATION_TIME / TIME_STEP);

    OUTER_SHELL_RADIUS = INNER_SHELL_RADIUS + 1.0;

    std::cout << INNER_SHELL_RADIUS << '\t' << OUTER_SHELL_RADIUS << '\t'
	      << INTEGRATION_TIME << '\t' << TIME_STEP << '\t' << BOX_WIDTH << std::endl;

    try {

	std::vector<opt::hist_frame> this_file;
    	opt::read_cmd(this_file, true);
    	std::copy(this_file.begin(), this_file.end(),
    		  std::back_inserter(frames));

//	std::cerr <<" CMD : " << this_file.size() << std::endl;

    } catch (const std::exception& e) {
	std::cerr << e.what() << std::endl;
	return EXIT_FAILURE;
    }

    std::vector<std::string> atom_labels;
    std::vector<double> frame_x_vel;
    std::vector<double> frame_y_vel;
    std::vector<double> frame_z_vel;

    std::ifstream velocity_cmd_file;
    velocity_cmd_file.open("VELOCITY_CMD");

    while(!velocity_cmd_file.eof()){

	std::string label;
	double v_x, v_y, v_z;
	
	velocity_cmd_file >> label >> v_x >> v_y >> v_z;

	atom_labels.push_back(label);
	frame_x_vel.push_back(v_x);
	frame_y_vel.push_back(v_y);
	frame_z_vel.push_back(v_z);
	
    }

//    Open files for Writing

    std::cout.setf(std::ios_base::showpoint);
    std::cout.precision(9);

    std::ofstream bulk_vel_file;
    bulk_vel_file.open("c_vv_bulk.dat");

    //compute the box atomic velocity tcf
    //LIMIT IS HARDCODED FOR 5ps USING A 2fs DUMPING FREQUENCY

    double num_entries[num_data];
    double cor_vv_bulk[num_data];

    std::fill(num_entries, num_entries + num_data, 0.0);
    std::fill(cor_vv_bulk, cor_vv_bulk + num_data, 0.0);

    for(size_t tau = 0; tau < frames.size(); ++tau){

	for(size_t n = tau; n < (tau + num_data) && n < frames.size(); ++n){

	    //compute the dot-product of the vector at time tau with
	    //box atomic velocity at time n

	    double v0_dot_vt = 0.0;

	    for(size_t vel_index = 0; vel_index < frames[0].natm; ++vel_index){

		v0_dot_vt += frame_x_vel[(tau*frames[0].natm) + vel_index]
			   * frame_x_vel[(  n*frames[0].natm) + vel_index];
		v0_dot_vt += frame_y_vel[(tau*frames[0].natm) + vel_index]
			   * frame_y_vel[(  n*frames[0].natm) + vel_index];
		v0_dot_vt += frame_z_vel[(tau*frames[0].natm) + vel_index]
			   * frame_z_vel[(  n*frames[0].natm) + vel_index];

	    }

	    size_t time = n - tau;
	    cor_vv_bulk[time] += v0_dot_vt;
	    ++num_entries[time];

	}

    }

    // changed size_t to int
    for(int index = 0; index < num_data; ++index){

	bulk_vel_file << (double(index) * TIME_STEP) << '\t'
		      << (cor_vv_bulk[index] / (cor_vv_bulk[0]*num_entries[index])) << '\n';

    }

// write cluster velocity moments

//change for-loop limit from frames... -> 1
//to create an ion centered cluster defined by 
//radial cut-off macros

    for(size_t i = 0; i < frames[0].natm; i+=3){

	std::ofstream cluster_file;
	std::string base = "vel_corr_cluster_";
	std::string tail = ".dat";

	std::ostringstream oss;
	oss << (i/3);
	std::string file_num;
	file_num = oss.str();

	std::string filename = base + file_num + tail;

	
	//cycle through the trajectory to store dynamic cluster information

	std::vector<double> cluster_x_vel; 
        std::vector<double> cluster_y_vel; 
	std::vector<double> cluster_z_vel; 
	
	for(size_t k = 0; k < frames.size(); ++k){
	    
	    std::vector<double> distances; 

	    for(size_t m = 0; m < frames[k].natm; ++m){

		size_t m3 = 3*m; // m = index of atom wrt cluster center
		size_t i3 = 3*i; // index of cluster center oxygen

		double d_xyz[3];

		for(size_t o = 0; o < 3; ++o){

		    d_xyz[o] = frames[k].xyz[m3 + o] - frames[k].xyz[i3 + o];

		    if (d_xyz[o] >   BOX_WIDTH/2.0) d_xyz[o] -= BOX_WIDTH;		
		    if (d_xyz[o] <= -BOX_WIDTH/2.0) d_xyz[o] += BOX_WIDTH;		

		}

		double dist_from_cluster_center = std::sqrt((d_xyz[0] * d_xyz[0])
				+ (d_xyz[1] * d_xyz[1]) + (d_xyz[2] * d_xyz[2]));

		distances.push_back(dist_from_cluster_center);

	    }
	    //build up the cluster velocity vectors using the distances 

	    for(size_t p = 0; p < frames[k].natm; ++p){

		double weight = switching_function(distances[p]);
		
		double weighted_v_x_contribution = 
		    weight * frame_x_vel[(k*frames[0].natm) + p];
		double weighted_v_y_contribution = 
		    weight * frame_y_vel[(k*frames[0].natm) + p];
		double weighted_v_z_contribution = 
		    weight * frame_z_vel[(k*frames[0].natm) + p];

		cluster_x_vel.push_back(weighted_v_x_contribution);	
		cluster_y_vel.push_back(weighted_v_y_contribution);	
		cluster_z_vel.push_back(weighted_v_z_contribution);	

		if (i == 0){

		    std::cerr << std::scientific;

	    	    std::cerr << cluster_x_vel[p] << '\t' << cluster_y_vel[p] 
	    		      << '\t' << cluster_z_vel[p] << '\t'
	    		      << distances[p] << '\t' << weight << std::endl;
		}

	    }

	}

	//data now stored in cluster_vel[(frame_index)*num_atoms + atom_index]
	//compute the autocorrellation

	double num_clust_entries[num_data];
	double corr_vv_cluster[num_data];
	
	std::fill(num_clust_entries, num_clust_entries + num_data, 0.0);
	std::fill(corr_vv_cluster, corr_vv_cluster + num_data, 0.0);

	for(size_t tau = 0; tau < frames.size(); ++tau){

	    for(size_t n = tau; n < (tau + num_data) && n < frames.size(); ++n){

	    //compute the dot-product of the vector at time tau with
	    //cluster atomic velocity at time n

		double v0_dot_vt = 0.0;

		for(size_t vel_index = 0; vel_index < frames[0].natm; ++vel_index){

		    v0_dot_vt += cluster_x_vel[(tau * frames[0].natm) + vel_index]
		    	       * cluster_x_vel[(  n * frames[0].natm) + vel_index];
		    v0_dot_vt += cluster_y_vel[(tau * frames[0].natm) + vel_index]
			       * cluster_y_vel[(  n * frames[0].natm) + vel_index];
 		    v0_dot_vt += cluster_z_vel[(tau * frames[0].natm) + vel_index]
			       * cluster_z_vel[(  n * frames[0].natm) + vel_index];

 		}
	    
		size_t time = n - tau;
		corr_vv_cluster[time] += v0_dot_vt;
		++num_clust_entries[time];
	    }
	}
	
	cluster_file.open(filename.c_str());

	// changed size_t to int
	for(int index = 0; index < num_data; ++index){

	    cluster_file << (double(index) * TIME_STEP) << '\t'
			 << (corr_vv_cluster[index] / (corr_vv_cluster[0]*num_clust_entries[index])) << '\n';

	}
	
	cluster_file.close();

    }

    return 0;

}
