#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <sstream>
#include <cstdlib>
#include <iomanip>
#include <fstream>
#include <iostream>

#include <cmath>

#include "fgh.h"
#include "interpolation.h"

#include "mat_tools.h"
#include "constants.h"

#include "read-scan.h"

namespace {

//----------------------------------------------------------------------------//

const double PI = M_PI;
const double au2kcal = 627.509;
const double au2ang = 0.529177249;
const size_t nsample = 128;

const double amu2au = 1822.88839;
const double massH = 1.00794*amu2au;    // au
const double massD = 2.01410178*amu2au; // au
const double massO = 15.999*amu2au;     // au

double reduced_mass = 1;

//----------------------------------------------------------------------------//

void compute_eigenvalues(tools::scan_frame& scan, double* eig)
{

    double E[scan.npoints];
    double rOX[scan.npoints];
    for(size_t i = 0; i < scan.npoints; ++i){
	E[i]   = scan.E[i]   / au2kcal;
	rOX[i] = scan.rOX[i] / au2ang;
    }

    const double length = (rOX[scan.npoints-1] - rOX[0]);
    const double     dx = length/nsample;

    double v_interpolated[nsample];

    double min(1e40);
    for(size_t i = 0; i < nsample; ++i){
	double xq = dx*i + rOX[0];
	v_interpolated[i] = tools::cubic_spline(scan.npoints, rOX, E, xq);

	if(v_interpolated[i] < min)
	    min = v_interpolated[i];
    }

    for(size_t i = 0; i < nsample; ++i)
	v_interpolated[i] -= min;

    double H[nsample*nsample];
    std::fill(H, H + nsample*nsample, 0.0);

    tools::cpt_H(nsample, length, v_interpolated, reduced_mass, H);

    int status = tools::eigenvalues(nsample, H, eig);

    if(status != 0){
	std::cerr << "Failed to diagonlize matrix, time = " 
	          << scan.time << std::endl;

	for(size_t i = 0; i < scan.npoints; ++i)
	    std::cerr << scan.rOX[i] << "  " << scan.E[i] << std::endl;

    }

}

//----------------------------------------------------------------------------//

void calculate_frequencies(std::ifstream& ifs)
{

    size_t frameno(0);

    while(!ifs.eof()){ // read through the file

	// read scan position and energies for a single frame
	size_t lineno(0);
	tools::scan_frame scan;
	tools::read_single_scan(lineno, ifs, scan);

	if(scan.npoints > 0){

    	    // do spline interpolation, calculate and diagonalize Hamiltonian
	    double eig[nsample];
	    compute_eigenvalues(scan, eig);

    	    // Report
	    std::cout << std::setw(15) << scan.time;
	    for(size_t i = 1; i < 3; ++i)
		std::cout << std::setw(15) 
		          << (eig[i] - eig[i-1])*constants::Eh_cm1;
	    std::cout << std::endl;
	}

    }

}

//----------------------------------------------------------------------------//

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{

    if(argc < 3){
     	std::cerr << "usage: ./freq-calculator isotope# scan.dat > freq.dat\n" 
	          << "       isotope#: 1 == H, 2 == D"
	          << std::endl;
	exit(1);
    }

    { // Determine reduced mass

	std::istringstream iss(argv[1]);
	int isotope;
	iss >> isotope;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[1]
	 	      << "' to an integer" << std::endl;
	    return EXIT_FAILURE;
	}
	double ma;
	if(isotope == 2)
	    ma = massD;
	else
	    ma = massH;
	const double mb = massO;
	reduced_mass = ma*mb/(ma+mb);

    }

    std::cout.precision(6);
    std::cout.setf( std::ios::fixed, std:: ios::floatfield );

    for(int i = 2; i < argc; ++i){
	
	std::ifstream ifs(argv[2]);

	calculate_frequencies(ifs);

    }



    return 0;
}
