#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <mpi.h>

#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <utility>
#include <algorithm>

#include "read-hist.h"
#include "read-xyz.h"
#include "constants.h"

#include "ttm4-dip.h"
#include "ltp2011.h"

#include "dip_poly2b5.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////


namespace t = ttm;

typedef ttm::ttm4_dip ind_model;
static h2o::ltp2011 dip1b_model;

const double r2i = 5.5;
const double r2f = 7.5;

//----------------------------------------------------------------------------//

inline double distance(size_t imcon, const double* cell, 
		       const double* r1, const double* r2)
{
    assert(imcon == 0 || imcon == 1);

    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    dx -= cell[0]*std::floor(dx/cell[0] + 0.5);
    dy -= cell[4]*std::floor(dy/cell[4] + 0.5);
    dz -= cell[8]*std::floor(dz/cell[8] + 0.5);

    return std::sqrt(dx*dx + dy*dy + dz*dz);
}

//----------------------------------------------------------------------------//

inline void image(const double* cell, const double* r1, const double* r2, 
		  double* crd_out)
{
    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    double facx = std::floor(dx/cell[0] + 0.5);
    double facy = std::floor(dy/cell[4] + 0.5);
    double facz = std::floor(dz/cell[8] + 0.5);

    crd_out[0] = r2[0] + cell[0]*facx;
    crd_out[1] = r2[1] + cell[4]*facy;
    crd_out[2] = r2[2] + cell[8]*facz;

}

//----------------------------------------------------------------------------//

inline void image_dimer(const double* cell, 
      			const double* r1, const double* r2,
		      	double* dimer)
{
    std::fill(dimer, dimer+18, 0.0);

    // take r1 oxygen as central atom, image its two hydrogens
     
    for(size_t i = 0; i < 3; ++i)
	dimer[i] = r1[i];
    
    for(size_t a = 1; a < 3; ++a)
	image(cell, r1, r1 + 3*a, dimer + 3*a);

    // Now image the oxygen of the second molecule
     
    image(cell, r1, r2, dimer + 9);

    // Finally, reattach hydrogens of molec two to oxygen of molec 2
    for(size_t a = 1; a < 3; ++a)
	image(cell, dimer+9, r2 + 3*a, dimer + 9 + 3*a);
}

//----------------------------------------------------------------------------//

double f_switch2(const double& r)
{
    if (r > r2f)
	return 0.0;
    else if (r > r2i) {
	const double x = (r - r2i)/(r2f - r2i);
	return 1 + x*x*(2*x - 3);
    } else
	return 1.0;
}

////////////////////////////////////////////////////////////////////////////////

  static int world_rank(0);
  static int world_size(0);

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{

    MPI_Init(&argc, &argv);
    
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &world_size);

    if (argc != 5) {
        std::cerr << "usage: bulk_dipole POSITION_CMD DIPMOL_CMD DIPIND_CMD mb" << std::endl;
        MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }

    // Open files for Writing

    std::cout.setf(std::ios_base::showpoint);
    std::cout.precision(9);

    size_t lineno(0);

    ind_model pot_ind;

    double* dimer = new double[18];

    // begin the parallelization over frames
    
    std::ifstream ifs_cmd(argv[1]);
    std::ifstream ifs_mol(argv[2]);
    std::ifstream ifs_ind(argv[3]);
    
    // Toggles names: 1b, 1bnb, mb
    std::string mb = argv[4];
    std::transform(mb.begin(), mb.end(), mb.begin(), ::tolower);

    while(!ifs_cmd.eof() && !ifs_mol.eof() && !ifs_ind.eof()) {

        std::vector<opt::hist_frame> frames;

        // reads in the number of frames equal to num of processors for MPI
        while((int) frames.size() < world_size && !ifs_cmd.eof() && !ifs_mol.eof() && !ifs_ind.eof()){
            opt::hist_frame this_frame;
            opt::read_cmd_frame(lineno, ifs_cmd, this_frame);
            if (ifs_cmd.eof()) break;
            if (this_frame.natm == 0) break;
            opt::collect_multipoles(ifs_mol, ifs_ind, this_frame);
            if (ifs_mol.eof() || ifs_ind.eof()) break;
            frames.push_back(this_frame);
        }

        int n_frames = frames.size();

        if (n_frames == 0) break;

        opt::hist_frame frame;
   
        double total_dip[3];
        double dip_mag = 0.0;

        if (n_frames > world_rank) {

            frame = frames[world_rank];

            const size_t nmolec = frame.natm/3;
            std::fill (total_dip, total_dip + 3, 0.0);
      
            double dip2b_corr[3];
            double dip2b_ind[3];

            // ONLY FOR MB, two body
            if (mb.compare("mb") == 0) {
                std::fill(dip2b_corr, dip2b_corr + 3, 0.0);
                std::fill(dip2b_ind, dip2b_ind + 3, 0.0);

	        for (size_t i = 0; i < nmolec; ++i) {
	            for (size_t j = i+1; j < nmolec; ++j) {
    		        const size_t i9 = 9*i;
    		        const size_t j9 = 9*j;
    		        double rOO = distance(frame.imcon, frame.cell, 
			                      frame.xyz + i9, frame.xyz + j9);

		        if (rOO < r2f) {
		            const double s = f_switch2(rOO);
		            image_dimer(frame.cell, 
				        frame.xyz + i9, frame.xyz + j9, dimer);

		            // Get corr two-body dipole
		            {
		       	        double tmp_dip2b[3];
			        h2o::dip_poly2b5::eval(dimer, dimer + 9, tmp_dip2b);

			        for (size_t b = 0; b < 3; ++b)
		    	            dip2b_corr[b] += s*tmp_dip2b[b];
		            }

		            // Get TTM4 two-body dipole
		            {
			        double tmp_dip2b[3];
			        pot_ind(2, dimer, tmp_dip2b);

			        for(size_t b = 0; b < 3; ++b)
		    	            dip2b_ind[b] += s*tmp_dip2b[b];

		            }

		        }

	            }
	        }
            }

            // for 1BNB and 1B
            if (mb.compare("mb") != 0) {
                for (size_t i = 0; i < nmolec; ++i) {
                    size_t i3 = 3*i;
                    for (size_t j = 0; j < 3; ++j) {
                        total_dip[j]  += frame.dipmol[i3 + j];
                        
                        // for 1BNB
                        if (mb.compare("1b") != 0) 
                            total_dip[j] += frame.dipind[i3 + j];
                    }
                }
            }
        
            // for MB
            if (mb.compare("mb") == 0) {
	        for(size_t i = 0; i < nmolec; ++i){
	            size_t i3 = 3*i;
	            size_t i9 = 9*i;

	            double monomer[9];
	            std::copy(frame.xyz + i9, frame.xyz + i9 + 3, monomer);
	            for(size_t a = 1; a < 3; ++a)
		        image(frame.cell, frame.xyz + i9,
		              frame.xyz + i9 + 3*a, monomer + 3*a);

	            double dip_mol[3];
	            dip1b_model.cartesian(1, monomer, dip_mol);

	            for (size_t j = 0; j < 3; ++j) {
		        total_dip[j] += dip_mol[j];
                        total_dip[j] += frame.dipind[i3 + j];
	            }
	        }
            }

	    // Substitute two-body dipole contribution from induction with corr
	    // for MB
	    if (mb.compare("mb") == 0) {
	        for (size_t j = 0; j < 3; ++j) {
	            total_dip[j] += (dip2b_corr[j] - dip2b_ind[j]);
    	        }
            }

	    dip_mag = std::sqrt(total_dip[0]*total_dip[0]
			   	      + total_dip[1]*total_dip[1]
				      + total_dip[2]*total_dip[2]);

        }

        double all_frame_time[1 * world_size];
        MPI_Gather(&frame.time, 1, MPI_DOUBLE, &all_frame_time, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

        double all_frame_dip[3 * world_size];
        MPI_Gather(&total_dip, 3, MPI_DOUBLE, &all_frame_dip, 3, MPI_DOUBLE, 0, MPI_COMM_WORLD);

        double all_frame_dip_mag[1 * world_size];
        MPI_Gather(&dip_mag, 1, MPI_DOUBLE, &all_frame_dip_mag, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

        if (world_rank == 0){
            for (int i = 0; i < n_frames; ++i)
            std::cout << all_frame_time[i] << ' '
                      << std::setw(14) << all_frame_dip[i * 3 + 0] << ' ' 
                      << std::setw(14) << all_frame_dip[i * 3 + 1] << ' ' 
                      << std::setw(14) << all_frame_dip[i * 3 + 2] << ' ' 
                      << std::setw(14) << all_frame_dip_mag[i] << std::endl;
        }
    
    }

    delete[] dimer;
    MPI_Finalize();
}

