#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include "read-hist.h"
#include "read-xyz.h"
#include "constants.h"

#include "qtip4pf.h"

//#define HIST yes

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static std::vector<opt::hist_frame> frames;

//----------------------------------------------------------------------------//

inline double distance(size_t imcon, const double* cell, 
		       const double* r1, const double* r2)
{
    assert(imcon == 0 || imcon == 1);

    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    dx -= cell[0]*std::floor(dx/cell[0] + 0.5);
    dy -= cell[4]*std::floor(dy/cell[4] + 0.5);
    dz -= cell[8]*std::floor(dz/cell[8] + 0.5);

    return std::sqrt(dx*dx + dy*dy + dz*dz);
}

//----------------------------------------------------------------------------//

inline void image(const double* cell, const double* r1, const double* r2, 
		  double* crd_out)
{
    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    double facx = std::floor(dx/cell[0] + 0.5);
    double facy = std::floor(dy/cell[4] + 0.5);
    double facz = std::floor(dz/cell[8] + 0.5);

    crd_out[0] = r2[0] + cell[0]*facx;
    crd_out[1] = r2[1] + cell[4]*facy;
    crd_out[2] = r2[2] + cell[8]*facz;
}

//----------------------------------------------------------------------------//

inline void image_monomer(const double* cell, 
			  const double* r1, double* mon)
{
    std::fill(mon, mon+9, 0.0);

    // take r1 oxygen as central atom, image its two hydrogens
     
    for(size_t i = 0; i < 3; ++i)
	mon[i] = r1[i];
    
    for(size_t a = 1; a < 3; ++a)
	image(cell, r1, r1 + 3*a, mon + 3*a);
}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 1) {
        std::cerr << "usage: bulk_dipole ..." << std::endl;
        return EXIT_FAILURE;
    }

    try {

#ifdef HIST
	opt::read_hist(argv[1], frames, false);
#else
    	opt::read_cmd(frames, false);
#endif

	std::cerr <<" CMD : " << frames.size() << std::endl;

    } catch (const std::exception& e) {
	std::cerr << e.what() << std::endl;
	return EXIT_FAILURE;
    }

//    Open files for Writing

    std::ofstream dipout;
    dipout.setf(std::ios_base::showpoint);
    dipout.precision(9);
    dipout.open("dip.qtip4pf");

    h2o::qtip4pf model;

    for(size_t n = 0; n < frames.size(); ++n){

	const size_t nmolec = frames[n].natm/3;
	double total_dip[3];
	std::fill(total_dip, total_dip + 3, 0.0);

//	std::cout << 256*3 << std::endl << std::endl;
	for(size_t i = 0; i < nmolec; ++i){
	    const size_t i9 = 9*i;

	    double mon[9];
	    image_monomer(frames[n].cell, frames[n].xyz + i9, mon);
//	    for(size_t a = 0; a < 3; ++a){
//	        if(a = 0)
//	            std::cout << "O ";
//	        else
//	            std::cout << "H ";
//		for(size_t b = 0; b < 3; ++b)
//		    std::cout << mon[3*a + b] << "  ";
//		std::cout << std::endl;
//	    }

	    double tmp_dip[3];
	    model(1, mon, tmp_dip);

	    for(size_t j = 0; j < 3; ++j)
		total_dip[j] += tmp_dip[j];

	}

	const double dip_mag = std::sqrt(total_dip[0]*total_dip[0]
			   	       + total_dip[1]*total_dip[1]
				       + total_dip[2]*total_dip[2]);

	dipout << frames[n].time << ' ' 
	         << total_dip[0]   << ' ' 
	         << total_dip[1]   << ' ' 
		 << total_dip[2]   << ' '
		 << dip_mag        << std::endl;

    }

    dipout.close();
}

