#include <cmath>
#include <vector>
#include <iomanip>
#include <iostream>

#include <cassert>
#include <cstdlib>
#include <sstream>

namespace {

bool normalize = false;
double tau = 0.0;
double alpha = 0.0;
double max_int = 0.0;
const size_t num_nu = 10000; // frequency grid
const double omega_max = 9000.0;
const double ps2s = 1.0e-12; // s -> ps
const double pi2 = 2.0*M_PI;

const double c = 2.99792458e-02; // x cm/ps (speed of light)

const double hplanck = 6.62606876e-34;
const double hbar = hplanck/(2.0*M_PI);
const double kboltz = 1.3806503e-23;
const double epsilon_vacuum = 8.8541878e-12;
const double light_velocity = 2.99792e8; // m/s
const double conv_freq =  2.99792e10;

//const double length = 19.7291394*1e-10; // m
//const double volume = length*length*length; // m^3

//const double T = 298.15; // K

} // namespace

int main(int argc, char** argv)
{
    double volume(0.0);
    double T(0.0);

    if (argc != 6) {
	std::cerr << "usage: cos-transform T volume tau alpha max < corr_dip.dat > spectrum.dat" << std::endl;
	return EXIT_FAILURE;
    }

    {
        std::istringstream iss(argv[1]);
        iss >> T;
	if(T == 0){
	    T = 298.15;
	}
        if (!iss || !iss.eof()) {
    	std::cerr << "could not convert '" << argv[1]
    	    << "' to real number" << std::endl;
    	return EXIT_FAILURE;
        }
    
        assert(T > 0.0);
    }
    {
        std::istringstream iss(argv[2]);
        iss >> volume;
	if(volume == 0){
	    normalize = true;
	    volume = 7679.34930337;
	}
        if (!iss || !iss.eof()) {
    	std::cerr << "could not convert '" << argv[2]
    	    << "' to real number" << std::endl;
    	return EXIT_FAILURE;
        }
    
        assert(volume > 0.0);
	volume *= 1e-30; // A^3 -> m^3
    }
    {
        std::istringstream iss(argv[3]);
        iss >> tau;
        if (!iss || !iss.eof()) {
    	std::cerr << "could not convert '" << argv[3]
    	    << "' to real number" << std::endl;
    	return EXIT_FAILURE;
        }
    }
    {
        std::istringstream iss(argv[4]);
        iss >> alpha;
        if (!iss || !iss.eof()) {
    	std::cerr << "could not convert '" << argv[4]
    	    << "' to real number" << std::endl;
    	return EXIT_FAILURE;
        }
    }
    {
        std::istringstream iss(argv[5]);
        iss >> max_int;
        if (!iss || !iss.eof()) {
    	std::cerr << "could not convert '" << argv[5]
    	    << "' to real number" << std::endl;
    	return EXIT_FAILURE;
        }
    }

    std::vector<double> time;
    std::vector<double> corr;

    size_t Ndata = 0;

    while (std::cin) {
    	double t, c;
	std::cin >> t >> c >> std::ws;
	if(max_int > 0 && t >= max_int){
	    time.push_back(t);
	    corr.push_back(0.0);
	}else if (std::cin) {
	    ++Ndata;
	    time.push_back(t);
	    corr.push_back(c);
	}
    }

    std::cerr << "Ndata : " << Ndata << std::endl;
    corr[Ndata] = 0.0;


    double arg = 2.0 * M_PI / (double(2 * Ndata - 1));

    double kT = kboltz * T;
    double beta = 1/kT;
    double norm =(3.0 * hbar * light_velocity * volume * epsilon_vacuum);
    norm = 2.0 * M_PI / norm;

    double dt = (time[1] - time[0]) * ps2s;
    double dnu = 1.0 / (double(2 * Ndata - 1)) / dt;
    double arg1 = beta * hplanck * dnu;

    if(tau != 0)
        for (size_t t = 0; t < Ndata; ++t)
            corr[t] *= std::exp(-(t*dt)/(tau*ps2s));
    if(alpha != 0){
        const double alpha2 = alpha*alpha;
        for (size_t t = 0; t < Ndata; ++t){
            const double time2 = (t*dt)*(t*dt)/(ps2s*ps2s);
            corr[t] *= std::exp(-alpha2*time2);
	}
    }

    for (size_t inu = 0; inu < Ndata; ++inu){

	double nu = dnu * inu;
	double beta_hplanck_nu = arg1 * inu;
	double tmp = beta_hplanck_nu / 2.0;
	double factor_quantum = nu * tmp;

	if (inu == 0) factor_quantum = 1.0;

	double absorbance = 0.0;

	for (int itime = Ndata - 1; itime > 0; --itime){

	    double nu_time = (double(itime)) * (double(inu)) * arg;
	    absorbance += std::cos(nu_time) * corr[itime];

	}

	for (int itime = 0; itime < Ndata; ++itime){

	    double nu_time = (double(itime)) * (double(inu)) * arg;
	    absorbance += std::cos(nu_time) * corr[itime];

	}

	factor_quantum *= norm * dt;
	absorbance *= factor_quantum/100;
	std::cout << (nu/conv_freq) << '\t'<< absorbance << std::endl;

    }

    return 0;
    
}

