#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <cmath>
#include <algorithm>

#include <limits.h>

#include "read-hist.h"
#include "read-bin.h"
#include "pbc-tools.h"

#include "interpolation.h"
//#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>

#include "struct-instant-interface.h"

#define SHIFT_HALF_CELL yes

#define HIST yes

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

const double cg_length = 2.4;
const double cg_cutoff = cg_length*3;

// Grid for calculating coarse grained density (spacing is dx, dy, dz)
int ngrid_x = 0;
int ngrid_y = 0;
int ngrid_z = 0;

const double dx = 0.25;
const double dy = 0.25;
const double dz = 0.25;

// Step size for calculating gradient of CG density
const double eps = 0.0001;

bool first(true); // some values set when reading first frame
double half_z_length = 0;
double one_over_SA = 1;

// Read from command line
int n_skip(1);
double isovalue(0.0937);

// Used to discretize the density profiles
const double dr = 0.01;
const double r_max = 30.0;
const double r_min = -10.0;

const int nbins =  (int)((r_max-r_min)/dr);
double bin_instant[nbins];
//double bin_instant[(int)((r_max-r_min)/dr)];
size_t natm;

double xi = cg_length;
double xi2 = xi*xi;
double prefactor = std::pow(2*M_PI*xi2, -3/2);

//----------------------------------------------------------------------------//

bool crossed_isosurface(const double& v1, const double& v0)
{
    if( (v1 - isovalue)*(v0 - isovalue) < 0)
	return true;
    else
	return false;
}

//----------------------------------------------------------------------------//

double cg_density(const double& r)
{
    return prefactor*std::exp(-r*r/(2*xi2));
}

//----------------------------------------------------------------------------//

double calc_cg_value(const double* grid_point, const double* atm_xyz,
		     size_t& natm, size_t& imcon,
		     const double* cell, const double* rcell)
{
    double value(0);
    for(size_t j = 0; j < natm; j += 3){ // O only
	const double* qj = atm_xyz + 3*j;
	const double rij = tools::distance3(imcon, cell, rcell,
		    			    grid_point, qj);
	if(rij < cg_cutoff)
	    value += cg_density(rij);

    }
    return value;
}

//----------------------------------------------------------------------------//

size_t do_frames(size_t& lineno, std::ifstream& ifs)
{

    size_t frameno(0);

    while(!ifs.eof()){

	opt::hist_frame frame;

#ifdef HIST     
	opt::read_hist_frame(lineno, ifs, frame);
#else
	opt::read_cmd_frame(lineno, ifs, frame);
#endif

	if(frame.natm > 0){
	    ++frameno;

#ifdef SHIFT_HALF_CELL
	    tools::center_slab_at_origin(frame.natm, frame.cell, frame.mass,
					 frame.xyz, true);
#else
	    tools::center_slab_at_origin(frame.natm, frame.cell, frame.mass,
					 frame.xyz, false);
#endif

	    if(frameno % n_skip == 0){

	    // set up grid: dim_1 = cell_x / dx
	    //              dim_2 = cell_y / dy
	    //              dim_3 = cell_z / dz

    	    if(first){
		first = false;

		half_z_length = frame.cell[8]/2.0;
		one_over_SA = 1.0/(frame.cell[0]*frame.cell[4]);

		natm = frame.natm;

    		ngrid_x = (int)(frame.cell[0]/dx);
    		ngrid_y = (int)(frame.cell[4]/dy);
    		ngrid_z = (int)(frame.cell[8]/dz);
	    }

	    std::vector<instant::grid_point> surface_points;

	    // Calculate the coarse grained density on a grid. Locate gridpoints
	    // where the instantaneous interface lies and add those to the
	    // surface_points vector
	    for(int ix = 0; ix < ngrid_x; ++ix){
		const double x = (ix + 0.5)*dx;

		for(int iy = 0; iy < ngrid_y; ++iy){
		    const double y = (iy + 0.5)*dy;

		    double prev_value(0);
		    for(int iz = 0; iz < ngrid_z; ++iz){
			const double z = (iz + 0.5)*dz - half_z_length;

			if(std::fabs(z) < 15){
			    double grid_point[3] = {x, y, z};
			    double value = calc_cg_value(grid_point, frame.xyz,
							 frame.natm,
							 frame.imcon,
							 frame.cell,
							 frame.rcell);

    			    if( crossed_isosurface(value, prev_value) ){

    				instant::grid_point here;
				here.ix = ix;
				here.iy = iy;
				// linear interpolation to the isovalue
				double ratio = (isovalue - prev_value)
				                         / (value - prev_value);
				here.rz = ratio*dz + (z - dz);

    				surface_points.push_back(here);

			    }

			    prev_value = value;

    			    //avg += value;
			    //++num;
			}

		    } // for(int iz ...
		} // for(int iy ...
	    } // for(int ix ...

	    // calculate the gradient of the CG density at each surface point
	    for(size_t i = 0; i < surface_points.size(); ++i){
		const double x = (surface_points[i].ix + 0.5)*dx;
	    	const double y = (surface_points[i].iy + 0.5)*dy;
	    	const double z =  surface_points[i].rz;

		double dxyz[3] = {0.0, 0.0, 0.0};

		double grid_point[3] = {x, y, z};
		//const double v = calc_cg_value(grid_point, frame.xyz,
		//			       frame.natm, frame.imcon,
		//			       frame.cell, frame.rcell);

		// df/dq ~= [ f(q+h) - f(q-h) ] / 2h
		for(size_t j = 0; j < 3; ++j){
		    double q0 = grid_point[j];

		    grid_point[j] = q0 + eps;
		    const double vforw = calc_cg_value(grid_point, frame.xyz,
						       frame.natm, frame.imcon,
				      		       frame.cell, frame.rcell);

		    grid_point[j] = q0 - eps;
		    const double vback = calc_cg_value(grid_point, frame.xyz,
						       frame.natm, frame.imcon,
				      		       frame.cell, frame.rcell);

		    grid_point[j] = q0;
		    dxyz[j] = (vforw - vback)/(2*eps);
		}
		double norm = std::sqrt(dxyz[0]*dxyz[0] 
				      + dxyz[1]*dxyz[1] 
				      + dxyz[2]*dxyz[2]);
		for(size_t j = 0; j < 3; ++j){
		    dxyz[j] /= norm;
		    surface_points[i].drho[j] = dxyz[j];
		}

	    }

	    // Now loop over every oxygen and locate the closest surface point
	    for(size_t i = 0; i < frame.natm; i += 3){ // O only
		const double* qi = frame.xyz + 3*i;

		double closest_distance(1000);
		double closest_drij[3];
		int closest_index(0);

		for(size_t j = 0; j < surface_points.size(); ++j){
		    const double x = (surface_points[j].ix + 0.5)*dx;
		    const double y = (surface_points[j].iy + 0.5)*dy;
		    const double z =  surface_points[j].rz;

		    double surface_point[3] = {x, y, z};
		    double drij[3];
		    const double rij = tools::distance3(frame.imcon,
	    						frame.cell,
							frame.rcell,
							qi, surface_point,
							drij);

		    if(rij < closest_distance){
			closest_distance = rij;
			closest_index = j;
			for(size_t k = 0; k < 3; ++k)
			    closest_drij[k] = drij[k]; // displacement vector
		    }
		}

		{ // Add the closest distance to the bin (use correct Expr.)
		    double proximity(0);
		    for(size_t k = 0; k < 3; ++k)
			proximity += (closest_drij[k])
			                * surface_points[closest_index].drho[k];
		    
		    ++bin_instant[(int)((proximity-r_min)/dr)];
	    	}
	    }

	    } // if(frameno % n_skip == 0){

	} // if(frame.natm > 0){
    } // while(!ifs.eof()){
    return frameno;

}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 3) {
        std::cerr << "usage: dividing_surfaces n_skip POSITION_CMD POSITION_CMD ... > profile.dat" 
	          << std::endl;
        return EXIT_FAILURE;
    }

    // parse command line: n_skip, isovalue, position_files...
    { // read n_skip
	std::istringstream iss(argv[1]);
	size_t tmp;
	iss >> tmp;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[1]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
	if(tmp > 0){ // cannot be < 1
	    n_skip = tmp;
	}
    }

    { // read isovalue
	std::istringstream iss(argv[2]);
	double tmp;
	iss >> tmp;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[2]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
	if(tmp > 0){ // cannot be < 1
	    isovalue = tmp;
	}
    }

    size_t tot_frameno(0);

    std::cout << std::setprecision(10) << std::scientific;

    for(int i = 3; i < argc; ++i){

	std::ifstream ifs(argv[i]);
    	std::string line;
    	size_t lineno(0);
#ifdef HIST
	std::getline(ifs, line);
    	std::getline(ifs, line);
    	lineno += 2;
#endif

	const size_t frameno = do_frames(lineno, ifs);
	tot_frameno += frameno;

	std::cerr << argv[i] << " : " << frameno << std::endl;

    }

    for(size_t i = 0; i < nbins; ++i){
	const double z = dr*(i + 0.5) + r_min;

	std::cout << std::setw(25) << z
	          << std::setw(25) << bin_instant[i]/(natm/3.0)/tot_frameno
	          << std::setw(25) << bin_instant[i]/(natm/3.0)
	          << std::setw(25) << tot_frameno
		  << std::endl;
    }

}
