#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <cmath>
#include <algorithm>

#include <limits.h>

#include "read-hist.h"
#include "read-bin.h"
#include "pbc-tools.h"
#include "orient.h"

#include "struct-instant-interface.h"

#define SHIFT_HALF_CELL yes

#define HIST yes

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

const double cg_length = 2.4;
const double cg_cutoff = cg_length*3;

// Grid for calculating coarse grained density (spacing is dx, dy, dz)
int ngrid_x = 0;
int ngrid_y = 0;
int ngrid_z = 0;

const double dx = 0.25;
const double dy = 0.25;
const double dz = 0.25;

// Step size for calculating gradient of CG density
const double eps = 0.0001;

bool first(true); // some values set when reading first frame
double half_z_length = 0;
double one_over_SA = 1;

// Read from command line
int n_skip(1);
double isovalue(0.0937);

// HB definitions
const double hb_rOO(3.5); // Angstrom
const double hb_aHOO(30); // Degress

// Used to discretize the density profiles
const double dr = 0.01;
const double r_max = 30.0;
const double r_min = -10.0;

const int nbins =  (int)((r_max-r_min)/dr);
int bin_total[nbins];
int bin_other[nbins]; // all other HB types

int bin_0n[nbins];    // 0 HBs total,    non donor

int bin_1n[nbins];    // 1 HBs total,    non donor
int bin_1s[nbins];    // 1 HBs total, single donor

int bin_2n[nbins];    // 2 HBs total,    non donor
int bin_2s[nbins];    // 2 HBs total, single donor
int bin_2d[nbins];    // 2 HBs total, double donor

int bin_3n[nbins];    // 3 HBs total,    non donor
int bin_3s[nbins];    // 3 HBs total, single donor
int bin_3d[nbins];    // 3 HBs total, double donor

int bin_4n[nbins];    // 4 HBs total,    non donor
int bin_4s[nbins];    // 4 HBs total, double donor
int bin_4d[nbins];    // 4 HBs total, double donor

//int bin_total[(int)((r_max-r_min)/dr)];
//int bin_other[(int)((r_max-r_min)/dr)]; // all other HB types
//
//int bin_0n[(int)((r_max-r_min)/dr)];    // 0 HBs total,    non donor
//
//int bin_1n[(int)((r_max-r_min)/dr)];    // 1 HBs total,    non donor
//int bin_1s[(int)((r_max-r_min)/dr)];    // 1 HBs total, single donor
//
//int bin_2n[(int)((r_max-r_min)/dr)];    // 2 HBs total,    non donor
//int bin_2s[(int)((r_max-r_min)/dr)];    // 2 HBs total, single donor
//int bin_2d[(int)((r_max-r_min)/dr)];    // 2 HBs total, double donor
//
//int bin_3n[(int)((r_max-r_min)/dr)];    // 3 HBs total,    non donor
//int bin_3s[(int)((r_max-r_min)/dr)];    // 3 HBs total, single donor
//int bin_3d[(int)((r_max-r_min)/dr)];    // 3 HBs total, double donor
//
//int bin_4n[(int)((r_max-r_min)/dr)];    // 4 HBs total,    non donor
//int bin_4s[(int)((r_max-r_min)/dr)];    // 4 HBs total, double donor
//int bin_4d[(int)((r_max-r_min)/dr)];    // 4 HBs total, double donor

size_t natm;
size_t nw;

double xi = cg_length;
double xi2 = xi*xi;
double prefactor = std::pow(2*M_PI*xi2, -3/2);

//----------------------------------------------------------------------------//

void dump_1d_distrib(const int n_frame, const char* filename, const int* bin,
		     const int nbins, const double dr, const double min)
{
    std::ofstream ofs;
    ofs.open(filename);
    ofs.precision(9);

    ofs << std::scientific;

    for(int i = 0; i < nbins; ++i){
	const double z = dr*(i + 0.5) + min;

	ofs << std::setw(25) << z
	    << std::setw(25) << bin[i]/((double)(natm/3.0))/((double)n_frame)
    	    << std::setw(25) << bin[i]/(natm/3.0)
      	    << std::setw(25) << n_frame
	    << std::endl;
    }

    ofs.close();
}

//----------------------------------------------------------------------------//

bool crossed_isosurface(const double& v1, const double& v0)
{
    if( (v1 - isovalue)*(v0 - isovalue) < 0)
	return true;
    else
	return false;
}

//----------------------------------------------------------------------------//

double cg_density(const double& r)
{
    return prefactor*std::exp(-r*r/(2*xi2));
}

//----------------------------------------------------------------------------//

double calc_cg_value(const double* grid_point, const double* atm_xyz,
		     size_t& natm, size_t& imcon,
		     const double* cell, const double* rcell)
{
    double value(0);
    for(size_t j = 0; j < natm; j += 3){ // O only
	const double* qj = atm_xyz + 3*j;
	const double rij = tools::distance3(imcon, cell, rcell,
		    			    grid_point, qj);
	if(rij < cg_cutoff)
	    value += cg_density(rij);

    }
    return value;
}

//----------------------------------------------------------------------------//

double hb_angle(size_t& imcon, const double* cell, const double* rcell,
		const double* atm_xyz,
		const size_t iOd, const size_t iHd, double* r_OdOa)
{
    const double* qOd = atm_xyz + 3*iOd;
    const double* qHd = atm_xyz + 3*iHd;

    double r_OdHd[3];
    tools::distance3(imcon, cell, rcell, qOd, qHd, r_OdHd);

    return tools::angle_vect(r_OdHd, r_OdOa);
}

//----------------------------------------------------------------------------//

size_t do_frames(size_t& lineno, std::ifstream& ifs)
{

    size_t frameno(0);

    while(!ifs.eof()){

	opt::hist_frame frame;

#ifdef HIST     
	opt::read_hist_frame(lineno, ifs, frame);
#else
	opt::read_cmd_frame(lineno, ifs, frame);
#endif

	if(frame.natm > 0){
	    ++frameno;

#ifdef SHIFT_HALF_CELL
	    tools::center_slab_at_origin(frame.natm, frame.cell, frame.mass,
					 frame.xyz, true);
#else
	    tools::center_slab_at_origin(frame.natm, frame.cell, frame.mass,
					 frame.xyz, false);
#endif

	    if(frameno % n_skip == 0){

	    // set up grid: dim_1 = cell_x / dx
	    //              dim_2 = cell_y / dy
	    //              dim_3 = cell_z / dz

    	    if(first){
		first = false;

		half_z_length = frame.cell[8]/2.0;
		one_over_SA = 1.0/(frame.cell[0]*frame.cell[4]);

		natm = frame.natm;
		nw = frame.natm/3;

    		ngrid_x = (int)(frame.cell[0]/dx);
    		ngrid_y = (int)(frame.cell[4]/dy);
    		ngrid_z = (int)(frame.cell[8]/dz);
	    }

	         ////////////////////////////////////////////
	         //  Determine Hydrogen Bonding Topology   //
	         ////////////////////////////////////////////
	    
	    int total_hb[nw];
	    std::fill(total_hb, total_hb + nw, 0);
	    int donating_X_hb[nw];
	    std::fill(donating_X_hb, donating_X_hb + nw, 0);

	    for(size_t iw = 0; iw < nw; ++iw){ // O only
		const size_t i = 3*iw;
		const double* qi = frame.xyz + 3*i;

		for(size_t jw = iw + 1; jw < nw; ++jw){ // O only
		    const size_t j = 3*jw;
		    const double* qj = frame.xyz + 3*j;

		    double dr_OiOj[3];
		    const double rij = tools::distance3(frame.imcon,
	    						frame.cell,
							frame.rcell,
							qi, qj, dr_OiOj);

		    if(rij < hb_rOO){ // Oxygens are close enough, evalute aHOO

			//   H_i(a/b)
			//    /
			//   Oi --- Oj
			{
			    double aHOO;
			    aHOO = hb_angle(frame.imcon, frame.cell,
		     			    frame.rcell, frame.xyz,
					    i, i + 1, dr_OiOj);
			    if(aHOO < hb_aHOO){
				++donating_X_hb[iw];
				++total_hb[iw];
				++total_hb[jw];
			    }

			    aHOO = hb_angle(frame.imcon, frame.cell,
		     			    frame.rcell, frame.xyz,
				 	    i, i + 2, dr_OiOj);
			    if(aHOO < hb_aHOO){
				++donating_X_hb[iw];
				++total_hb[iw];
				++total_hb[jw];
			    }
			}

			//   H_j(a/b)
			//    /
			//   Oj --- Oi
			{
			    // Change direction of the r(OiOj) vector
			    double dr_OjOi[3];
			    for(size_t a = 0; a < 3; ++a)
				dr_OjOi[a] = -dr_OiOj[a];

			    double aHOO;
			    aHOO = hb_angle(frame.imcon, frame.cell,
		     			    frame.rcell, frame.xyz,
				 	    j, j + 1, dr_OjOi);
			    if(aHOO < hb_aHOO){
				++donating_X_hb[jw];
				++total_hb[jw];
				++total_hb[iw];
			    }

			    aHOO = hb_angle(frame.imcon, frame.cell,
		     			    frame.rcell, frame.xyz,
				 	    j, j + 2, dr_OjOi);
			    if(aHOO < hb_aHOO){
				++donating_X_hb[jw];
				++total_hb[jw];
				++total_hb[iw];
			    }
			}

		    } // if(rij < hb_rOO)
		} // for(size_t jw = 0 ...
	    } // for(size_t iw = 0 ...

#if 0
	    for(size_t iw = 0; iw < nw; ++iw){
		if(total_hb[iw] == 0){
		    const size_t i = 3*iw;
		    const double* qi = frame.xyz + 3*i;

		    std::vector<int> oxygens;
		    oxygens.push_back(iw);

		    for(size_t jw = 0; jw < nw; ++jw){
			if(iw != jw){
			    const size_t j = 3*jw;
			    const double* qj = frame.xyz + 3*j;

			    const double rij = tools::distance3(frame.imcon,
								frame.cell,
								frame.rcell,
								qi, qj);

			    if(rij < 3.5)
				oxygens.push_back(iw);
			}
		    }

		    // print
		    std::cout << oxygens.size()*3 << std::endl << std::endl;

		    for(size_t i = 0; i < oxygens.size(); ++i){
			const double* qi = frame.xyz + 9*oxygens[i];
			double mon[9];
			tools::image_monomer(frame.cell, qi, mon);

			for(size_t a = 0; a < 3; ++a){ // atom
			    if(a == 0)
				std::cout << "O  ";
			    else
				std::cout << "H  ";
			    for(size_t b = 0; b < 3; ++b){ // coord
				std::cout << mon[3*a + b] << "   ";
			    }
			    std::cout << std::endl;
			}
		    }
		}
	    }
#endif


	    // hb_type values:
	    //   0: other
	    //   1: 2n (  two HBs,    non donor)
	    //   2: 2s (  two HBs, single donor)
	    //   3: 2d (  two HBs, double donor)
	    //   4: 3n (three HBs,    non donor)
	    //   5: 3s (three HBs, single donor)
	    //   6: 3d (three HBs, double donor)
	    //   7: 4n ( four HBs,    non donor)
	    //   8: 4s ( four HBs, single donor)
	    //   9: 4d ( four HBs, double donor)
#if 0
	    int hb_type[nw];
	    std::fill(hb_type, hb_type + nw, 0);

	    for(size_t iw = 0; iw < nw; ++iw){

		if(total_hb[iw] == 0){
		    hb_type[iw] = 1;
		}else if(total_hb[iw] == 1){
		    if(donating_X_hb[iw] == 0)
		     	hb_type[iw] = 2;
		    else if(donating_X_hb[iw] == 1)
		     	hb_type[iw] = 3;
		}else if(total_hb[iw] == 2){
		    if(donating_X_hb[iw] == 0)
		     	hb_type[iw] = 4;
		    else if(donating_X_hb[iw] == 1)
		     	hb_type[iw] = 5;
		    else if(donating_X_hb[iw] == 2)
		     	hb_type[iw] = 6;
		}else if(total_hb[iw] == 3){
		    if(donating_X_hb[iw] == 0)
		     	hb_type[iw] = 7;
		    else if(donating_X_hb[iw] == 1)
		     	hb_type[iw] = 8;
		    else if(donating_X_hb[iw] == 2)
		     	hb_type[iw] = 9;
		}else if(total_hb[iw] == 4){
		    if(donating_X_hb[iw] == 0)
		     	hb_type[iw] = 10;
		    else if(donating_X_hb[iw] == 1)
		     	hb_type[iw] = 11;
		    else if(donating_X_hb[iw] == 2)
		     	hb_type[iw] = 12;
		}

//		std::cout << iw << ' ' << donating_X_hb[iw] << ' '
//		          << total_hb[iw] << ' ';
//		          << hb_type[iw] << std::endl;
	    }

//	    exit(1);
#endif

	         ////////////////////////////////////////////
	         // Calculate the intstantaneous interface //
	         ////////////////////////////////////////////

	    std::vector<instant::grid_point> surface_points;

	    // Calculate the coarse grained density on a grid. Locate gridpoints
	    // where the instantaneous interface lies and add those to the
	    // surface_points vector
	    for(int ix = 0; ix < ngrid_x; ++ix){
		const double x = (ix + 0.5)*dx;

		for(int iy = 0; iy < ngrid_y; ++iy){
		    const double y = (iy + 0.5)*dy;

		    double prev_value(0);
		    for(int iz = 0; iz < ngrid_z; ++iz){
			const double z = (iz + 0.5)*dz - half_z_length;

			if(std::fabs(z) < 15){
			    double grid_point[3] = {x, y, z};
			    double value = calc_cg_value(grid_point, frame.xyz,
							 frame.natm,
							 frame.imcon,
							 frame.cell,
							 frame.rcell);

    			    if( crossed_isosurface(value, prev_value) ){

    				instant::grid_point here;
				here.ix = ix;
				here.iy = iy;
				// linear interpolation to the isovalue
				double ratio = (isovalue - prev_value)
				                         / (value - prev_value);
				here.rz = ratio*dz + (z - dz);

    				surface_points.push_back(here);

			    }

			    prev_value = value;

    			    //avg += value;
			    //++num;
			}

		    } // for(int iz ...
		} // for(int iy ...
	    } // for(int ix ...

	    // calculate the gradient of the CG density at each surface point
	    for(size_t i = 0; i < surface_points.size(); ++i){
		const double x = (surface_points[i].ix + 0.5)*dx;
	    	const double y = (surface_points[i].iy + 0.5)*dy;
	    	const double z =  surface_points[i].rz;

		double dxyz[3] = {0.0, 0.0, 0.0};

		double grid_point[3] = {x, y, z};
		//const double v = calc_cg_value(grid_point, frame.xyz,
		//			       frame.natm, frame.imcon,
		//			       frame.cell, frame.rcell);

		// df/dq ~= [ f(q+h) - f(q-h) ] / 2h
		for(size_t j = 0; j < 3; ++j){
		    double q0 = grid_point[j];

		    grid_point[j] = q0 + eps;
		    const double vforw = calc_cg_value(grid_point, frame.xyz,
						       frame.natm, frame.imcon,
				      		       frame.cell, frame.rcell);

		    grid_point[j] = q0 - eps;
		    const double vback = calc_cg_value(grid_point, frame.xyz,
						       frame.natm, frame.imcon,
				      		       frame.cell, frame.rcell);

		    grid_point[j] = q0;
		    dxyz[j] = (vforw - vback)/(2*eps);
		}
		double norm = std::sqrt(dxyz[0]*dxyz[0] 
				      + dxyz[1]*dxyz[1] 
				      + dxyz[2]*dxyz[2]);
		for(size_t j = 0; j < 3; ++j){
		    dxyz[j] /= norm;
		    surface_points[i].drho[j] = dxyz[j];
		}

	    }

	    // Now loop over every oxygen and locate the closest surface point
	    for(size_t iw = 0; iw < nw; ++iw){ // O only
		const size_t i = 3*iw;
		const double* qi = frame.xyz + 3*i;

		double closest_distance(1000);
		double closest_drij[3];
		int closest_index(0);

		for(size_t j = 0; j < surface_points.size(); ++j){
		    const double x = (surface_points[j].ix + 0.5)*dx;
		    const double y = (surface_points[j].iy + 0.5)*dy;
		    const double z =  surface_points[j].rz;

		    double surface_point[3] = {x, y, z};
		    double drij[3];
		    const double rij = tools::distance3(frame.imcon,
	    						frame.cell,
							frame.rcell,
							qi, surface_point,
							drij);

		    if(rij < closest_distance){
			closest_distance = rij;
			closest_index = j;
			for(size_t k = 0; k < 3; ++k)
			    closest_drij[k] = drij[k]; // displacement vector
		    }
		}

		{ // Add the closest distance to the bin (use correct Expr.)
		    double proximity(0);
		    for(size_t k = 0; k < 3; ++k)
			proximity += (closest_drij[k])
			                * surface_points[closest_index].drho[k];
		    
		    ++bin_total[(int)((proximity-r_min)/dr)];

		    // Break the distance distributions into HB type
		    if(total_hb[iw] == 0){ // 0n
			++bin_0n[(int)((proximity-r_min)/dr)];
		    }else if(total_hb[iw] == 1){
			if(donating_X_hb[iw] == 0) // 1n
			    ++bin_1n[(int)((proximity-r_min)/dr)];
			else if(donating_X_hb[iw] == 1) // 1s
			    ++bin_1s[(int)((proximity-r_min)/dr)];
		    }else if(total_hb[iw] == 2){
			if(donating_X_hb[iw] == 0) // 2n
			    ++bin_2n[(int)((proximity-r_min)/dr)];
			else if(donating_X_hb[iw] == 1) // 2s
			    ++bin_2s[(int)((proximity-r_min)/dr)];
			else if(donating_X_hb[iw] == 2) // 2d
			    ++bin_2d[(int)((proximity-r_min)/dr)];
		    }else if(total_hb[iw] == 3){
			if(donating_X_hb[iw] == 0) // 3n
			    ++bin_3n[(int)((proximity-r_min)/dr)];
			else if(donating_X_hb[iw] == 1) // 3s
			    ++bin_3s[(int)((proximity-r_min)/dr)];
			else if(donating_X_hb[iw] == 2) // 3d
			    ++bin_3d[(int)((proximity-r_min)/dr)];
		    }else if(total_hb[iw] == 4){
			if(donating_X_hb[iw] == 0) // 4n
			    ++bin_4n[(int)((proximity-r_min)/dr)];
			else if(donating_X_hb[iw] == 1) // 4s
			    ++bin_4s[(int)((proximity-r_min)/dr)];
			else if(donating_X_hb[iw] == 2) // 4d
			    ++bin_4d[(int)((proximity-r_min)/dr)];
		    }else{
    			++bin_other[(int)((proximity-r_min)/dr)];
		    }
	    	}
	    }

	    } // if(frameno % n_skip == 0){

	} // if(frame.natm > 0){
//	if(frameno > 5)
//	    exit(0);
    } // while(!ifs.eof()){
    return frameno;

}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 3) {
        std::cerr << "usage: dividing_surfaces n_skip POSITION_CMD POSITION_CMD ... > profile.dat" 
	          << std::endl;
        return EXIT_FAILURE;
    }

    // parse command line: n_skip, isovalue, position_files...
    { // read n_skip
	std::istringstream iss(argv[1]);
	size_t tmp;
	iss >> tmp;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[1]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
	if(tmp > 0){ // cannot be < 1
	    n_skip = tmp;
	}
    }

    { // read isovalue
	std::istringstream iss(argv[2]);
	double tmp;
	iss >> tmp;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[2]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
	if(tmp > 0){ // cannot be < 1
	    isovalue = tmp;
	}
    }

    std::fill(bin_total, bin_total + nbins, 0);
    std::fill(bin_other, bin_other + nbins, 0);

    std::fill(   bin_0n,    bin_0n + nbins, 0);

    std::fill(   bin_1n,    bin_1n + nbins, 0);
    std::fill(   bin_1s,    bin_1s + nbins, 0);

    std::fill(   bin_2n,    bin_2n + nbins, 0);
    std::fill(   bin_2s,    bin_2s + nbins, 0);
    std::fill(   bin_2d,    bin_2d + nbins, 0);

    std::fill(   bin_3n,    bin_3n + nbins, 0);
    std::fill(   bin_3s,    bin_3s + nbins, 0);
    std::fill(   bin_3d,    bin_3d + nbins, 0);

    std::fill(   bin_4n,    bin_4n + nbins, 0);
    std::fill(   bin_4s,    bin_4s + nbins, 0);
    std::fill(   bin_4d,    bin_4d + nbins, 0);

    size_t tot_frameno(0);

    for(int i = 3; i < argc; ++i){

	std::ifstream ifs(argv[i]);
    	std::string line;
    	size_t lineno(0);
#ifdef HIST
	std::getline(ifs, line);
    	std::getline(ifs, line);
    	lineno += 2;
#endif

	const size_t frameno = do_frames(lineno, ifs);
	tot_frameno += frameno;

	std::cerr << argv[i] << " : " << frameno << std::endl;

    }

    dump_1d_distrib(tot_frameno, "dist_total.dat", bin_total, nbins, dr, r_min);
    dump_1d_distrib(tot_frameno, "dist_other.dat", bin_other, nbins, dr, r_min);

    dump_1d_distrib(tot_frameno,    "dist_0n.dat",    bin_0n, nbins, dr, r_min);

    dump_1d_distrib(tot_frameno,    "dist_1n.dat",    bin_1n, nbins, dr, r_min);
    dump_1d_distrib(tot_frameno,    "dist_1s.dat",    bin_1s, nbins, dr, r_min);

    dump_1d_distrib(tot_frameno,    "dist_2n.dat",    bin_2n, nbins, dr, r_min);
    dump_1d_distrib(tot_frameno,    "dist_2s.dat",    bin_2s, nbins, dr, r_min);
    dump_1d_distrib(tot_frameno,    "dist_2d.dat",    bin_2d, nbins, dr, r_min);

    dump_1d_distrib(tot_frameno,    "dist_3n.dat",    bin_3n, nbins, dr, r_min);
    dump_1d_distrib(tot_frameno,    "dist_3s.dat",    bin_3s, nbins, dr, r_min);
    dump_1d_distrib(tot_frameno,    "dist_3d.dat",    bin_3d, nbins, dr, r_min);

    dump_1d_distrib(tot_frameno,    "dist_4n.dat",    bin_4n, nbins, dr, r_min);
    dump_1d_distrib(tot_frameno,    "dist_4s.dat",    bin_4s, nbins, dr, r_min);
    dump_1d_distrib(tot_frameno,    "dist_4d.dat",    bin_4d, nbins, dr, r_min);

}
