#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <cmath>
#include <algorithm>

#include <limits.h>

#include "read-hist.h"
#include "read-bin.h"
#include "pbc-tools.h"

//#define SHIFT_HALF_CELL yes

#define HIST yes

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

const double cg_length = 2.4;
const double cg_cutoff = cg_length*3;

const double dx = 1.00;
const double dy = 1.00;
const double dz = 1.00;

double z_length = 0;
double half_z_length = 0;

double avg(0);
int num(0);
int print_frame_num(0);

int n_skip(1);
int max_frames(INT_MAX);

int nbins_x = 1000;
int nbins_y = 1000;
int nbins_z = 1000;

//int natm = 0;

bool first(true);

//----------------------------------------------------------------------------//

double cg_density(const double r, const double xi)
{
    const double xi2 = xi*xi;

    return std::pow(2*M_PI*xi2, -3/2)*std::exp(-r*r/(2*xi2));
}

//----------------------------------------------------------------------------//

size_t do_frames(size_t lineno, std::ifstream& ifs)
{

    size_t frameno(0);

    while(!ifs.eof()){

	opt::hist_frame frame;

#ifdef HIST     
	opt::read_hist_frame(lineno, ifs, frame);
#else
	opt::read_cmd_frame(lineno, ifs, frame);
#endif
	std::cerr << frame.natm << ' ';

	if(frame.natm > 0){
	    ++frameno;

#ifdef SHIFT_HALF_CELL
	    tools::center_slab_at_origin(frame.natm, frame.cell, frame.mass,
					 frame.xyz, true);
#else
	    tools::center_slab_at_origin(frame.natm, frame.cell, frame.mass,
					 frame.xyz, false);
#endif

	    if(frameno % n_skip == 0){

	    // set up grid: dim_1 = cell_x / dx
	    //              dim_2 = cell_y / dy
	    //              dim_3 = cell_z / dz

    	    if(first){
		first = false;

    		z_length = frame.cell[8]; 
		half_z_length = z_length/2.0;
//
//		natm = frame.natm;

    		nbins_x = (int)(frame.cell[0]/dx);
    		nbins_y = (int)(frame.cell[4]/dy);
    		nbins_z = (int)(frame.cell[8]/dz);
	    }

	    //Prepare file name: "frame_0000XXXXXX.cube"

	    std::string base = "frame_";
	    std::string tail = ".cube";

	    std::ostringstream oss;
	    oss << std::setw(10) << std::setfill('0') << print_frame_num;
	    std::string file_num;
	    file_num = oss.str();

	    std::string filename = base + file_num + tail;

	    // Open file
	    std::ofstream cube_file;
	    cube_file.open(filename.c_str());
	    cube_file << std::setprecision(10) << std::scientific;

	    // two description lines
	    cube_file << std::endl << "Timestep: " << frame.time << std::endl;

	    const double ang2au = 1.0/0.529177;

	    cube_file << std::fixed << std::setprecision(6);

	    // #atoms, origin_x, origin_y, origin_z
	    cube_file << std::setw(5) << frame.natm
		      << std::setw(12) << (0.5*dx - frame.cell[0]/2)*ang2au
		      << std::setw(12) << (0.5*dy - frame.cell[4]/2)*ang2au
		      << std::setw(12) << (0.5*dz - half_z_length)*ang2au
		      << std::endl;

	    // # of grid points and x/y/z increment vectors
	    cube_file << std::setw(5) << nbins_x;
	    for(size_t i = 0; i < 3; ++i)
		cube_file << std::setw(12) << frame.cell[i    ]/nbins_x*ang2au;
	    cube_file << std::endl;

	    cube_file << std::setw(5) << nbins_y;
	    for(size_t i = 0; i < 3; ++i)
		cube_file << std::setw(12) << frame.cell[i + 3]/nbins_y*ang2au;
	    cube_file << std::endl;

	    cube_file << std::setw(5) << nbins_z;
	    for(size_t i = 0; i < 3; ++i)
		cube_file << std::setw(12) << frame.cell[i + 6]/nbins_z*ang2au;
	    cube_file << std::endl;

	    for(size_t i = 0; i < frame.natm; ++i){
            if (frame.atomic_num[i] != 8) continue;
		    tools::image_monomer(frame.cell, frame.xyz + 3*i);
        }

	    for(size_t i = 0; i < frame.natm; ++i){
            cube_file << std::setw(5) << frame.atomic_num[i];

		cube_file << std::setw(12) << 0
	                  << std::setw(12) << frame.xyz[3*i + 0] * ang2au
	                  << std::setw(12) << frame.xyz[3*i + 1] * ang2au
	                  << std::setw(12) << frame.xyz[3*i + 2] * ang2au
			  << std::endl;
	    }

	    cube_file << std::scientific;
	    int count_columns(0);

	    for(int ix = 0; ix < nbins_x; ++ix){
		const double x = (ix + 0.5)*dx;

		for(int iy = 0; iy < nbins_y; ++iy){
		    const double y = (iy + 0.5)*dy;

		    for(int iz = 0; iz < nbins_z; ++iz){
			const double z = (iz + 0.5)*dz - half_z_length;

		      //if(std::fabs(z) < 7.5){
			double grid_point[3] = {x, y, z};
			double value(0);
			for(size_t j = 0; j < frame.natm; ++j){ 
                if (frame.atomic_num[j] < 4) continue; // heavy atoms only
			    const double* qj = frame.xyz + 3*j;
			    const double rij = tools::distance3(frame.imcon,
		    						frame.cell,
								frame.rcell,
								grid_point, qj);
			    if(rij < cg_cutoff)
				value += cg_density(rij, cg_length);
			}
			avg += value;
			++num;

			++count_columns;
			cube_file << std::setw(13) << value;
			if(count_columns % 6 == 0)
			    cube_file << std::endl;
//		      }
		    }
		}
	    }

	    ++print_frame_num;
	    cube_file.close();
	    if(print_frame_num > max_frames - 1)
		exit(0);
	    }

	}
    }
    return frameno;

}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 2) {
        std::cerr << "usage: dividing_surfaces n_skip max_frames POSITION_CMD POSITION_CMD ..." 
	          << std::endl;
        return EXIT_FAILURE;
    }

    // parse command line: n_skip, max_frames, position_files...
    { // read n_skip
	std::istringstream iss(argv[1]);
	size_t tmp;
	iss >> tmp;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[1]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
	if(tmp > 0){ // cannot be < 1
	    n_skip = tmp;
	}
    }

    { // read max_frames
	std::istringstream iss(argv[2]);
	size_t tmp;
	iss >> tmp;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[2]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
	if(tmp > 0){ // cannot be < 1. if < 1, uses default (INT_MAX)
	    max_frames = tmp;
	}
    }
    std::cerr << n_skip << " " << max_frames << std::endl;

    size_t tot_frameno(0);

    std::cout << std::setprecision(10) << std::scientific;

    for(int i = 3; i < argc; ++i){

	std::ifstream ifs(argv[i]);
    	std::string line;
    	size_t lineno(0);
#ifdef HIST
	std::getline(ifs, line);
    	std::getline(ifs, line);
    	lineno += 2;
#endif

	const size_t frameno = do_frames(lineno, ifs);
	tot_frameno += frameno;

	std::cerr << argv[i] << " : " << frameno << std::endl;

	std::cerr << "average density: " << avg/num << std::endl
	          << "averaged over " << tot_frameno << " frames" << std::endl;

    }

    std::cout << "average density: " << avg/num << std::endl
	      << "averaged over " << tot_frameno << " frames" << std::endl;


}

