#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <cmath>
#include <algorithm>

#include "read-hist.h"
#include "read-bin.h"
#include "struct-raman-tcc.h"

#define DO_TCC yes
#define no_BINARY yes

#include "avila.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static x2o::avila pol_model;

std::vector<raman::eprop> eprops;
std::vector<raman::connect> connects;

const double zc1(1.0);
const double zc2(2.0);
const double rOO_max(4);

int int_time_ind(0);
double dt(0);

//----------------------------------------------------------------------------//

double tr_mat_prod(double* a0, double* at)
{

    double xx = a0[0]*at[0] + a0[1]*at[3] + a0[2]*at[6];
    double yy = a0[3]*at[1] + a0[4]*at[4] + a0[5]*at[7];
    double zz = a0[6]*at[2] + a0[7]*at[5] + a0[8]*at[8];

    return (xx + yy + zz);

}

//----------------------------------------------------------------------------//

inline double distance(size_t imcon, const double* cell, 
		       const double* r1, const double* r2)
{
    assert(imcon == 0 || imcon == 1);

    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    dx -= cell[0]*std::floor(dx/cell[0] + 0.5);
    dy -= cell[4]*std::floor(dy/cell[4] + 0.5);
    dz -= cell[8]*std::floor(dz/cell[8] + 0.5);

    return std::sqrt(dx*dx + dy*dy + dz*dz);
}

//----------------------------------------------------------------------------//

inline void image(const double* cell, const double* r1, const double* r2, 
		  double* crd_out)
{
    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    double facx = std::floor(dx/cell[0] + 0.5);
    double facy = std::floor(dy/cell[4] + 0.5);
    double facz = std::floor(dz/cell[8] + 0.5);

    crd_out[0] = r2[0] + cell[0]*facx;
    crd_out[1] = r2[1] + cell[4]*facy;
    crd_out[2] = r2[2] + cell[8]*facz;
}

//----------------------------------------------------------------------------//

inline void image_monomer(const double* cell, 
			  const double* r1, double* mon)
{
    std::fill(mon, mon+9, 0.0);

    // take r1 oxygen as central atom, image its two hydrogens
     
    for(size_t i = 0; i < 3; ++i)
	mon[i] = r1[i];
    
    for(size_t a = 1; a < 3; ++a)
	image(cell, r1, r1 + 3*a, mon + 3*a);
}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 3) {
        std::cerr << "usage: response_raman_tcc_1b int_max filename" 
	          << std::endl;
        return EXIT_FAILURE;
    }

    double int_max(0);
    {
	std::istringstream iss(argv[1]);
	iss >> int_max;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[1]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }

    { // attempt to control scope of frames
    std::vector<opt::hist_frame> frames;
    for(int i = 2; i < argc; ++i){
	try {

#ifdef BINARY
	    opt::read_bin(argv[i], frames);
#else
	    opt::read_cmd(argv[i], frames, false);
#endif

	    std::cerr <<" # Frames : " << frames.size() << std::endl;

	} catch (const std::exception& e) {
	    std::cerr << e.what() << std::endl;
	    return EXIT_FAILURE;
	}
    }

    // First, go over trajectory and pre-compute \mu_z and \alpha_{xx}

    dt = (frames[1].time - frames[0].time);
    int_time_ind = (int) (int_max / dt);

    std::cerr << dt << "   " << int_time_ind << std::endl;

    for(size_t n = 0; n < frames.size(); ++n){

	const size_t nmolec = frames[n].natm/3;

        { // calculate electrostatic properties
	    raman::eprop eprop_this_frame;
	    eprop_this_frame.nmolec    = nmolec;
	    eprop_this_frame.pol_iso   = new double[  eprop_this_frame.nmolec];
	    eprop_this_frame.pol_aniso = new double[9*eprop_this_frame.nmolec];

	    for(size_t i = 0; i < nmolec; ++i){
		const size_t i9 = 9*i;

		double mon[9];
		image_monomer(frames[n].cell, frames[n].xyz + i9, mon);


		// xx-component of the polarizability
		double pol_mol[9];
		pol_model.alpha(mon, pol_mol);

		double a_iso = (pol_mol[0] + pol_mol[4] + pol_mol[8])/3.0;
		eprop_this_frame.pol_iso[i] = a_iso;

		// set anisotropic polarizability to total polarizability
		for(size_t j = 0; j < 9; ++j)
		    eprop_this_frame.pol_aniso[i9 + j] = pol_mol[j];

		// remove isotropic part from diagonal elements
		eprop_this_frame.pol_aniso[i9 + 0] -= a_iso;
		eprop_this_frame.pol_aniso[i9 + 4] -= a_iso;
		eprop_this_frame.pol_aniso[i9 + 8] -= a_iso;

	    }

	    eprops.push_back(eprop_this_frame);
	}

	{ // determine connectivity

	    raman::connect connect_this_frame;
	    connect_this_frame.nmolec = nmolec;

	    for(size_t i = 0; i < nmolec; ++i){

#ifdef DO_TCC //Begin partner determination
		const size_t i9 = 9*i;

		std::vector<int> mol;

		for(size_t j = 0; j < nmolec; ++j){
	    	    double avg_rij(0);
	    	    size_t count(0);
	    	    if(j != i){

			// Assume that int_time_ind is short enough to 
		    	// prohibit diffusion between solvation shells
		    	// and that the average_rij cannot be < rOO_max
		    	// unless rij at the initial time is 
		    	// less than 1.5*rOO_max
		    	if(distance(frames[n].imcon, frames[n].cell,
		    		    frames[n].xyz + i9, frames[n].xyz + j*9)
			        < 2.0*rOO_max){

			    // calculate average rij distance over the next
		    	    // int_time_ind frames
		    	    for(size_t a = n;
		    		       a < std::min( frames.size(),
			    	       	             n + int_time_ind + 1);
				       a += 250){

				avg_rij += distance(frames[n].imcon, 
			    			    frames[n].cell,
					    	    frames[n].xyz + i9, 
					    	    frames[a].xyz + j*9);
				++count;

			    }
		    	    if((avg_rij/count) < rOO_max){
				mol.push_back((int)j);
			    }
			}
		    }
		}

		connect_this_frame.partners.push_back(mol);

#endif // end partner determination

		connect_this_frame.central.push_back(i);

	    }

	    connects.push_back(connect_this_frame);
	}

    }

    if(eprops.size() != frames.size()){
	std::cerr << "eprops.size() != frames.size()" << std::endl;
	exit(1);
    }

    std::cerr << "Deallocating frames" << std::endl;
//    frames.erase(frames.begin(), frames.end());
    std::vector<opt::hist_frame>().swap(frames);
    }

    // With electrostatic properties loaded in eprops
    // Calculated the 1B approximation to the SFG with the 
    // truncated cross-correlation approximation of dx.doi.org/10.1021/jz400683v
    
    double corr_iso[int_time_ind];
    std::fill(corr_iso, corr_iso + int_time_ind, 0.0);

    double corr_aniso[int_time_ind];
    std::fill(corr_aniso, corr_aniso + int_time_ind, 0.0);

    int ntime[int_time_ind];
    std::fill(ntime, ntime + int_time_ind, 0.0);

    std::cerr << "Beginning correlation function" << std::endl;

    for(size_t m = 0; m < eprops.size(); ++m){ // time = 0
	for(size_t n = m; // time = 't'
	        n < std::min( eprops.size(), m + int_time_ind); ++ n){ 

	    int time = n - m;

	    double* pol_iso_0   = eprops[m].pol_iso;
	    double* pol_aniso_0 = eprops[m].pol_aniso;

	    double* pol_iso_t   = eprops[n].pol_iso;
	    double* pol_aniso_t = eprops[n].pol_aniso;

	    for(size_t i = 0; i < connects[m].central.size(); ++i){
		const size_t i_ind = connects[m].central[i];

		// autocorrelation, j == i
		{

	    	    corr_iso[time]   += pol_iso_0[i_ind]*pol_iso_t[i_ind];

	    	    corr_aniso[time] += tr_mat_prod(pol_aniso_0 + 9*i_ind,
						    pol_aniso_t + 9*i_ind);

		}

#ifdef DO_TCC
		// truncated cross-correlation
		for(size_t j = 0; j < connects[m].partners[i].size(); ++j){
	    	    const size_t j_ind = connects[m].partners[i][j];

		    corr_iso[time]   += pol_iso_0[i_ind]*pol_iso_t[j_ind];

	    	    corr_aniso[time] += tr_mat_prod(pol_aniso_0 + 9*i_ind,
						    pol_aniso_t + 9*j_ind);
		}
#endif
	    }
	    ++ntime[time];

	}
    }

    // print correlation function for isotropic spectrum

    std::ofstream ofs_iso;
    ofs_iso.open("response_raman_iso.dat");

    ofs_iso << std::setprecision(10) << std::scientific;

    for(int i = 0; i < int_time_ind; ++i)
	ofs_iso << std::setw(20) << dt*i 
	        << std::setw(20) << corr_iso[i]/((double)ntime[i])
	        << std::setw(20) << corr_iso[i]
	        << std::setw(20) << ((double)ntime[i])
		<< std::endl;

    ofs_iso.close();

    // print correlation function for anisotropic spectrum

    std::ofstream ofs_aniso;
    ofs_aniso.open("response_raman_aniso.dat");

    ofs_aniso << std::setprecision(10) << std::scientific;

    for(int i = 0; i < int_time_ind; ++i)
	ofs_aniso << std::setw(20) << dt*i 
	        << std::setw(20) << corr_aniso[i]/((double)ntime[i])
	        << std::setw(20) << corr_aniso[i]
	        << std::setw(20) << ((double)ntime[i])
		<< std::endl;

    ofs_aniso.close();

}

