#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <cmath>
#include <algorithm>

#include "read-hist.h"
#include "struct-sfg.h"

#include "ttm4-dip.h"

#include "alpha-ttm4.h"

#include "ps.h"
#include "ltp2011.h"
#include "avila.h"

#define LTP yes

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static x2o::avila pol_model;
static h2o::ltp2011 dip_model;
typedef ttm::ttm4_dip ind_dip;

const double zc1(2.0);
const double zc2(3.0);
const double rOO_max(3.7);

const double r2i = 3.3;
const double r2f = 3.7;

//----------------------------------------------------------------------------//


double f_switch2(const double& r)
{
    if (r > r2f)
	return 0.0;
    else if (r > r2i) {
	const double x = (r - r2i)/(r2f - r2i);
	return 1 + x*x*(2*x - 3);
    } else
	return 1.0;
}

//----------------------------------------------------------------------------//

inline double distance(size_t imcon, const double* cell, 
		       const double* r1, const double* r2)
{
    assert(imcon == 0 || imcon == 1);

    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    dx -= cell[0]*std::floor(dx/cell[0] + 0.5);
    dy -= cell[4]*std::floor(dy/cell[4] + 0.5);
    dz -= cell[8]*std::floor(dz/cell[8] + 0.5);

    return std::sqrt(dx*dx + dy*dy + dz*dz);
}

//----------------------------------------------------------------------------//

inline void image(const double* cell, const double* r1, const double* r2, 
		  double* crd_out)
{
    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    double facx = std::floor(dx/cell[0] + 0.5);
    double facy = std::floor(dy/cell[4] + 0.5);
    double facz = std::floor(dz/cell[8] + 0.5);

    crd_out[0] = r2[0] + cell[0]*facx;
    crd_out[1] = r2[1] + cell[4]*facy;
    crd_out[2] = r2[2] + cell[8]*facz;
}

//----------------------------------------------------------------------------//

inline void image_monomer(const double* cell, 
			  const double* r1, double* mon)
{
    std::fill(mon, mon+9, 0.0);

    // take r1 oxygen as central atom, image its two hydrogens
     
    for(size_t i = 0; i < 3; ++i)
	mon[i] = r1[i];
    
    for(size_t a = 1; a < 3; ++a)
	image(cell, r1, r1 + 3*a, mon + 3*a);
}

//----------------------------------------------------------------------------//

inline void image_dimer(const double* cell,
			const double* r1, const double* r2,
			double* dimer)
{
    std::fill(dimer, dimer+18, 0.0);

    // take r1 oxygen as central atom, image its two hydrogens

    for(size_t i = 0; i < 3; ++i)
	dimer[i] = r1[i];

    for(size_t a = 1; a < 3; ++a)
	image(cell, r1, r1 + 3*a, dimer + 3*a);

    // Now image the oxygen of the second molecule
    
    image(cell, r1, r2, dimer + 9);

    // Finally, reattach hydrogens of molec two to oxygen of molec 2
    for(size_t a = 1; a < 3; ++a)
	image(cell, dimer+9, r2 + 3*a, dimer + 9 + 3*a);
}

//----------------------------------------------------------------------------//

inline double g_scale(const double* crd)
{

    double z = std::fabs(crd[2]);
    if(z < zc1)
	return 0.0;
    else if(z > zc2)
	return copysign(1.0, crd[2]);
    else{
	const double c = std::cos(M_PI*(z-zc2)/(2*(zc1-zc2)));
	return copysign(c*c, crd[2]); // multiply switch by sign of z (crd[2])
    }

}

//----------------------------------------------------------------------------//

inline double g2_scale(const double* crd1, const double* crd2)
{

    double scale1 = g_scale(crd1);
    double scale2 = g_scale(crd2);

    return copysign(scale1*scale2, scale1);

}

//----------------------------------------------------------------------------//

inline double g2avg_scale(const double* crd1, const double* crd2)
{

    double avg[3];
    for(size_t i = 0; i < 3; ++i)
	avg[i] = (crd1[i] + crd2[i])/2.0;

    return g_scale(avg);

}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 1) {
        std::cerr << "usage: dipol_sfg_fcf_1b > dipol.dat" 
	          << std::endl;
        return EXIT_FAILURE;
    }

    // Set up i/o

    std::ifstream ifs_cmd("POSITION_CMD");

    std::ofstream dipol_12;
    dipol_12.open("dipol_12.dat");
    dipol_12 << std::scientific << std::setprecision(15);

#ifdef EFF
    std::ofstream dipol_tcc_12;
    dipol_tcc_12.open("dipol_tcc_12.dat");
    dipol_tcc_12 << std::scientific << std::setprecision(15);
#endif

    std::cout << std::scientific << std::setprecision(9);

    ind_dip my_ind_dip;

                       ///////////////////////////
                       //   Begin calculation   //
                       ///////////////////////////

    size_t lineno(0);
    while(!ifs_cmd.eof()){

	opt::hist_frame frame;
	opt::read_cmd_frame(lineno, ifs_cmd, frame);
	if(frame.natm == 0)
	    break;

	const size_t nmolec = frame.natm/3;

	{ // set the center of mass of the cell as the origin

	    double com[3] = {0.0, 0.0, 0.0};
	    double tot_mass(0);
	    for(size_t i = 0; i < frame.natm; ++i){
		for(size_t j = 0; j < 3; ++j)
		    com[j] += frame.mass[i]*frame.xyz[i*3 + j];

		tot_mass +=frame.mass[i];
	    }

	    for(size_t j = 0; j < 3; ++j)
		com[j] /= tot_mass;

	    for(size_t i = 0; i < frame.natm; ++i)
		for(size_t j = 0; j < 3; ++j)
		    frame.xyz[i*3 + j] -= com[j];

	}

        // calculate electrostatic properties
	
        double dip_1b(0), pol_1bx(0), pol_1by(0);
	double dip_2b(0), pol_2bx(0), pol_2by(0);

#ifdef EFF
	double dip_eff_2b[nmolec];
	std::fill(dip_eff_2b, dip_eff_2b + nmolec, 0.0);

	double dip_eff_Nb[nmolec];
	std::fill(dip_eff_Nb, dip_eff_Nb + nmolec, 0.0);

	double pol_eff_2b[nmolec];
	std::fill(pol_eff_2b, pol_eff_2b + nmolec, 0.0);
#endif

	for(size_t i = 0; i < nmolec; ++i){
    	    const size_t i9 = 9*i;

    	    double mon[9];
	    image_monomer(frame.cell, frame.xyz + i9, mon);

    	    // scaling that accounts for the double-interface
	    double g_i = g_scale(frame.xyz + i9);

            /////
	    // 1B Dipole/polarizability
	    /////

	    { // z-component of dipole

	        double dip_mon[3];
#ifdef LTP
		dip_model.cartesian(1, mon, dip_mon);
#else
	        ps::dipole(mon, dip_mon);
#endif

	        dip_1b += g_i*dip_mon[2];
#ifdef EFF
		dip_eff_2b[i] = dip_mon[2];
#endif

	    }

	    { // xx-component of the polarizability

	        double pol_mol[9];
	        pol_model.alpha(mon, pol_mol);

		pol_1bx += pol_mol[0];
		pol_1by += pol_mol[4];
#ifdef EFF
		pol_eff_2b[i] = pol_mol[0];
#endif

	    }

            /////
	    // 2B Dipole/polarizability
	    /////

	    for(size_t j = i + 1; j < nmolec; ++j){
		const size_t j9 = 9*j;

		double rOO = distance(frame.imcon, frame.cell,
				      frame.xyz + i9, frame.xyz + j9);

		if(rOO < r2f){

		    const double s2b = f_switch2(rOO);
		    //const double g_j = g_scale(frame.xyz + j9);
		    const double g_ij = g2_scale(frame.xyz + i9, 
						 frame.xyz + j9);
//		    const double g_ij = g2avg_scale(frame.xyz + i9, 
//					            frame.xyz + j9);

		    double dimer[18];
		    image_dimer(frame.cell,
				frame.xyz + i9, frame.xyz + j9, dimer);

		    { // Dipole 2-body

		        double ind_dip2b[9];
			my_ind_dip(2, dimer,  ind_dip2b);

			dip_2b +=  g_ij*s2b*ind_dip2b[2];

#ifdef EFF
			dip_eff_2b[i] += (s2b*(ind_dip2b[2]));
			dip_eff_2b[j] += (s2b*(ind_dip2b[2]));
#endif

		    }

		    { // Polarizability 2-body

		        double  ind_pol2b[9];
		        ttm::alpha_no_1b(2, dimer,  ind_pol2b);

			pol_2bx +=  s2b*ind_pol2b[0];
			pol_2by +=  s2b*ind_pol2b[4];

#ifdef EFF
			pol_eff_2b[i] += ( s2b*(ind_pol2b[0]));
			pol_eff_2b[j] += ( s2b*(ind_pol2b[0]));
#endif

		    }

		}


	    }

	}

	dipol_12  << frame.time << ' '
	          << std::setw(25) << dip_1b + dip_2b  << ' ' 
	          << std::setw(25) << pol_1bx + pol_2bx   << ' ' 
	          << std::setw(25) << pol_1by + pol_2by   << ' ' 
		  << std::endl;

#ifdef EFF
	dipol_tcc_12  << nmolec << std::endl;
	for(size_t i = 0; i < nmolec; ++i){

	    dipol_tcc_12  << std::setw(5)  << i
	                  << std::setw(25) << dip_eff_2b[i]
			  << std::setw(25) << pol_eff_2b[i] << std::endl;

	}
#endif

    }

}

