#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <cmath>
#include <algorithm>

#include "read-hist.h"
#include "struct-sfg.h"

#include "ps.h"
#include "avila.h"

#define ADD_NB_DIP yes

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static x2o::avila pol_model;

const double zc1(2.0);
const double zc2(3.0);
const double rOO_max(4);

//----------------------------------------------------------------------------//

inline double distance(size_t imcon, const double* cell, 
		       const double* r1, const double* r2)
{
    assert(imcon == 0 || imcon == 1);

    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    dx -= cell[0]*std::floor(dx/cell[0] + 0.5);
    dy -= cell[4]*std::floor(dy/cell[4] + 0.5);
    dz -= cell[8]*std::floor(dz/cell[8] + 0.5);

    return std::sqrt(dx*dx + dy*dy + dz*dz);
}

//----------------------------------------------------------------------------//

inline void image(const double* cell, const double* r1, const double* r2, 
		  double* crd_out)
{
    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    double facx = std::floor(dx/cell[0] + 0.5);
    double facy = std::floor(dy/cell[4] + 0.5);
    double facz = std::floor(dz/cell[8] + 0.5);

    crd_out[0] = r2[0] + cell[0]*facx;
    crd_out[1] = r2[1] + cell[4]*facy;
    crd_out[2] = r2[2] + cell[8]*facz;
}

//----------------------------------------------------------------------------//

inline void image_monomer(const double* cell, 
			  const double* r1, double* mon)
{
    std::fill(mon, mon+9, 0.0);

    // take r1 oxygen as central atom, image its two hydrogens
     
    for(size_t i = 0; i < 3; ++i)
	mon[i] = r1[i];
    
    for(size_t a = 1; a < 3; ++a)
	image(cell, r1, r1 + 3*a, mon + 3*a);
}

//----------------------------------------------------------------------------//

inline double g_scale(const double* crd)
{

    double z = std::fabs(crd[2]);
    if(z < zc1)
	return 0.0;
    else if(z > zc2)
	return copysign(1.0, crd[2]);
    else{
	const double c = std::cos(M_PI*(z-zc2)/(2*(zc1-zc2)));
	return copysign(c*c, crd[2]); // multiply switch by sign of z (crd[2])
    }

}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 1) {
        std::cerr << "usage: elect_sfg_fcf_1b > dipol.dat" 
	          << std::endl;
        return EXIT_FAILURE;
    }

    std::ifstream ifs_cmd("POSITION_CMD");
#ifdef ADD_NB_DIP
    std::ifstream ifs_mind("DIPIND_CMD");
#endif

    std::cout << std::scientific << std::setprecision(9);

    size_t lineno(0);
    while(!ifs_cmd.eof()){

	opt::hist_frame frame;
	opt::read_cmd_frame(lineno, ifs_cmd, frame);
	if(frame.natm == 0)
	    break;

#ifdef ADD_NB_DIP
	opt::collect_ind_multipoles(ifs_mind, frame);
#endif

	const size_t nmolec = frame.natm/3;

	double total_dip[3];
	std::fill(total_dip, total_dip + 3, 0.0);

	double total_pol[9];
	std::fill(total_pol, total_pol + 9, 0.0);

	{ // set the center of mass of the cell as the origin

	    double com[3] = {0.0, 0.0, 0.0};
	    double tot_mass(0);
	    for(size_t i = 0; i < frame.natm; ++i){
		for(size_t j = 0; j < 3; ++j)
		    com[j] += frame.mass[i]*frame.xyz[i*3 + j];

		tot_mass +=frame.mass[i];
	    }

	    for(size_t j = 0; j < 3; ++j)
		com[j] /= tot_mass;

	    for(size_t i = 0; i < frame.natm; ++i)
		for(size_t j = 0; j < 3; ++j)
		    frame.xyz[i*3 + j] -= com[j];

	}

        // calculate electrostatic properties

	for(size_t i = 0; i < nmolec; ++i){
    	    const size_t i9 = 9*i;

    	    double mon[9];
	    image_monomer(frame.cell, frame.xyz + i9, mon);

    	    // scaling that accounts for the double-interface
	    double g_i = g_scale(frame.xyz + i9);

    	    // z-component of the dipole
	    double dip_mon[3];
	    ps::dipole(mon, dip_mon);

    	    total_dip[0] +=     dip_mon[0];
	    total_dip[1] +=     dip_mon[1];
#ifdef ADD_NB_DIP
	    total_dip[2] += g_i*(dip_mon[2]+frame.dipind[3*i+2]);
#else
	    total_dip[2] += g_i*dip_mon[2];
#endif

    	    // xx-component of the polarizability
	    double pol_mol[9];
	    pol_model.alpha(mon, pol_mol);

    	    for(size_t j = 0; j < 9; ++j)
		total_pol[j] += pol_mol[j];

	}

	std::cout << frame.time << ' '
	          << std::setw(25) << total_dip[2]   << ' ' //  z component
	          << std::setw(25) << total_pol[0]   << ' ' // xx component
	          << std::setw(25) << total_pol[4]   << ' ' // yy component
		  << std::endl;

    }

}

