#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <omp.h>

#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <cmath>
#include <algorithm>

#include "read-hist.h"
#include "read-bin.h"
#include "pbc-tools.h"
#include "struct-layered-sfg.h"

#define DO_TCC yes
//#define SHIFT_HALF_CELL yes
#define CENTER_CELL yes
//#define BINARY yes

//#define TYPE2 yes
//#define JUST_SWITCH yes

#include "ps.h"
#include "avila.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static std::vector<opt::hist_frame> frames;
static x2o::avila pol_model;

std::vector<sfg::eprop> eprops;
std::vector<sfg::connect> connects;

const double rOO_max(4);

//----------------------------------------------------------------------------//

// Lower cutoff function
inline double g_scale_a(const double* crd, const double za1, const double za2)
{

    double z = std::fabs(crd[2]);
    if(z < za1)
	return 0.0;
    else if(z > za2)
	return copysign(1.0, crd[2]);
    else{
	const double c = std::cos(M_PI*(z-za2)/(2*(za1-za2)));
	return copysign(c*c, crd[2]); // multiply switch by sign of z (crd[2])
    }

}

// Upper cutoff function
inline double g_scale_b(const double* crd, const double zb1, const double zb2)
{

    double z = std::fabs(crd[2]);
    if(z > zb2)
	return copysign(-1.0, crd[2]);
    else if(z < zb1)
	return 0.0;
    else{
	const double c = std::cos(M_PI*(z-zb2)/(2*(zb1-zb2)));
	return copysign(c*c - 1.0, crd[2]); // multiply switch by sign of z (crd[2])
    }

}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 7) {
        std::cerr << "usage: response_sfg_tcc_cutoff threads za1 za2 zb1 zb2 int_max nb POSITION_CMD > response.dat" 
	          << std::endl;
        return EXIT_FAILURE;
    }

    size_t threads(0);
    {
	std::istringstream iss(argv[1]);
	iss >> threads;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[1]
		<< "' to integer" << std::endl;
	    return EXIT_FAILURE;
	}
    }

    double za1(0);
    {
	std::istringstream iss(argv[2]);
	iss >> za1;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[2]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }
    
    double za2(0);
    {
	std::istringstream iss(argv[3]);
	iss >> za2;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[3]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }
    double zb1(0);
    {
	std::istringstream iss(argv[4]);
	iss >> zb1;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[4]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }
    
    double zb2(0);
    {
	std::istringstream iss(argv[5]);
	iss >> zb2;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[5]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }
    
    double int_max(0);
    {
	std::istringstream iss(argv[6]);
	iss >> int_max;
	if (!iss || !iss.eof()) {
	    std::cerr << "could not convert '" << argv[6]
		<< "' to real number" << std::endl;
	    return EXIT_FAILURE;
	}
    }
    
    std::string nb = argv[7];
    std::transform(nb.begin(), nb.end(), nb.begin(), ::tolower);

    for(int i = 8; i < argc; ++i){
	try {
            if (nb.compare("1b") != 0)        
	        opt::read_cmd(argv[i], frames, true);

            else
	        opt::read_cmd(argv[i], frames);

	    std::cerr <<" # Frames : " << frames.size() << std::endl;

	} catch (const std::exception& e) {
	    std::cerr << e.what() << std::endl;
	    return EXIT_FAILURE;
	}
    }

    // Open files for Writing

    // First, go over trajectory and pre-compute \mu_z and \alpha_{xx}

    const double dt = (frames[1].time - frames[0].time);
    int int_time_ind = (int) (int_max / dt);

    std::cerr << dt << "   " << int_time_ind << std::endl;

#ifdef CENTER_CELL
    #pragma omp parallel for num_threads(threads)
    for (size_t n = 0; n < frames.size(); ++n) {
        #ifdef SHIFT_HALF_CELL
        tools::center_slab_at_origin(frames[n].natm, frames[n].cell,
				     frames[n].mass, frames[n].xyz, true);
        #else
    	tools::center_slab_at_origin(frames[n].natm, frames[n].cell,
				     frames[n].mass, frames[n].xyz, false);
        #endif
    }
#endif

    for (size_t n = 0; n < frames.size(); ++n) {

	const size_t nmolec = frames[n].natm/3;

        { // calculate electrostatic properties
	    sfg::eprop eprop_this_frame;
	    eprop_this_frame.nmolec = nmolec;
	    eprop_this_frame.dip_z  = new double[eprop_this_frame.nmolec];
	    eprop_this_frame.pol_xx = new double[eprop_this_frame.nmolec];
	    eprop_this_frame.gscale_a = new double[eprop_this_frame.nmolec];
	    eprop_this_frame.gscale_b = new double[eprop_this_frame.nmolec];

	    #pragma omp parallel for num_threads(threads)
	    for(size_t i = 0; i < nmolec; ++i){
                //std::cerr << "threads = "<< omp_get_num_threads() << std::endl;
		const size_t i9 = 9*i;

		double mon[9];
		tools::image_monomer(frames[n].cell, frames[n].xyz + i9, mon);

                #ifdef JUST_SWITCH
		eprop_this_frame.dip_z[i] = 1.0;
		eprop_this_frame.pol_xx[i] = 1.0;
                #else
	
      	        // z-component of the dipole
		double tmp_dip[3];
		ps::dipole(mon, tmp_dip);

                if (nb.compare("1b") != 0) {
		    double moldip_z = frames[n].dipmol[3*i+2];
		    double inddip_z = frames[n].dipind[3*i+2];
//		    eprop_this_frame.dip_z[i] = tmp_dip[2] + inddip_z;
		    eprop_this_frame.dip_z[i] = moldip_z + inddip_z;
                }
                else
 		    eprop_this_frame.dip_z[i] = tmp_dip[2];
//		    double moldip_z = frames[n].dipmol[3*i+2];
//		    eprop_this_frame.dip_z[i] = moldip_z;

		// xx-component of the polarizability
		double pol_mol[9];
		pol_model.alpha(mon, pol_mol);

		eprop_this_frame.pol_xx[i] = pol_mol[0];

                #endif
		
                // scaling that accounts for the double-interface
		double ga_i = g_scale_a(frames[n].xyz + i9, za1, za2);
		double gb_i = g_scale_b(frames[n].xyz + i9, zb1, zb2);
		eprop_this_frame.gscale_a[i] = ga_i;
		eprop_this_frame.gscale_b[i] = gb_i;
	    }

	    eprops.push_back(eprop_this_frame);
	}

        { // determine connectivity

	    sfg::connect connect_this_frame;
	    connect_this_frame.nmolec = nmolec;

            //#pragma omp parallel for num_threads(threads)
	    for (size_t i = 0; i < nmolec; ++i) { 

		if ((eprops[n].gscale_a[i] + eprops[n].gscale_b[i]) != 0.0) {

                    #ifdef DO_TCC //Begin partner determination
		    const size_t i9 = 9*i;
		    std::vector<int> mol;

		    #pragma omp parallel for num_threads(threads)
    		    for(size_t j = 0; j < nmolec; ++j){
			double avg_rij(0);
			size_t count(0);
			if(j != i){

			    // Assume that int_time_ind is short enough to 
			    // prohibit diffusion between solvation shells
			    // and that the average_rij cannot be < rOO_max
			    // unless rij at the initial time is 
			    // less than 1.5*rOO_max
			    if(tools::distance3(frames[n].imcon, frames[n].cell,
						frames[n].rcell,
						frames[n].xyz + i9,
						frames[n].xyz + j*9)
			        < 2.0*rOO_max){

				// calculate average rij distance over the next
				// int_time_ind frames
				for(size_t a = n;
				    a < std::min( frames.size(),
				    		  n + int_time_ind + 1);
				    a += 250){

				    #pragma omp atomic
    				    avg_rij += tools::distance3(frames[n].imcon,
				    			         frames[n].cell,
								frames[n].rcell,
							     frames[n].xyz + i9,
							   frames[a].xyz + j*9);

				    #pragma omp atomic
				    ++count;
				}
 
				if ((avg_rij/count) < rOO_max) {
				    #pragma omp critical
				    {
			            mol.push_back((int)j);
	  		            }
				}
			    }
			}
		    }

    		    connect_this_frame.partners.push_back(mol);
                    #endif // end partner determination

		    connect_this_frame.central.push_back(i);
		}
	    }

 	    connects.push_back(connect_this_frame);
	}
    }

    if (eprops.size() != frames.size()) {
	std::cerr << "eprops.size() != frames.size()" << std::endl;
	exit(1);
    }

    std::cerr << "Deallocating frames" << std::endl;
    //frames.erase(frames.begin(), frames.end());
    std::vector<opt::hist_frame>().swap(frames);

    // With electrostatic properties loaded in eprops
    // Calculated the 1B approximation to the SFG with the 
    // truncated cross-correlation approximation of dx.doi.org/10.1021/jz400683v
    
    double corr[int_time_ind];
    std::fill(corr, corr + int_time_ind, 0.0);
    int ntime[int_time_ind];
    std::fill(ntime, ntime + int_time_ind, 0.0);

    std::cerr << "Beginning correlation function" << std::endl;

    //omp_set_dynamic(0);
    //std::cerr << "threads = "<< omp_get_num_threads() << std::endl;
    #pragma omp parallel for num_threads(threads)
    for(size_t m = 0; m < eprops.size(); ++m){ // dipole, time = 0
	for(size_t n = m; //polarizability, time = 't'
	           n < std::min( eprops.size(), m + int_time_ind); ++ n){ 

	    int time = n - m;

	    double* dip_z_0  = eprops[m].dip_z;
	    double* pol_xx_t = eprops[n].pol_xx;

	    for(size_t i = 0; i < connects[m].central.size(); ++i){
		const size_t i_ind = connects[m].central[i];

		const double ga_i0 = eprops[m].gscale_a[i_ind];
		const double gb_i0 = eprops[m].gscale_b[i_ind];
		const double g_i0 = ga_i0 + gb_i0;

		// autocorrelation, j == i
		{

                    #ifdef TYPE2
    		    const double ga_it = eprops[n].gscale_a[i_ind];
    		    const double gb_it = eprops[n].gscale_b[i_ind];
		    const double g_it = ga_it + gb_it;
		    const double g3 = g_i0*g_it*g_it;
                    #else
	    	    const double g3 = g_i0*g_i0*g_i0;
                    #endif

		    {
	    	    #pragma omp atomic 
		    corr[time] += g3*dip_z_0[i_ind]*pol_xx_t[i_ind];
		    }

		}

                #ifdef DO_TCC
		// truncated cross-correlation
		for(size_t j = 0; j < connects[m].partners[i].size(); ++j){
	    	    const size_t j_ind = connects[m].partners[i][j];

		    const double ga_j0 = eprops[n].gscale_a[j_ind];
		    const double gb_j0 = eprops[n].gscale_b[j_ind];
		    const double g_j0 = ga_j0 + gb_j0;
		
                    {
		    #pragma omp atomic
		    corr[time] += g_i0*g_j0*g_j0*dip_z_0[i_ind]*pol_xx_t[j_ind];
		    }
		}
                #endif
	    }

	    #pragma omp atomic
	    ++ntime[time];
	}
    }

    std::cout << std::setprecision(10) << std::scientific;

    for(int i = 0; i < int_time_ind; ++i)
	std::cout << std::setw(20) << dt*i 
	          << std::setw(20) << corr[i]/((double)ntime[i])
	          << std::setw(20) << corr[i]
	          << std::setw(20) << ((double)ntime[i])
		  << std::endl;

}

