#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include "read-hist.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

const double r2f = 7.5;
const double dr = 0.01;
const double r_max = 7.5;

int pairs_per_frame = 0;
double invvolm = 0;

//----------------------------------------------------------------------------//

inline double norm(const int nentry, const double* arr)
{
    double val(0);
    for(int i = 0; i < nentry; ++i)
	val += arr[i]*arr[i];

    return std::sqrt(val);
}

//----------------------------------------------------------------------------//

inline double distance(int imcon, const double* cell, 
		       const double* r1, const double* r2)
{
    assert(imcon == 0 || imcon == 1);

    double dx = r1[0] - r2[0];
    double dy = r1[1] - r2[1];
    double dz = r1[2] - r2[2];

    dx -= cell[0]*std::floor(dx/cell[0] + 0.5);
    dy -= cell[4]*std::floor(dy/cell[4] + 0.5);
    dz -= cell[8]*std::floor(dz/cell[8] + 0.5);

    return std::sqrt(dx*dx + dy*dy + dz*dz);
}

//----------------------------------------------------------------------------//

void dump_1d_distrib_norm(const char* filename, const int* bin,
			  const double dr, const int nbins)
{
    std::ofstream ofs;
    ofs.open(filename);
    ofs.precision(9);

    ofs << std::scientific;

    const double norm_fac = 1.0/invvolm/pairs_per_frame;

    for(int i = 0; i < nbins; ++i){
	const double r = dr*(i+0.5);

	const double rm = r - dr/2;
	const double rp = r + dr/2;

	const double dvol = 4.0/3.0*M_PI*(std::pow(rp, 3) - std::pow(rm, 3));

	ofs << std::setw(20) << r
	    << std::setw(20) << norm_fac*bin[i]/dvol << std::endl;
    }

    ofs.close();
}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 1) {
        std::cerr << "usage: bulk_dipole ..." << std::endl;
        return EXIT_FAILURE;
    }

    const int nbin_r    = (int)(r_max/dr);
    int bin_rdf[nbin_r];
    std::fill(bin_rdf, bin_rdf + nbin_r, 0.0);

    for(int i = 1; i < argc; ++i){

	std::ifstream ifs(argv[i]);
	size_t lineno(0);
	int frameno = 0;

	while(!ifs.eof()){

	    opt::hist_frame frame;
	    opt::read_cmd_frame(lineno, ifs, frame);
	    if(frame.natm > 0){
		const int nmolec = frame.natm/3;

		int pairs_this_frame(0);
    		for(int i = 0; i < nmolec; ++i){
		    for(int j = i+1; j < nmolec; ++j){
			const int i9 = 9*i;
			const int j9 = 9*j;

    			double rOO = distance(frame.imcon, frame.cell, 
			    		      frame.xyz + i9, 
					      frame.xyz + j9);

    			if(rOO < r2f){
			    const int ind_Roo = (int)(rOO/dr);
			    ++bin_rdf[ind_Roo];
			}
			++pairs_this_frame;
		    }
		}

		if(pairs_per_frame == 0)
		    pairs_per_frame = pairs_this_frame;
		else if(pairs_this_frame != pairs_per_frame){
		    std::cerr << "Dying. Not a constant number of atm per frame"
			<< std::endl;
		    exit(1);
		}

		invvolm += 1.0/(frame.vol);
		++frameno;
	    } // frame.natm > 0
	}
	std::cerr << "Read " << frameno << " frames from " << argv[i]
	          << std::endl;

    }

//    dump_1d_distrib("unnorm_rdf.dat", bin_rdf, dr, nbin_r);
    dump_1d_distrib_norm("norm_rdf.dat", bin_rdf, dr, nbin_r);

    return 0;
}
