#ifdef HAVE_CONFIG_H
#   include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>
#include <cstddef>

#include "avila.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

//----------------------------------------------------------------------------//

const double a0 = 0.529177249; // A
const double toA3 = (10.0/1.648778)*std::pow(a0, 3);

const double sqrt2 = std::sqrt(2.0);

//----------------------------------------------------------------------------//

struct term {
    int i:8;
    int j:8;
    int k:8;

    double v;
};

//----------------------------------------------------------------------------//

static const term aijk_xz[] = {
    {0, 0, 1,  1.17904},
    {0, 1, 1, -0.37700},
    {1, 0, 1,  1.85559},
    {0, 0, 3,  0.00698},
    {0, 2, 1, -0.36099},
    {1, 1, 1, -0.00202},
    {2, 0, 1,  0.62070},
    {0, 1, 3,  0.53070},
    {1, 0, 3, -0.64252},
    {1, 2, 1, -0.63496},
    {0, 3, 1, -0.12823},
    {2, 1, 1,  1.40356},
    {3, 0, 1, -1.24044}
};

//----------------------------------------------------------------------------//

static const term aijk_xx[] = {
    {0, 0, 0,  1.58787},
    {0, 1, 0, -0.07614},
    {1, 0, 0,  1.60710},
    {0, 0, 2,  0.71119},
    {0, 2, 0,  0.13331},
    {2, 0, 0,  0.61842},
    {1, 1, 0, -0.70521},
    {0, 3, 0,  0.05328},
    {3, 0, 0, -0.20959},
    {0, 1, 2, -1.08201},
    {1, 0, 2,  0.35624},
    {1, 2, 0,  0.00892},
    {2, 1, 0, -0.35495},
    {0, 0, 4, -0.22179},
    {0, 4, 0, -0.03182},
    {4, 0, 0, -0.51709},
    {0, 2, 2,  0.34644},
    {1, 1, 2,  0.01374},
    {2, 0, 2, -1.25979},
    {2, 2, 0, -0.03825},
    {3, 1, 0,  0.18326},
    {1, 3, 0,  0.17426}
};

//----------------------------------------------------------------------------//

static const term aijk_yy[] = {
    {0, 0, 0,  1.53818},
    {0, 1, 0,  0.08361},
    {1, 0, 0,  0.82863},
    {0, 0, 2, -0.04232},
    {0, 2, 0,  0.05666},
    {2, 0, 0,  0.08172},
    {1, 1, 0,  0.08557},
    {0, 3, 0, -0.00271},
    {3, 0, 0, -0.11892},
    {0, 1, 2, -0.06926},
    {1, 0, 2, -0.10806},
    {1, 2, 0,  0.08190},
    {2, 1, 0,  0.06999},
    {0, 0, 4, -0.05890},
    {0, 4, 0, -0.00800},
    {4, 0, 0, -0.11111},
    {0, 2, 2,  0.12995},
    {1, 1, 2, -0.09119},
    {2, 0, 2, -0.48361},
    {2, 2, 0, -0.03088},
    {3, 1, 0, -0.09411},
    {1, 3, 0,  0.05992}
};

//----------------------------------------------------------------------------//

static const term aijk_zz[] = {
    {0, 0, 0,  1.64720},
    {0, 1, 0,  0.24730},
    {1, 0, 0,  2.39394},
    {0, 0, 2,  0.89549},
    {0, 2, 0,  0.03954},
    {2, 0, 0,  1.99483},
    {1, 1, 0,  0.99680},
    {0, 3, 0, -0.08900},
    {3, 0, 0,  0.53619},
    {0, 1, 2,  0.96599},
    {1, 0, 2,  0.41179},
    {1, 2, 0, -0.08290},
    {2, 1, 0,  1.62467},
    {0, 0, 4,  0.04290},
    {0, 4, 0, -0.02203},
    {4, 0, 0, -0.62916},
    {0, 2, 2, -0.20222},
    {1, 1, 2,  3.29320},
    {2, 0, 2, -2.27559},
    {2, 2, 0,  0.15971},
    {3, 1, 0,  1.20526},
    {1, 3, 0, -0.18405}
};

//----------------------------------------------------------------------------//

} // namespace

////////////////////////////////////////////////////////////////////////////////

namespace x2o {

//----------------------------------------------------------------------------//

const double avila::r0 = 0.95843;
const double avila::phi0 = 104.44*M_PI/180.0;

//----------------------------------------------------------------------------//

#define DEFINE_AKL(kl) \
double avila::a_##kl(const double& dS1, const double& dS2, const double& dS3) \
{ \
    double result(0); \
    for (size_t n = 0; n < sizeof(aijk_##kl)/sizeof(aijk_##kl[0]); ++n) \
        result += aijk_##kl[n].v*std::pow(dS1, aijk_##kl[n].i) \
                                *std::pow(dS2, aijk_##kl[n].j) \
                                *std::pow(dS3, aijk_##kl[n].k); \
    return result*toA3; \
}

DEFINE_AKL(xx)
DEFINE_AKL(yy)
DEFINE_AKL(zz)
DEFINE_AKL(xz)

#undef DEFINE_AKL

//----------------------------------------------------------------------------//

void avila::cartesian
    (const double& dS1, const double& dS2, const double& dS3, double xyz[3])
{
    const double r1 = r0 + (dS1 - dS3)/sqrt2;
    const double r2 = r0 + (dS1 + dS3)/sqrt2;

    const double phi2 = (phi0 + dS2)/2;

    // O
    xyz[0] = xyz[1] = xyz[2] = 0.0;

    // H1
    xyz[3] = -r1*std::cos(phi2);
    xyz[4] = 0.0;
    xyz[5] =  r1*std::sin(phi2);

    // H2
    xyz[6] = -r2*std::cos(phi2);
    xyz[7] = 0.0;
    xyz[8] = -r2*std::sin(phi2);
}

//----------------------------------------------------------------------------//

void avila::alpha(const double xyz[9], double a[9])
{
    double oh1[3];
    double oh2[3];

    double oh1sq(0), oh2sq(0), oh1oh2(0);

    for (int i = 0; i < 3; ++i) {
        oh1[i] = xyz[3 + i] - xyz[i];
        oh1sq += oh1[i]*oh1[i];

        oh2[i] = xyz[6 + i] - xyz[i];
        oh2sq += oh2[i]*oh2[i];

        oh1oh2 += oh1[i]*oh2[i];
    }

    const double oh1n = std::sqrt(oh1sq);
    const double oh2n = std::sqrt(oh2sq);

    const double dr1 = oh1n - r0;
    const double dr2 = oh2n - r0;
    const double cosphi = oh1oh2/(oh1n*oh2n);
    const double phi = std::acos(cosphi);

    const double dS1 = (dr1 + dr2)/sqrt2;
    const double dS2 = phi - phi0;
    const double dS3 = (dr2 - dr1)/sqrt2;

    double am[9] = {
        a_xx(dS1, dS2, dS3),                 0.0, a_xz(dS1, dS2, dS3),
                        0.0, a_yy(dS1, dS2, dS3),                 0.0,
                        0.0,                 0.0, a_zz(dS1, dS2, dS3)
    };

    am[6] = am[2]; // zx = xz

    // transformation matrix

    double T[9];

    for (int i = 0; i < 3; ++i) {
        oh1[i] /= oh1n;
        oh2[i] /= oh2n;

        // ex -- along HOH bisector, towards O
        T[0 + i] = - (oh1[i] + oh2[i])/std::sqrt(2 + 2*cosphi);

        // ez -- towards H1
        T[6 + i] = (oh1[i] - oh2[i])/std::sqrt(2 - 2*cosphi);
    }

    // ey = ez x ex

    T[3 + 0] = T[6 + 1]*T[0 + 2] - T[6 + 2]*T[0 + 1];
    T[3 + 1] = T[6 + 2]*T[0 + 0] - T[6 + 0]*T[0 + 2];
    T[3 + 2] = T[6 + 0]*T[0 + 1] - T[6 + 1]*T[0 + 0];

    // a = T^{t} * am * T
    for (int i = 0; i < 3; ++i)
        for (int j = 0; j < 3; ++j) {
            a[3*i + j] = 0.0;
            for (int k = 0; k < 3; ++k)
                for (int l = 0; l < 3; ++l)
                    a[3*i + j] += T[3*k + i]*T[3*l + j]*am[3*k + l];
        }
}

//----------------------------------------------------------------------------//

void avila::alpha6(const double xyz[9], double a[6])
{
    double a3x3[9];
    alpha(xyz, a3x3);

    a[0] = a3x3[0];
    a[1] = a3x3[4];
    a[2] = a3x3[8];
    a[3] = a3x3[1];
    a[4] = a3x3[2];
    a[5] = a3x3[5];
}

//----------------------------------------------------------------------------//

} // namespace x2o

////////////////////////////////////////////////////////////////////////////////
