#include <cmath>
#include <iostream>

#include <limits>
#include <algorithm> // for max

#include "interpolation.h"

////////////////////////////////////////////////////////////////////////////////

namespace tools {

////////////////////////////////////////////////////////////////////////////////

double lagrange_interpolation(const size_t& ndata, const double* x, 
  			      const double* f, double& a)
{

//
// x are independent variables with f(x) known
// a is x value to interpolate the b value 
//

    double l[ndata];

    for(size_t i = 0; i < ndata; ++i){
	l[i] = 1.0;
	for(size_t j = 0; j < ndata; ++j){
	    if(j != i){
		l[i] *= (a - x[j])/(x[i] - x[j]);
	    }
	}
    }

    double b(0);
    for(size_t i = 1; i < ndata; ++i){
	b += f[i]*l[i];
    }

    return b;
}

/////
//
//  Cubic Splines ------------------------------------------------------------//
//
/////

double cubic_spline(const size_t ndata, const double* x, 
		    const double* y, double xq)
{

//
// x are independent variables with y(x) known
// xq is x value to interpolate the b value 
//
// First begin by constructing the Tridiagonal matrix for determining 
// second derivatives
//    

    double e[ndata];
    double f[ndata];
    double g[ndata];
    double b[ndata];

    // First treat non-boundaries

    for(size_t i = 1; i < ndata - 1; ++i){
	size_t h = i - 1;
	size_t j = i + 1;

	const double delta0 = 1.0/(x[i] - x[h]);
	const double delta1 = 1.0/(x[j] - x[i]);

	e[i] = delta0;
	f[i] = 2.0*(delta0 + delta1);
	g[i] = delta1;

	b[i] = 3.0*(  delta0*delta0*(y[i] - y[h])
	            + delta1*delta1*(y[j] - y[i]));

    }	

    // For simplicity, assume natural boundary conditions
    // First, i = 0

    {

	size_t i = 0;
	size_t j = i + 1;

	const double delta1 = 1.0/(x[j] - x[i]);

	e[i] = 0.0;
	f[i] = 2.0*(delta1);
	g[i] = delta1;

	b[i] = 3.0*(delta1*delta1*(y[j] - y[i]));

    }

    // Finally, treat i = ndata - 1

    {

	size_t i = ndata - 1;
	size_t h = i - 1;

	const double delta0 = 1.0/(x[i] - x[h]);

	e[i] = delta0;
	f[i] = 2.0*(delta0);
	g[i] = 0.0;

	b[i] = 3.0*(  delta0*delta0*(y[i] - y[h]));

    }

    // With the tridiagonal matrices populated, solve for second derivatives
    // using the thomas algorithm

    double g2[ndata];

    ThomasTriDiag_efg(ndata, e, f, g, b, g2);

    // With the second derivatives, we can obtain the value of the spline
   
    for(size_t i = 1; i < ndata; ++i){
	size_t h = i - 1;
//	print h, "-", i,"  :  ", x[h], " < ", xq, " < ",x[i], "\n";

	if(xq > x[h]){
	    if(xq <= x[i]){
		const double delta0 = x[i] - x[h];
		const double      t = (xq - x[h])/(delta0);
		const double     ai =  g2[h]*(delta0) - (y[i] - y[h]);
		const double     bi = -g2[i]*(delta0) + (y[i] - y[h]);
		const double      q = (1.0-t)*y[h] + t*y[i]
		                    + t*(1.0-t)*(ai*(1.0-t) + bi*t);

		return q;
	    }else if(i == ndata - 1){ // approximate point outside 
		                      // postive bound with 
				      // natrual runoff
		const double delta0 = x[i] - x[h];
		const double      t = (xq - x[h])/(delta0);
		const double     ai =   g2[h]*(delta0) - (y[i] - y[h]);
		const double     bi =  -g2[i]*(delta0) + (y[i] - y[h]);
		const double      q = (1.0-t)*y[h] + t*y[i]
	                            + t*(1.0-t)*(ai*(1.0-t) + bi*t);

		return q;
	    }
	}else if(xq <= x[i] && i == 1){ // approximate point outside negative
	                                // bound w/ natural runoff
	    const double delta0 = x[i] - x[h];
	    const double      t = (xq - x[h])/(delta0);
	    const double     ai =  g2[h]*(delta0) - (y[i] - y[h]);
	    const double     bi = -g2[i]*(delta0) + (y[i] - y[h]);
	    const double      q = (1.0-t)*y[h] + t*y[i]
	                        + t*(1.0-t)*(ai*(1.0-t) + bi*t);

	    return q;
	}
    }

    std::cerr << "Requested interpolated point, xq, not bounded by data"
	      << std::endl;
    exit(1);
}

/////
//
//  Thomas Algorithm ---------------------------------------------------------//
//
/////

// Thomas Algorithm for tri-diagonal matrix 'A' uses vectors e, f, and g
// f_1  g_1  --- ---
// e_2  f_2  g_2 ---
// ---  e_3  f_3 g_3
// ---  ---  e_4 f_4

void ThomasTriDiag_efg(const size_t ndata, double* e, double* f, 
		       const double* g, double* b, double* x)
{

// Note, arrays e, f, and b are overwritten

    for(size_t i = 1; i < ndata; ++i){
	size_t j = i - 1;

    	const double fac = e[i]/f[j];
	e[i] = fac;
	f[i] = f[i] - fac*g[j]; // only need to worry about f for tri-diag
                                // b/c nothing above g_i and e_i must be zero
	b[i] = b[i] - fac*b[j];
    }

    //
    // Now calculate x through back-substitution
    //

    x[ndata-1] = b[ndata-1]/f[ndata-1]; // special case
    for(int i = ndata-2; i >= 0; --i){
	x[i] = (b[i] - g[i]*x[i+1])/f[i];
    }

//   for(size_t i = 0; i < ndata; ++i)
//     std::cout << e[i] << ' ' <<  f[i] << ' ' <<  g[i] << ' ' <<  x[i]
//               << std::endl;
}

#if 0
void ThomasTriDiag_matA(const size_t m, const double* A, const double* b, 
			const double* x)
{

#% Thomas Algorithm for tri-diagonal matrix 'A' uses vectors e, f, and g
#% f_1  g_1  --- ---
#% e_2  f_2  g_2 ---
#% ---  e_3  f_3 g_3
#% ---  ---  e_4 f_4
#%

    my @e = ();
    my @f = ();
    my @g = ();

    for(my i = 0; i < m; ++i){
	x[i] = 0;
	my h = i - 1;
	my j = i + 1;

    	if(i > 0){
	    e[i] = A{"i:h"}
	}else{
	    e[i] = 0;
	}

    	f[i] = A{"i:i"};

    	if(i < m-1){
	    g[i] = A{"i:j"};
        }else{
    	    g[i] = 0;
	}
    }

    for(my i = 1; i < m; ++i){
	my j = i - 1;

    	my fac = e[i]/f[j];
	e[i] = fac ;
	f[i] = f[i] - fac*g[j]; # only need to worry about f for tri-diag
	                            # b/c nothing above g_i and e_i must be zero
	b[i] = b[i] - fac*b[j];
    }

    #
    # Now calculate x through back-substitution
    #

    x[m-1] = b[m-1]/f[m-1];#% special case
    for(my i = m-2; i >= 0; i -= 1){
	x[i] = (b[i] - g[i]*x[i+1])/f[i];
    }

#    for(my i = 0; i < m; ++i){
#	printf "%12.6f %12.6f %12.6f %12.6f\n", e[i], f[i], g[i], x[i];
#    }
}
#endif

////////////////////////////////////////////////////////////////////////////////

} //namespace
