#include <cassert>

#include <cstring>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <algorithm>
#include <stdexcept>

#include "read-bin.h"
#include "io-tools.h"

////////////////////////////////////////////////////////////////////////////////

namespace opt {

//----------------------------------------------------------------------------//

void read_bin_frame(std::istream& is, opt::hist_frame& frame) 
{
    assert(is && !is.eof());
    assert(frame.natm == 0); // empty

    is.read((char*)&(frame.timestep), sizeof(int));
    {
	float tmp;
	is.read((char*)(&tmp)    , sizeof(float));
	frame.time = (double) tmp;
    }
    is.read((char*)&(frame.natm)    , sizeof(int));

    if (frame.natm == 0)
	return;

    frame.xyz        = new double[3*frame.natm];
    frame.name       = new char[frame.natm*MAX_CHAR_LENGTH];
    frame.atomic_num = new int[frame.natm];
    frame.mass       = new double[frame.natm];
    frame.cell       = new double[9];
    frame.rcell       = new double[9];
    frame.dipind       = new double[frame.natm];
    frame.dipmol       = new double[frame.natm];

    std::fill(frame.name, frame.name + frame.natm*MAX_CHAR_LENGTH, '0');

    std::fill(frame.cell, frame.cell + 9, 0.0);
    for(size_t i = 0; i < 9; ++i){
     	float a;
	is.read((char*)(&a)    , sizeof(float));
	frame.cell[i] = (double) a;
    }

    frame.imcon = 3;

    {
	const double xx = frame.cell[0];
	const double xy = frame.cell[1];
	const double xz = frame.cell[2];
	const double yx = frame.cell[3];
	const double yy = frame.cell[4];
	const double yz = frame.cell[5];
	const double zx = frame.cell[6];
	const double zy = frame.cell[7];
	const double zz = frame.cell[8];

	const double cross_x = yy*zz - yz*zy;
	const double cross_y = -xy*zz + zy*xz;
	const double cross_z = xy*yz - yy*xz;

	frame.vol = xx*cross_x + yx*cross_y + zx*cross_z; // [a] . [b] x [c]

	opt::calculate_rcell(0, frame.cell, frame.rcell);
    }

    for(size_t i = 0; i < frame.natm; i++) {  // Loop over interfacial waters
	// read atom name
        { 
	    char c_tmp[8];
	    std::fill(c_tmp, c_tmp + 8, '0');
	    for(size_t j = 0; j < 4; ++j){
		char my_char;
		is.read((char*)(&my_char)  , sizeof(char));
		c_tmp[j] = my_char;
	    }

	    std::string strname(c_tmp);

	    if(strname.length() > (MAX_CHAR_LENGTH - 1)){
		std::stringstream error;
		error << "Atom Type " << strname
		    << " exceeds maximum character length of "
		    << MAX_CHAR_LENGTH-1 << "\n";
		std::runtime_error(error.str());
	    }

	    if((strname.find("OW")) != std::string::npos){
		frame.mass[i] = 15.994900;
		frame.atomic_num[i] = 8;
	    } else if((strname.find("Na")) != std::string::npos){
		frame.mass[i] = 22.989769;
		frame.atomic_num[i] = 11;
	    } else if((strname.find("Br")) != std::string::npos){
		frame.mass[i] = 79.904;  
		frame.atomic_num[i] = 35;
	    } else if((strname.find("Cl")) != std::string::npos){
		frame.mass[i] = 35.453;  
		frame.atomic_num[i] = 17;
	    } else if((strname.find("I")) != std::string::npos){
		frame.mass[i] = 126.90447;
		frame.atomic_num[i] = 53;
	    } else if((strname.find("HW")) != std::string::npos){
		frame.mass[i] = 1.007900;
		frame.atomic_num[i] = 1;
	    }else{
		frame.mass[i] = -1;
		frame.atomic_num[i] = 0;
	    }

	    std::strcpy(frame.name + MAX_CHAR_LENGTH*i, strname.c_str());

	}

        // Loop over X, Y, Z
        for(size_t j = 0; j < 3; j++) {    
	    float tmp;
	    is.read((char*)(&tmp)  , sizeof(float));
	    frame.xyz[3*i + j] = (double) tmp;
	}
    }

    for (size_t n = 0; n < frame.natm; ++n) {
        if (is.eof()) 
    	    opt::print_error(0, 1);

	frame.dipmol[n] = 0.0;
	frame.dipind[n] = 0.0;
    }
}

//----------------------------------------------------------------------------//

size_t read_bin
    (const char* filename, std::vector<hist_frame>& frames)
{
    std::ifstream ifs(filename, std::ios_base::in | std::ios_base::binary);

    if (!ifs)
        throw std::runtime_error("could not open file for reading");

    size_t frameno(0);

    while (!ifs.eof()) {
        hist_frame frame;

	read_bin_frame(ifs, frame);

        if (frame.natm > 0) {

            frames.push_back(frame);
            ++frameno;

        }
    }

    return frameno;
}

////////////////////////////////////////////////////////////////////////////////

} // namespace opt

////////////////////////////////////////////////////////////////////////////////
