#include <cassert>

#include <cstring>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <algorithm>
#include <stdexcept>

#include "read-hist.h"
#include "io-tools.h"

////////////////////////////////////////////////////////////////////////////////

namespace opt {

    std::string    pos_name = "POSITION_CMD";
    std::string dipmol_name = "DIPMOL_CMD";
    std::string dipind_name = "DIPIND_CMD";

    bool replace(std::string& str, const std::string& from, const std::string& to) {
	size_t start_pos = str.find(from);
	if(start_pos == std::string::npos)
	    return false;
	str.replace(start_pos, from.length(), to);
	return true;
    }

//----------------------------------------------------------------------------//

void read_hist_frame(size_t& lineno, std::istream& is, opt::hist_frame& frame)
{
    assert(is && !is.eof());
    assert(frame.natm == 0); // empty


    std::string string_junk;
    size_t int_junk;
    // parse first line of frame

    std::string line;
    std::getline(is, line); 
    if (is.eof())
        return;
    ++lineno;

    std::istringstream iss(line);
    iss >> string_junk >> frame.timestep >> frame.natm >> int_junk 
	>> frame.imcon >> frame.time;
    if (iss.fail()) 
	print_error(lineno, 0);

    if (frame.natm == 0)
        throw std::runtime_error("No atoms in the XYZ frame");

    frame.xyz        = new double[3*frame.natm];
    frame.name       = new char[frame.natm*MAX_CHAR_LENGTH];
    frame.atomic_num = new int[frame.natm];
    frame.mass       = new double[frame.natm];
    frame.cell       = new double[9];
    frame.rcell      = new double[9];
    frame.dipind     = new double[frame.natm]; // 3 dim * n_wat
    frame.dipmol     = new double[frame.natm]; // 3 dim * n_wat

    std::fill(frame.name, frame.name + frame.natm*MAX_CHAR_LENGTH, '0');

    // parse unit cell dimensions (lines 2-4 of frame)
    double xx(0), xy(0), xz(0);
    double yx(0), yy(0), yz(0);
    double zx(0), zy(0), zz(0);

    std::getline(is, line); 
    ++lineno;
    iss.clear();
    iss.str(line);
    iss >> xx >> xy >> xz;
    if (iss.fail()) 
	print_error(lineno, 0);
    frame.cell[0] = xx;
    frame.cell[1] = xy;
    frame.cell[2] = xz;

    std::getline(is, line); 
    ++lineno;
    iss.clear();
    iss.str(line);
    iss >> yx >> yy >> yz;
    if (iss.fail()) 
	print_error(lineno, 0);
    frame.cell[3] = yx;
    frame.cell[4] = yy;
    frame.cell[5] = yz;

    std::getline(is, line); 
    ++lineno;
    iss.clear();
    iss.str(line);
    iss >> zx >> zy >> zz;
    if (iss.fail()) 
	print_error(lineno, 0);
    frame.cell[6] = zx;
    frame.cell[7] = zy;
    frame.cell[8] = zz;

    if(frame.imcon < 3){// orthorhombic (or orthorhombic)

	frame.vol = xx*yy*zz;

	std::fill(frame.rcell, frame.rcell + 9, 0.0);
	frame.rcell[0] = 1.0/frame.cell[0];
	frame.rcell[4] = 1.0/frame.cell[4];
	frame.rcell[8] = 1.0/frame.cell[8];

    }else if(frame.imcon == 3){ // non-orthorhombic
	double cross_x = yy*zz - yz*zy; 
	double cross_y = -xy*zz + zy*xz;
	double cross_z = xy*yz - yy*xz;

	frame.vol = xx*cross_x + yx*cross_y + zx*cross_z; // [a] . [b] x [c]

	calculate_rcell(lineno, frame.cell, frame.rcell);

    }else{
	std::ostringstream oss;
	oss << "Do not know how to calculate volume for imcon = " 
     	    << frame.imcon << std::endl;
	throw std::runtime_error(oss.str());
    }

    for (size_t n = 0; n < frame.natm; ++n) {
        if (is.eof()) 
    	    print_error(lineno, 1);

        std::getline(is, line);
        ++lineno;

	std::string strname;

	iss.clear();
    	iss.str(line);
	iss >> strname;
        if (iss.fail()) 
    	    print_error(lineno, 0);

	if(strname.length() > (MAX_CHAR_LENGTH - 1)){
	    std::stringstream error;
	    error << "Atom Type " << strname
		<< " exceeds maximum character length of "
		<< MAX_CHAR_LENGTH-1 << "\n";
	    std::runtime_error(error.str());
	}

    	if((strname.find("OW")) != std::string::npos){
	    frame.mass[n] = 15.994900;
	    frame.atomic_num[n] = 8;
	} else if((strname.find("HW")) != std::string::npos){
	    frame.mass[n] = 1.007900;
	    frame.atomic_num[n] = 1;
    } else if((strname.find("Cl")) != std::string::npos){
        frame.mass[n] = 35.453;
        frame.atomic_num[n] = 17;
    } else if((strname.find("Br")) != std::string::npos){
        frame.mass[n] = 79.904;
        frame.atomic_num[n] = 35;
    } else if((strname.find("I")) != std::string::npos){
        frame.mass[n] = 126.90447;
        frame.atomic_num[n] = 53;
    } else if((strname.find("Na")) != std::string::npos){
        frame.mass[n] = 22.989769;
        frame.atomic_num[n] = 11;
	}else{
	    frame.mass[n] = -1;
	    frame.atomic_num[n] = 0;
	}

	std::strcpy(frame.name + MAX_CHAR_LENGTH*n, strname.c_str());

    	std::getline(is, line);
        ++lineno;
    	iss.clear();
    	iss.str(line);
    	iss >> frame.xyz[3*n + 0] >> frame.xyz[3*n + 1] >> frame.xyz[3*n + 2];
        if (iss.fail()) 
    	    print_error(lineno, 0);

	frame.dipmol[n] = 0.0;
	frame.dipind[n] = 0.0;
    }
}

//----------------------------------------------------------------------------//

void read_cmd_frame(size_t& lineno, std::istream& is, opt::hist_frame& frame) 
{
    assert(is && !is.eof());
    assert(frame.natm == 0); // empty

    std::string string_junk;
    // parse first line of frame

    std::string line;
    std::getline(is, line); 
    if (is.eof()) // check eof before defining/allocating variables
        return;
    ++lineno;

    std::istringstream iss(line);
    iss >> frame.timestep >> frame.time >> frame.natm;
    if (iss.fail()) 
	print_error(lineno, 0);

    if (frame.natm == 0)
        throw std::runtime_error("No atoms in the XYZ frame");

    frame.xyz        = new double[3*frame.natm];
    frame.name       = new char[frame.natm*MAX_CHAR_LENGTH];
    frame.atomic_num = new int[frame.natm];
    frame.mass       = new double[frame.natm];
    frame.cell       = new double[9];
    frame.rcell       = new double[9];
    frame.dipind       = new double[frame.natm]; // 3 dim * n_wat
    frame.dipmol       = new double[frame.natm]; // 3 dim * n_wat

    std::fill(frame.name, frame.name + frame.natm*MAX_CHAR_LENGTH, '0');

    // read unit cell dimensions (lines 2-4 of frame)
    for(size_t i = 0; i < 3; ++i){

	std::string line;
	std::getline(is, line); 
	if (is.eof()) // check eof before defining/allocating variables
	    return;
	++lineno;

	std::istringstream iss(line);
	iss >> frame.cell[3*i] >> frame.cell[3*i + 1] >> frame.cell[3*i + 2];
	if (iss.fail()) 
	    print_error(lineno, 0);

    }

    frame.imcon = 1;

    frame.vol = frame.cell[0]*frame.cell[4]*frame.cell[8];

    std::fill(frame.rcell, frame.rcell + 9, 0.0);
    frame.rcell[0] = 1.0/frame.cell[0];
    frame.rcell[4] = 1.0/frame.cell[4];
    frame.rcell[8] = 1.0/frame.cell[8];

    for (size_t n = 0; n < frame.natm; ++n) {
        if (is.eof()) 
    	    print_error(lineno, 1);

        std::getline(is, line);
        ++lineno;

	std::string strname;

	iss.clear();
    	iss.str(line);
	iss >> strname >> frame.xyz[3*n + 0] 
	    >> frame.xyz[3*n + 1] >> frame.xyz[3*n + 2];
        if (iss.fail()) 
    	    print_error(lineno, 0);

	if(strname.length() > (MAX_CHAR_LENGTH - 1)){
	    std::stringstream error;
	    error << "Atom Type " << strname
		<< " exceeds maximum character length of "
		<< MAX_CHAR_LENGTH-1 << "\n";
	    std::runtime_error(error.str());
	}

    	if((strname.find("OW")) != std::string::npos){
	    frame.mass[n] = 15.994900;
	    frame.atomic_num[n] = 8;
	} else if((strname.find("HW")) != std::string::npos){
	    frame.mass[n] = 1.007900;
	    frame.atomic_num[n] = 1;
	}else{
	    frame.mass[n] = -1;
	    frame.atomic_num[n] = 0;
	}

	std::strcpy(frame.name + MAX_CHAR_LENGTH*n, strname.c_str());

	frame.dipmol[n] = 0.0;
	frame.dipind[n] = 0.0;
    }
}

//----------------------------------------------------------------------------//

void collect_multipoles(std::istream& is_mol, std::istream& is_ind, 
			opt::hist_frame& frame)
{
    assert(is_mol && !is_mol.eof());
    assert(is_ind && !is_ind.eof());
    assert(frame.natm != 0); // empty


    size_t atm;
    // parse first line of frame

    for(size_t i = 0; i < frame.natm/3; ++i){
	size_t i3 = 3*i;

	{//Molecular dipole
	    std::string line;
	    std::getline(is_mol, line); 
	    if (is_mol.eof())
		return;

	    std::istringstream iss(line);

	    iss >> atm >> frame.dipmol[i3] >> frame.dipmol[i3+1]
		>> frame.dipmol[i3+2];

	    if (iss.fail()) 
		print_error(12345, 0);

	    if(atm != i+1){
		std::cerr << "Counting is off, atm indices != in dipmol"
		          << std::endl;
	    }
	}

	{//Induced dipole
	    std::string line;
	    std::getline(is_ind, line); 
	    if (is_ind.eof())
		return;

	    std::istringstream iss(line);

	    iss >> atm >> frame.dipind[i3] >> frame.dipind[i3+1]
		>> frame.dipind[i3+2];

	    if (iss.fail()) 
		print_error(12345, 0);

	    if(atm != i+1){
		std::cerr << "Counting is off, atm indices != in dipind"
		          << std::endl;
	    }
	}
    }
}

//----------------------------------------------------------------------------//

void collect_ind_multipoles(std::istream& is_ind, 
			    opt::hist_frame& frame)
{
    assert(is_ind && !is_ind.eof());
    assert(frame.natm != 0); // empty


    size_t atm;
    // parse first line of frame

    for(size_t i = 0; i < frame.natm/3; ++i){
	size_t i3 = 3*i;

	{//Induced dipole
	    std::string line;
	    std::getline(is_ind, line); 
	    if (is_ind.eof())
		return;

	    std::istringstream iss(line);

	    iss >> atm >> frame.dipind[i3] >> frame.dipind[i3+1]
		>> frame.dipind[i3+2];

	    if (iss.fail()) 
		print_error(12345, 0);

	    if(atm != i+1){
		std::cerr << "Counting is off, atm indices != in dipind"
		          << std::endl;
	    }
	}
    }
}
//----------------------------------------------------------------------------//

size_t read_hist
    (const char* filename, std::vector<hist_frame>& frames, bool multipole)
{
    assert(filename);

    std::ifstream ifs(filename);
    std::string line;
    std::getline(ifs, line);
    std::getline(ifs, line);

    if (!ifs)
        throw std::runtime_error("could not open XYZ file for reading");

    std::ifstream mmol_ifs;
    std::ifstream mind_ifs;

    if(multipole){
	mmol_ifs.open("DIPMOL.01", std::ifstream::in);
	if (!mmol_ifs)
	    throw std::runtime_error("could not open DIPMOL.01 for reading");
	mind_ifs.open("DIPIND.01", std::ifstream::in);
	if (!mind_ifs)
	    throw std::runtime_error("could not open DIPIND.01 for reading");
    }

    size_t lineno(0);
    size_t frameno(0);

    while (!ifs.eof()) {
        hist_frame frame;

	read_hist_frame(lineno, ifs, frame);

        if (frame.natm > 0) {
	    if(multipole)
		collect_multipoles(mmol_ifs, mind_ifs, frame);
            frames.push_back(frame);
            ++frameno;
        }
    }

    return frameno;
}

//----------------------------------------------------------------------------//

size_t read_cmd
    (std::vector<hist_frame>& frames, bool multipole)
{
    std::ifstream ifs("POSITION_CMD");

    if (!ifs)
        throw std::runtime_error("could not open POSITION_CMD for reading");

    std::ifstream mmol_ifs;
    std::ifstream mind_ifs;

    if(multipole){
	mmol_ifs.open("DIPMOL_CMD", std::ifstream::in);
	if (!mmol_ifs)
	    throw std::runtime_error("could not open DIPMOL_CMD for reading");
	mind_ifs.open("DIPIND_CMD", std::ifstream::in);
	if (!mind_ifs)
	    throw std::runtime_error("could not open DIPIND_CMD for reading");
    }

    size_t lineno(0);
    size_t frameno(0);

    while (!ifs.eof()) {
        hist_frame frame;

	read_cmd_frame(lineno, ifs, frame);

        if (frame.natm > 0) {

	    if(multipole)
		collect_multipoles(mmol_ifs, mind_ifs, frame);

            frames.push_back(frame);
            ++frameno;
        }
    }

    return frameno;
}

//----------------------------------------------------------------------------//

//size_t read_cmd
//    (const char* c_dirname, std::vector<hist_frame>& frames, bool multipole)
//{
//
//    std::string dirname(c_dirname);
//
//    std::string pos_name = dirname+"/POSITION_CMD";
//    std::string dipmol_name = dirname+"/DIPMOL_CMD";
//    std::string dipind_name = dirname+"/DIPIND_CMD";
//
//    std::ifstream ifs(pos_name.c_str());
//    std::ifstream mmol_ifs(dipmol_name.c_str());
//    std::ifstream mind_ifs(dipind_name.c_str());
size_t read_cmd
    (const char* c_pos_file, std::vector<hist_frame>& frames, bool multipole)
{

    std::string    pos_file = c_pos_file;
    std::ifstream ifs(c_pos_file);
    std::ifstream mmol_ifs;
    std::ifstream mind_ifs;

    if(multipole){
	std::string dipmol_file = c_pos_file;
	std::string dipind_file = c_pos_file;

	if(!replace(dipmol_file, pos_name, dipmol_name))
	    std::cerr << "Could not rename " << pos_file << " to DIPMOL_CMD\n"
		      << " >>> " << dipmol_file  << std::endl;

	if(!replace(dipind_file, pos_name, dipind_name))
	    std::cerr << "Could not rename " << pos_file << " to DIPMOL_CMD\n"
		      << " >>> " << dipind_file  << std::endl;

	std::cerr << "Reading these files" << std::endl
	          <<    pos_file << std::endl
	          << dipmol_file << std::endl
	          << dipind_file << std::endl;

	mmol_ifs.open(dipmol_file.c_str());
	mind_ifs.open(dipind_file.c_str());
    }

    if (!ifs){
	std::ostringstream oss;
	oss << "could not open " << c_pos_file << " for reading\n";
	throw std::runtime_error(oss.str());
    }

    size_t lineno(0);
    size_t frameno(0);

    while (!ifs.eof()) {
        hist_frame frame;

	read_cmd_frame(lineno, ifs, frame);

        if (frame.natm > 0){ 

	    if(multipole)
		collect_multipoles(mmol_ifs, mind_ifs, frame);

	    // if more than one frame, make sure not a duplicate
	    if(frames.size() > 1){
		if(frame.time != frames[frames.size()-1].time) {
		    frames.push_back(frame);
		    ++frameno;
		}
	    }else{
		frames.push_back(frame);
		++frameno;
	    }
        }
    }

    return frameno;
}

//----------------------------------------------------------------------------//

void print_hist(std::ostream& os, const char* atom, const double* x)
{
    std::ios::fmtflags saved_flags = os.flags();

    os << std::setw(5) << std::left << atom
       << std::setprecision(9) << std::scientific
       << std::setw(18) << std::right << x[0]
       << std::setw(18) << std::right << x[1]
       << std::setw(18) << std::right << x[2]
       << std::endl;

    os.flags(saved_flags);
}

////////////////////////////////////////////////////////////////////////////////

} // namespace opt

////////////////////////////////////////////////////////////////////////////////

