#include <cassert>

#include <sstream>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <algorithm>
#include <stdexcept>

#include "read-scan.h"

////////////////////////////////////////////////////////////////////////////////

namespace tools {

////////////////////////////////////////////////////////////////////////////////

void read_single_scan(size_t& lineno, std::istream& is, 
		      tools::scan_frame& frame, bool check_curve)
{
    assert(is && !is.eof());
    assert(frame.npoints == 0); // empty

    std::string line;
    std::getline(is, line);

    if (is.eof())
        return;
    ++lineno;

    // number of atoms
    size_t npoints;
    double time;
    std::istringstream iss(line);
    iss >> npoints >> time;
    if (iss.fail()) {
	std::ostringstream oss;
    	oss << "unexpected text at line " << lineno << " of the scan file";
    	throw std::runtime_error(oss.str());
    }

    frame.npoints = npoints;
    frame.time    = time;
    frame.rOX     = new double[npoints];
    frame.E       = new double[npoints];

    for (size_t n = 0; n < npoints; ++n) {
        if (is.eof()) {
            std::ostringstream oss;
            oss << "unexpected EOF at line " << lineno << " of the scan file";
            throw std::runtime_error(oss.str());
        }

    	std::getline(is, line);
        ++lineno;

        double rOX, E;

        iss.clear();
        iss.str(line);
	iss >> rOX >> E;
        if (iss.fail()) {
            std::ostringstream oss;
            oss << "unexpected text at line " << lineno << " of the scan file";
            throw std::runtime_error(oss.str());
        }

        frame.rOX[n] = rOX;
        frame.E[n]   = E;
    }

    if(check_curve)
	check_for_multiple_minima(frame);

}

////////////////////////////////////////////////////////////////////////////////

size_t read_scans
    (const char* filename, std::vector<scan_frame>& frames, bool check_curve)
{
    assert(filename);

    std::ifstream ifs(filename);
    if (!ifs)
        throw std::runtime_error("could not open scan file for reading");

    size_t lineno(0);
    size_t frameno(0);

    while (!ifs.eof()) {
        scan_frame frame;

	read_single_scan(lineno, ifs, frame, check_curve);

        if (frame.npoints > 0) {
            frames.push_back(frame);
            ++frameno;
        }
    }

    return frameno;
}

////////////////////////////////////////////////////////////////////////////////

bool check_for_multiple_minima(scan_frame& frame)
{
    std::vector<int> faulty_points;

    double min(100000);
    size_t ind_min(0);
    for(size_t i = 0; i < frame.npoints; ++i){
	if(frame.E[i] < min){
	    min = frame.E[i];
	    ind_min = i;
	}
    }

    //ensure that each point goes up to the left
    for(int i = ind_min - 1; i >= 0; --i){
	if(frame.E[i] < frame.E[i+1])
	    faulty_points.push_back(i);
    }
    
    //ensure that each point goes up to the right
    for(int i = ind_min + 1; i < frame.npoints; ++i){
	if(frame.E[i] < frame.E[i-1])
	    faulty_points.push_back(i);
    }

    if(faulty_points.size() != 0){
	std::cerr << "Bad scan!!" << std::endl;
	for(size_t i = 0; i < frame.npoints; ++i)
	    std::cerr << std::setw(15) << frame.rOX[i] 
	  	      << std::setw(15) << frame.E[i] << std::endl;

	return true;
    }

    return false;
}

////////////////////////////////////////////////////////////////////////////////

} // namespace tools

////////////////////////////////////////////////////////////////////////////////
