#include <cassert>

#include <sstream>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <algorithm>
#include <stdexcept>

#include "read-xyz.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

void read_frame(size_t& lineno, std::istream& is, tools::xyz_frame& frame,
		bool do_dip, bool do_pol)
{
    assert(is && !is.eof());
    assert(frame.nw == 0); // empty

    std::string line;
    std::getline(is, line);

    if (is.eof())
        return;

    // number of atoms
    size_t natom;
    std::istringstream iss(line);
    iss >> natom;

    if (natom%3 != 0 || natom == 0)
        throw std::runtime_error("unexpected number of atoms in the XYZ file");

    frame.nw = natom/3;
    frame.xyz = new double[9*frame.nw];
    frame.dip = new double[3];
    frame.pol = new double[6];

    std::getline(is, line); // Extract energy, if provided
    ++lineno;

    if(do_dip){
	if (line.length() != 0){
	    double dip[3];

	    iss.clear();
	    iss.str(line);
	    iss >> dip[0] >> dip[1] >> dip[2];
	    if (iss.fail()) {
		std::ostringstream oss;
		oss << "unexpected text at line " << lineno 
		    << " of the XYZ file:" << std::endl << iss << std::endl;
		throw std::runtime_error(oss.str());
	    }

	    frame.dip[0] = dip[0];
	    frame.dip[1] = dip[1];
	    frame.dip[2] = dip[2];
	    std::fill(frame.pol, frame.pol + 6, 0.0);
	} else {
	    std::fill(frame.dip, frame.dip + 3, 0.0);
	    std::fill(frame.pol, frame.pol + 6, 0.0);
	}
    }else if(do_pol){
	if (line.length() != 0){
	    double pol[6];

	    iss.clear();
	    iss.str(line);
	    iss >> pol[0] >> pol[1] >> pol[2] >> pol[3] >> pol[4] >> pol[5];
	    if (iss.fail()) {
		std::ostringstream oss;
		oss << "unexpected text at line " << lineno 
		    << " of the XYZ file:" << std::endl << iss << std::endl;
		throw std::runtime_error(oss.str());
	    }

	    frame.pol[0] = pol[0];
	    frame.pol[1] = pol[1];
	    frame.pol[2] = pol[2];
	    frame.pol[3] = pol[3];
	    frame.pol[4] = pol[4];
	    frame.pol[5] = pol[5];
	    std::fill(frame.dip, frame.dip + 3, 0.0);
	} else {
	    std::fill(frame.dip, frame.dip + 3, 0.0);
	    std::fill(frame.pol, frame.pol + 6, 0.0);
	}
    }else{
	std::fill(frame.dip, frame.dip + 3, 0.0);
    	std::fill(frame.pol, frame.pol + 6, 0.0);
    }

    ++lineno;

    for (size_t n = 0; n < natom; ++n) {
        if (is.eof()) {
            std::ostringstream oss;
            oss << "unexpected EOF at line " << lineno << " of the XYZ file";
            throw std::runtime_error(oss.str());
        }

        std::getline(is, line);
        ++lineno;

        std::string name;
        double x, y, z;

        iss.clear();
        iss.str(line);
        iss >> name >> x >> y >> z;
        if (iss.fail()) {
            std::ostringstream oss;
            oss << "unexpected text at line " << lineno << " of the XYZ file";
            throw std::runtime_error(oss.str());
        }

        frame.xyz[3*n + 0] = x;
        frame.xyz[3*n + 1] = y;
        frame.xyz[3*n + 2] = z;
    }
}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

namespace tools {

////////////////////////////////////////////////////////////////////////////////

size_t read_xyz
    (const char* filename, std::vector<xyz_frame>& frames, 
     bool do_dip, bool do_pol)
{
    assert(filename);

    std::ifstream ifs(filename);
    if (!ifs)
        throw std::runtime_error("could not open XYZ file for reading");

    size_t lineno(0);
    size_t frameno(0);

    if(do_dip == true && do_pol == true)
	throw std::runtime_error("Cannot read dip and pol from xyz file at same time");

    while (!ifs.eof()) {
        xyz_frame frame;

	read_frame(lineno, ifs, frame, do_dip, do_pol);

        if (frame.nw > 0) {
            frames.push_back(frame);
            ++frameno;
        }
    }

    return frameno;
}

////////////////////////////////////////////////////////////////////////////////

void print_xyz(std::ostream& os, const size_t& nw, const double* x,
	       std::string comment)
{
    std::ios::fmtflags saved_flags = os.flags();

    if(!comment.empty())
	os << 3*nw << std::endl << comment << std::endl;
    else
	os << 3*nw << std::endl << std::endl;
    for(size_t n = 0; n < nw; ++n){
	const size_t natm = 3*n;
	for(size_t i = 0; i < 3; ++i){
	    const size_t natm3 = (natm + i)*3;
	    char name = ' ';
	    if(i == 0)
		name = 'O';
	    else
		name = 'H';
	
	    os << std::setw(5) << std::left << name
	       << std::setprecision(5)
       	       << std::setw(18) << std::right << x[natm3 + 0]
	       << std::setw(18) << std::right << x[natm3 + 1]
	       << std::setw(18) << std::right << x[natm3 + 2]
	       << std::endl;
	}
    }

    os.flags(saved_flags);
}

////////////////////////////////////////////////////////////////////////////////

void swap_xyz(double* a, double* b)
{
    if (a != b) {
     	std::swap(a[0], b[0]);
	std::swap(a[1], b[1]);
	std::swap(a[2], b[2]);
    }
}

////////////////////////////////////////////////////////////////////////////////

} // namespace tools

////////////////////////////////////////////////////////////////////////////////
