#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cstdlib>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <string>
#include <iostream>

#include <cmath>
#include <cassert>

#include <limits>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <limits>

namespace {

    const double PI = M_PI;

inline double distance(const double* x1, const double* x2)
{
    const double dx = x1[0] - x2[0];
    const double dy = x1[1] - x2[1];
    const double dz = x1[2] - x2[2];

    return std::sqrt(dx*dx + dy*dy + dz*dz);
}

inline double vect_mag(const double* x1, const double* x2, double* x3)
{
    x3[0] = x2[0] - x1[0];
    x3[1] = x2[1] - x1[1];
    x3[2] = x2[2] - x1[2];

    const double tmp = std::sqrt(x3[0]*x3[0] + x3[1]*x3[1] + x3[2]*x3[2]);

    x3[0] /= tmp;
    x3[1] /= tmp;
    x3[2] /= tmp;

    return tmp;
}

inline void vect(const double* x1, const double* x2, double* x3)
{
    x3[0] = x2[0] - x1[0];
    x3[1] = x2[1] - x1[1];
    x3[2] = x2[2] - x1[2];

    const double tmp = std::sqrt(x3[0]*x3[0] + x3[1]*x3[1] + x3[2]*x3[2]);

    x3[0] /= tmp;
    x3[1] /= tmp;
    x3[2] /= tmp;
}

inline void cross(const double *y, const double*z, double *x){
    double tmp;
    x[0] = z[2] * y[1] - z[1] * y[2];
    x[1] = z[0] * y[2] - z[2] * y[0];
    x[2] = z[1] * y[0] - z[0] * y[1];
    tmp = std::sqrt(x[0]*x[0] + x[1]*x[1] + x[2]*x[2]);
    
    for(size_t i = 0; i < 3; ++i){
      x[i] /= tmp;
    }    
}

} // namespace

namespace ccpol {

////////////////////////////////////////////////////////////////////////////////

//We need to convert from Cartesian to Euler:
//The z-y-z Euler angles of each monomer (alpha,beta,gamma) are defined in the
//coordinate system with the origin in the center of mass (COM) of the
//molecule such that the z axis points from the COM of monomer 1 to the
//COM of monomer 2. Monomer orientation with all three Euler angles
//equal to zero corresponds to the molecular plane coinciding with the
//x-z plane and the z axis pointing towards the O atom.  The rotation of
//the COM-COM axis is such that alpha1=0.
  
//Input should be: 
//COM-COM distance, beta1,gamma1, alpha2,beta2,gamma2  (bohr,degree)
  
void cart_to_euler
    (const double* crd, double* euler)
{
    //Euler: [r, betaA, gammaA, alphaB, betaB, gammaB]

    //This works only for dimers

    //Determine Centers of Mass
    const double O_mass = 15.9949;
    const double H_mass = 1.0079;
    const double mass[3] = { O_mass, H_mass, H_mass };
    const double tot_mass = O_mass + 2*H_mass;
    double comA[3] = {0}, comB[3] = {0};
    
    for(size_t i = 0; i < 3; ++i){
	comA[0] += mass[i]*crd[i*3]      / tot_mass;
	comA[1] += mass[i]*crd[i*3 + 1 ] / tot_mass;
	comA[2] += mass[i]*crd[i*3 + 2 ] / tot_mass;
	comB[0] += mass[i]*crd[i*3 + 9 ] / tot_mass;
	comB[1] += mass[i]*crd[i*3 + 10] / tot_mass;
	comB[2] += mass[i]*crd[i*3 + 11] / tot_mass;
    }

    //Dimer reference frame, pointing from A->B: zBA
    //Molecule A ref. frame, pointing from comA->Oa: zA
    //Molecule B ref. frame, pointing from comB->Ob: zB

    double zBA[3] = {0}, yBA[3] = {0}, xBA[3] = {0};
    double  zA[3] = {0},  yA[3] = {0},  xA[3] = {0};
    double  zB[3] = {0},  yB[3] = {0},  xB[3] = {0};

    euler[0] = vect_mag(comA, comB, zBA);
    //R(COM-COM) should be in Bohr!
    vect(comA, crd    , zA);
    vect(comB, crd + 9, zB);

    cross(zBA,  zA, yBA);
    cross(yBA, zBA, xBA);

    //molecule A: reference frame
    for(size_t i = 0; i < 3; ++i)
	yA[i] = yBA[i];

    cross(zA, yA, xA);

    //molecule B
    double hoB[3] = {0};
    vect(crd + 9, crd + 12, hoB);

    cross(hoB, zB, yB);
    cross( zB, yB, xB);

    gsl_matrix*    BA = gsl_matrix_alloc(3,3);
    gsl_matrix* invBA = gsl_matrix_alloc(3,3);
    gsl_matrix*  Amat = gsl_matrix_alloc(3,3);
    gsl_matrix*  Arot = gsl_matrix_alloc(3,3);
    gsl_matrix*  Bmat = gsl_matrix_alloc(3,3);
    gsl_matrix*  Brot = gsl_matrix_alloc(3,3);

    for(size_t i = 0; i < 3; ++i){
	gsl_matrix_set(BA,   i, 0, xBA[i]);
	gsl_matrix_set(BA,   i, 1, yBA[i]);
	gsl_matrix_set(BA,   i, 2, zBA[i]);
	gsl_matrix_set(Amat, i, 0, xA[i]);
	gsl_matrix_set(Amat, i, 1, yA[i]);
	gsl_matrix_set(Amat, i, 2, zA[i]);
	gsl_matrix_set(Bmat, i, 0, xB[i]);
	gsl_matrix_set(Bmat, i, 1, yB[i]);
	gsl_matrix_set(Bmat, i, 2, zB[i]);
    }

    gsl_permutation *p = gsl_permutation_alloc (3);
    int s;
    gsl_linalg_LU_decomp(BA, p, &s);
    gsl_linalg_LU_invert(BA, p, invBA);

    gsl_matrix_set_zero(Arot);
    gsl_matrix_set_zero(Brot);

    //Compute Arot = Amat * invBA
    gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, Amat, invBA, 0.0, Arot);
    gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, Bmat, invBA, 0.0, Brot);
    
    //Euler: [r, betaA, gammaA, alphaB, betaB, gammaB]

    // betaA
    euler[1] = std::acos( gsl_matrix_get(Arot,2,2));
    //alphaA
    double alphaA = std::acos( gsl_matrix_get(Arot,0,2) / std::sin( euler[1] ));
    //gammaA
    euler[2] = -std::acos( gsl_matrix_get(Arot,2,0) / std::sin( euler[1] ));

    // betaB
    euler[4] = std::acos( gsl_matrix_get(Brot,2,2) );
    //alphaB
    euler[3] = std::acos( gsl_matrix_get(Brot,0,2) / std::sin( euler[4] ));
    //gammaB
    euler[5] = -std::acos( gsl_matrix_get(Brot,2,0) / std::sin( euler[4] ));

    for(size_t i = 1; i < 6; ++i)
	euler[i] *= 180/PI;

    std::cout.precision(8);
    std::cout.setf(std::ios_base::showpoint);

    std::cout << "R(COM-COM) = " << euler[0] << std::endl
	      << "(Remember, R(COM-COM) must be in Bohr. Fix this!)" << std::endl
	      << "alphaA = " << alphaA << std::endl
	      << "betaA  = " << euler[1] << std::endl
	      << "gammaA = " << euler[2] << std::endl
	      << "alphaB = " << euler[3] << std::endl
	      << "betaB  = " << euler[4] << std::endl
	      << "gammaB = " << euler[5] << std::endl;



}

////////////////////////////////////////////////////////////////////////////////

} // namespace ccpol 

int main(void){
    double coord[18] ={0.109839948677069, 0.0000000000E+000, -6.077698915E-002,
  	              -0.167834869455959, 0.0000000000E+000,  1.75421148907991,
	              -1.57540492249411 , 0.0000000000E+000, -0.78963651258448,
     		      -0.102982479243547, 1.2611716357E-017,  5.43547774989713,
	    	       0.817203384803212, -1.45365196228170,  6.07691397344509,
		       0.817203384803212,  1.45365196228170,  6.07691397344509};

    double euler[6];

    ccpol::cart_to_euler(coord, euler);
    return 0;
}
