#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cstdlib>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <string>
#include <iostream>

#include <cmath>
#include <cassert>

#include <limits>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>

#include "read-xyz.h"
#include "orient.h"
#include "mat_tools.h"

#include "dsyevh3.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

//----------------------------------------------------------------------------//

const double mass_O = 15.9949;
const double mass_H =  1.0079;
const double smass_O = std::sqrt(mass_O); 
const double smass_H = std::sqrt(mass_H);

//----------------------------------------------------------------------------//

void pack_rotation_vector(D, xyz, QrA, QcA, QrB, QcB, smass){

    const double A = QrA[0]*xyz[0] + QrA[1]*xyz[1] + QrA[2]*xyz[2];
	const double B = QrB[0]*xyz[0] + QrB[1]*xyz[1] + QrB[2]*xyz[2];

    for (size_t j = 0; j < 3; ++j)
        D[j] = (A*QcA[j] + B*QcB[j])/smass;

}

//----------------------------------------------------------------------------//

} // namespace

////////////////////////////////////////////////////////////////////////////////

extern "C" {

    int dsyevh3(double *A, double *Q, double w[3]);

} // extern "C"

////////////////////////////////////////////////////////////////////////////////

int main(void){

    const size_t nw = 1;
    const size_t natm = 3*nw;
    const size_t natm3 = 9*nw;

//    double coord[natm3] = {-9.7848165530, -3.8736896807, 1.6057618942,
//	                -9.0708530959, -4.6086177117, 1.7032467123,
//			-10.681149484, -4.1697317829, 2.1411977815,
//			-8.3945381511, -1.6705736130, 2.8500799807,
//			-8.6914729338, -0.7170372729, 2.8172785149,
//			-8.8686659379, -2.3038775660, 2.3591184765};

//    double coord[natm3] = {-1.32843279,    0.51270096,    0.10611897,
//                        -1.15731791,    1.48945617,    0.23522728,
//                        -2.32282529,    0.43484004,    0.03455617,
//                         1.36690000,   -0.52095000,    0.14919000,
//                         1.50551464,   -1.50509511,    0.03853567,
//			 0.37629315,   -0.39228164,    0.10290254};

    double coord[natm3] = {0.0,  0.0, 0.0,
                          0.7,  0.7, 0.1,
			  0.7, -0.7, -0.1};

    double Ilab[9] = {0.0};
    double* xyz = new double[natm3];

    std::ofstream outfile;
    outfile.open("out.xyz");
    outfile.precision(8);
    outfile.setf(std::ios_base::showpoint);

    std::cout.precision(8);
    std::cout.setf(std::ios_base::showpoint);
//    std::cout << "Input coordinates\n";
    tools::print_xyz(outfile, nw, coord);

    tools::center(nw, coord, xyz);

//    std::cout << "\nCenter of Mass coordinates\n";
    tools::print_xyz(outfile, nw, xyz);

    tools::cpt_Ilab(nw, xyz, Ilab);

    std::cout << "\nTensor of inertia: I\n";
    tools::print_m_by_n(3, 3, Ilab);

    double Q[9], w[3];

    dsyevh3(Ilab, Q, w);

    std::cout << "\nEigenvalues: I_{\\lambda}\n";
    for (size_t i = 0; i < 3; ++i)
	std::cout << w[i] << std::endl;

    std::cout << "\nEigenvectors: Q\n";
    tools::print_m_by_n(3, 3, Q);

    double Qt[9];
    tools::transpose(3, 3, Q, Qt);

    std::cout << "\n Q^t I Q == diagonalized?\n";

    double tmp[9];
    tools::mat_mult(3, Qt, 3, Ilab, 3, tmp);
    tools::mat_mult(3, tmp, 3, Q, 3, Qt);

    tools::print_m_by_n(3, 3, Qt);

    // Now try to project out rotations
    double rotated[natm3];
    double D4[natm3];
    double D5[natm3];
    double D6[natm3];

    // reset Qt
    tools::transpose(3, 3, Q, Qt);

    for (size_t i = 0; i < natm; ++i){
	const size_t i3 = 3*i;
	const double smass = ( ( i%3 == 0 ) ? smass_O : smass_H );

	pack_rotation_vector(D4 + i3, xyz + i3, Q + 3, Qt + 6,
			                        Q + 6, Qt + 3, smass); 
			           
	pack_rotation_vector(D5 + i3, xyz + i3, Q + 6, Qt    ,
			                        Q    , Qt + 6, smass); 

	pack_rotation_vector(D6 + i3, xyz + i3, Q    , Qt + 3,
			                        Q + 3, Qt    , smass); 
    }

    double D4t[natm3];
    double D5t[natm3];
    double D6t[natm3];

    for (size_t i = 0; i < natm; ++i)
    	tools::rotate(Qt, xyz + 3*i, rotated + i*3);

    tools::print_xyz(outfile, nw, rotated);


    outfile.close();

    //    double euler[6];
    //
    //    ccpol::cart_to_euler(coord, euler);
    delete[] xyz;

    return 0;

}

////////////////////////////////////////////////////////////////////////////////
