#include <cmath>
#include <cassert>
#include <algorithm>
#include <cstdlib>
#include <iostream>

#include "ps.h"
#include "smear.h"
#include "constants.h"
#include "electrostatics.h"

#include "ttm4f.h"

namespace {

const double CHARGECON = constants::CHARGECON;

////////////////////////////////////////////////////////////////////////////////

inline void compute_M_site_crd
    (const double& gamma,
     const double O[3], const double H1[3], const double H2[3],
     double M[3])
{
    const double gamma1 = 1.0 - gamma;
    const double gamma2 = 0.5*gamma;

    for (size_t i = 0; i < 3; ++i)
        M[i] = gamma1*O[i] + gamma2*(H1[i] + H2[i]);
}

////////////////////////////////////////////////////////////////////////////////


inline void com_to_origin( const double* crd, double* com_crd)
{
    const double O_mass = 15.9949;
    const double H_mass = 1.0079;
    const double mass[3] = { O_mass, H_mass, H_mass };
    const double tot_mass = O_mass + 2*H_mass;
    double com[3] = {0};

    for(size_t i = 0; i < 3; ++i){
	com[0] += mass[i]*crd[i*3]     / tot_mass;
	com[1] += mass[i]*crd[i*3 + 1] / tot_mass;
	com[2] += mass[i]*crd[i*3 + 2] / tot_mass;
    }

    for(size_t i = 0; i < 4; ++i){
	com_crd[i*3] = crd[i*3] - com[0];
	com_crd[i*3 + 1] = crd[i*3 + 1] - com[1];
	com_crd[i*3 + 2] = crd[i*3 + 2] - com[2];
    }
}

////////////////////////////////////////////////////////////////////////////////


} // namespace

namespace ttm {

// M-site positioning (TTM2.1-F)
const double ttm4f::gammaM = 0.426706882;

ttm4f::ttm4f()
{
    set_vdw(true);
    set_calc_mon(true);
    set_new_damping(false);

    // polarizabilities
    polarO = 1.310;
    polarH = 0.294;
    polarM = 0.0;

    polfacO = polarO;
    polfacH = polarH;
    polfacM = polarO;

    m_nmol = 0;
}

ttm4f::~ttm4f()
{
    allocate(0);
}

void ttm4f::set_vdw(bool on) {
    if (on) {
        A6  = -0.126503e+5;
        A8  =  0.526347e+6;
        A10 = -0.964270e+7;
        A12 =  0.877792e+8;
        A14 = -0.367476e+9;
        A16 =  0.572395e+9;
        calc_vdw = true;
    } else {
        A6 = A8 = A10 = A12 = A14 = A16 = 0.0;
        calc_vdw = false;
    }
}

void ttm4f::set_calc_mon(bool on) {
    calc_mon = on;
}

void ttm4f::set_new_damping(bool on, double* a){
    if (on) {
	smear_ttm4 smr;

	if(a[0] == 0)
    	    aCC = smr.aCC();
	else
	    aCC = a[0];

	if(a[1] == 0)
	    aCD = smr.aCD();
	else
	    aCD = a[1];

	if(a[2] == 0)
	    aDD_inter = smr.aDD_inter();
	else
	    aDD_inter = a[2];

	if(a[3] == 0)
	    aDD_intra = smr.aDD_intra();
	else
	    aDD_intra = a[3];

      	change_damping = on;

    } else {
	smear_ttm4 smr;

	aCC = smr.aCC();
	aCD = smr.aCD();
	aDD_inter = smr.aDD_inter();
	aDD_intra = smr.aDD_intra();
    }
}


void ttm4f::allocate(size_t nmol)
{
    if (nmol == m_nmol)
        return;

    if (m_nmol > 0)
        delete[] m_memory;

    m_nmol = nmol;

    if (m_nmol == 0)
        return;

    m_memory = new double[4*m_nmol // charge
                        + 4*m_nmol  // polfac
                        + 4*m_nmol   // polar
                        + 3*4*m_nmol  // xyz
                        + 3*3*3*m_nmol // grdq
                        + 3*4*m_nmol  // dE
                        + 3*4*m_nmol  // grad_add
                        + 3*4*m_nmol];  // grad_nonadd

    // setup exclusions
    m_excluded.clear();
    for (size_t i = 0; i < m_nmol; ++i) {
        const size_t i4 = 4*i;
        m_excluded.insert(std::make_pair(i4 + 0, i4 + 1)); // O - H1
        m_excluded.insert(std::make_pair(i4 + 0, i4 + 2)); // O - H2
        m_excluded.insert(std::make_pair(i4 + 0, i4 + 3)); // O - M
        m_excluded.insert(std::make_pair(i4 + 1, i4 + 2)); // H1 - H2
        m_excluded.insert(std::make_pair(i4 + 1, i4 + 3)); // H1 - M
        m_excluded.insert(std::make_pair(i4 + 2, i4 + 3)); // H2 - M
    }
}

const double* ttm4f::grad_add() const
{
        return m_memory + m_nmol*(3*4 + 12 + 27 + 12);//see below
}

const double* ttm4f::grad_nonadd() const
{
        return m_memory + m_nmol*(3*4 + 12 + 27 + 12 + 9);//see below
}


////////////////////////////////////////////////////////////////////////////////

double ttm4f::operator()(size_t nmol, const double* crd, double* grad)
{
    assert(nmol && crd);

    allocate(nmol);

    // setup pointers

    double* charge = m_memory;
    double* polfac = charge + 4*nmol;
    double* polar = polfac + 4*nmol;
    double* xyz = polar + 4*nmol;
    double* grdq = xyz + 3*4*nmol;
    double* dE = grdq + 3*3*3*nmol;
    double* grad_add = dE + 3*4*nmol;
    double* grad_nonadd = grad_add + 3*3*nmol;

    // populate polfac and polar
    for (size_t n = 0; n < nmol; ++n) {
        const size_t n4 = 4*n;

        polfac[n4 + 0] = polfacO;
        polfac[n4 + 1] = polfacH;
        polfac[n4 + 2] = polfacH;
        polfac[n4 + 3] = polfacM;

        polar[n4 + 0] = polarO;
        polar[n4 + 1] = polarH;
        polar[n4 + 2] = polarH;
        polar[n4 + 3] = polarM;
    }

    //Initialize some stuff
    std::fill(grad_add, grad_add + 9*nmol, 0.0);
    std::fill(grad_nonadd, grad_nonadd + 9*nmol, 0.0);

    // calculate coordinates of the M-sites;
    // pack the coordinates in xyz as O H H M O H H M ...

    for (size_t i = 0; i < nmol; ++i) {
        // O H H
        std::copy(crd + 3*3*i, crd + 3*3*i + 9, xyz + 4*3*i);

        // and M
        compute_M_site_crd(gammaM,
            crd + 3*3*i, crd + 3*(3*i + 1), crd + 3*(3*i + 2),
                xyz + 4*3*i + 9);
    }

    // calculate the INTRA-molecular energy and Dipole Moment
    // Surface (Partridge-Schwenke)

    double Eint = 0.0;
    for (size_t n = 0; n < nmol; ++n) {
        double q3[3], dq3[27];
	const size_t n9 = 9*n;

        if (grad) {
            if(calc_mon){
                Eint += ps::pot_nasa(crd + n9, grad + n9);
		for(size_t a = 0; a < 9; ++a)
		    grad_add[n9 + a] = grad[n9 + a];
	    }else
                std::fill(grad + n9, grad + n9 + 9, 0.0);

            ps::dms_nasa(0.0, 0.0, 0.0,
                         crd + n9, q3, dq3, false);
        } else {
            if(calc_mon)
                Eint += ps::pot_nasa(crd + n9, 0);

            ps::dms_nasa(0.0, 0.0, 0.0,
                         crd + n9, q3, 0, false);
        }

        // TTM2.1-F assignment
        const double tmp = 0.5*gammaM/(1.0 - gammaM);

        charge[4*n + 0] = 0.0;                        // O
        charge[4*n + 1] = q3[1] + tmp*(q3[1] + q3[2]); // H1
        charge[4*n + 2] = q3[2] + tmp*(q3[1] + q3[2]); // H2
        charge[4*n + 3] = q3[0]/(1.0 - gammaM);       // M

        if (grad == 0)
            continue;

#define DQ3(i,j,k) dq3[k + 3*(j + 3*i)]
#define GRDQ(i,j,k) grdq[k + 3*(j + 3*(i + 3*n))]

        for (size_t k = 0; k < 3; ++k) {
            GRDQ(0, 0, k) = DQ3(0, 0, k) + tmp*(DQ3(0, 0, k) + DQ3(0, 1, k));
            GRDQ(1, 0, k) = DQ3(1, 0, k) + tmp*(DQ3(1, 0, k) + DQ3(1, 1, k));
            GRDQ(2, 0, k) = DQ3(2, 0, k) + tmp*(DQ3(2, 0, k) + DQ3(2, 1, k));

            GRDQ(0, 1, k) = DQ3(0, 1, k) + tmp*(DQ3(0, 1, k) + DQ3(0, 0, k));
            GRDQ(1, 1, k) = DQ3(1, 1, k) + tmp*(DQ3(1, 1, k) + DQ3(1, 0, k));
            GRDQ(2, 1, k) = DQ3(2, 1, k) + tmp*(DQ3(2, 1, k) + DQ3(2, 0, k));

            GRDQ(0, 2, k) = DQ3(0, 2, k) - 2*tmp*(DQ3(0, 0, k) + DQ3(0, 1, k));
            GRDQ(1, 2, k) = DQ3(1, 2, k) - 2*tmp*(DQ3(1, 0, k) + DQ3(1, 1, k));
            GRDQ(2, 2, k) = DQ3(2, 2, k) - 2*tmp*(DQ3(2, 0, k) + DQ3(2, 1, k));
        }
    }

    for (size_t n = 0; n < 4*nmol; ++n)
        charge[n] *= CHARGECON;

    if (grad) {
        for (size_t n = 0; n < 27*nmol; ++n)
            grdq[n] *= CHARGECON;
    }

    //---------------------------------------------------------------!
    // Calculate the vdW interactions for all atoms                  !
    //---------------------------------------------------------------!

    double Evdw(0);

    if (calc_vdw) {
        for (size_t i = 0; i < nmol - 1; ++i) {
            for (size_t j = i + 1; j < nmol; ++j) {

                // Calculate Oxygen-Oxygen interactions
                double dRijsq = 0.0, Rij[3];
                for (size_t k = 0; k < 3; ++k) {
                    Rij[k] = crd[3*3*i + k] - crd[3*3*j + k];
                    dRijsq += Rij[k]*Rij[k];
                }

                // Add Evdw van der Waal energy for O-O interaction
                const double x = 1/dRijsq;
                Evdw += x*x*x*(A6 + x*(A8 
                                  + x*(A10 + x*(A12 + x*(A14 + x*A16)))));

                if (grad) {
                    const double tmp =
                        x*x*x*x*(6*A6 + x*(8*A8 + x*(10*A10
                                 + x*(12*A12 + x*(14*A14 + x*16*A16)))));

                    for (size_t k = 0; k < 3; ++k) {
                        grad[3*3*i + k] -= tmp*Rij[k];
                        grad[3*3*j + k] += tmp*Rij[k];
                        grad_add[3*3*i + k] -= tmp*Rij[k];
                        grad_add[3*3*j + k] += tmp*Rij[k];
                    }
                } // grad
            }
        }
    }

    smear_ttm4 smr;

    m_int = Eint;
    m_vdw = Evdw;

//    std::cout << "The damping params: " << aCC << ' ' << aCD
//	      << ' ' << aDD_inter << ' ' << aDD_intra << std::endl;

    if (grad == 0)
	    return Eint + Evdw
		+ m_electrostatics(4*nmol, charge, polfac,
		      		   polar, xyz, m_excluded, smr, 0,
				   aCC, aCD, aDD_inter, aDD_intra);

    //-------------------------------------------------------------------------!
    // Calculate the remaining part of the derivatives                         !
    //-------------------------------------------------------------------------!

    const double Eelec =
        m_electrostatics(4*nmol, charge, polfac,
                         polar, xyz, m_excluded, smr, dE,
			 aCC, aCD, aDD_inter, aDD_intra);

    assert(m_electrostatics.dipoles_converged());
    const double* dE_add = m_electrostatics.dE_add();
    const double* dE_nonadd = m_electrostatics.dE_nonadd();

    // add electrostatic derivatives
    for (size_t n = 0; n < nmol; ++n) {
        // O H H
        for (size_t k = 0; k < 9; ++k){
            grad[9*n + k] += dE[12*n + k];
            grad_add[9*n + k] += dE_add[12*n + k];
            grad_nonadd[9*n + k] += dE_nonadd[12*n + k];
	}

        // redistribute the M-site derivatives
        for (size_t k = 0; k < 3; ++k) {
            grad[9*n + 0 + k] += (1.0 - gammaM)*dE[12*n + 9 + k]; // O
            grad[9*n + 3 + k] +=     0.5*gammaM*dE[12*n + 9 + k]; // H
            grad[9*n + 6 + k] +=     0.5*gammaM*dE[12*n + 9 + k]; // H
        }
	// redistribute the additive M-site derivatives
        for (size_t k = 0; k < 3; ++k) {
            grad_add[9*n + 0 + k] += (1.0 - gammaM)*dE_add[12*n + 9 + k];// O
            grad_add[9*n + 3 + k] +=     0.5*gammaM*dE_add[12*n + 9 + k];// H
            grad_add[9*n + 6 + k] +=     0.5*gammaM*dE_add[12*n + 9 + k];// H
        }
	// redistribute the non-additive M-site derivatives
        for (size_t k = 0; k < 3; ++k) {
            grad_nonadd[9*n+0+k]+=(1.0 - gammaM)*dE_nonadd[12*n + 9 + k];// O
            grad_nonadd[9*n+3+k]+=    0.5*gammaM*dE_nonadd[12*n + 9 + k];// H
            grad_nonadd[9*n+6+k]+=    0.5*gammaM*dE_nonadd[12*n + 9 + k];// H
        }
    }

    const double* phi = m_electrostatics.phi();
    const double* phi_add = m_electrostatics.phi_add();
    const double* phi_nonadd = m_electrostatics.phi_nonadd();

    // derivatives from the adjustable charges of the NASA PES
    for (size_t n = 0; n < nmol; ++n) {
        const size_t io  = 9*n + 0;
        const size_t ih1 = 9*n + 3;
        const size_t ih2 = 9*n + 6;

        for (size_t k = 0; k < 3; ++k) {
            grad[ih1 + k] += GRDQ(0, 0, k)*phi[4*n + 1]  // phi(h1)
                           + GRDQ(0, 1, k)*phi[4*n + 2]  // phi(h2)
                           + GRDQ(0, 2, k)*phi[4*n + 3]; // phi(M)

            grad[ih2 + k] += GRDQ(1, 0, k)*phi[4*n + 1]  // phi(h1)
                           + GRDQ(1, 1, k)*phi[4*n + 2]  // phi(h2)
                           + GRDQ(1, 2, k)*phi[4*n + 3]; // phi(M)

            grad[io + k] += GRDQ(2, 0, k)*phi[4*n + 1]  // phi(h1)
                          + GRDQ(2, 1, k)*phi[4*n + 2]  // phi(h2)
                          + GRDQ(2, 2, k)*phi[4*n + 3]; // phi(M)

            grad_add[ih1 + k] += GRDQ(0, 0, k)*phi_add[4*n + 1]  // phi(h1)
                                  + GRDQ(0, 1, k)*phi_add[4*n + 2]  // phi(h2)
                                  + GRDQ(0, 2, k)*phi_add[4*n + 3]; // phi(M)

            grad_add[ih2 + k] += GRDQ(1, 0, k)*phi_add[4*n + 1]  // phi(h1)
                                  + GRDQ(1, 1, k)*phi_add[4*n + 2]  // phi(h2)
                                  + GRDQ(1, 2, k)*phi_add[4*n + 3]; // phi(M)

            grad_add[io + k] += GRDQ(2, 0, k)*phi_add[4*n + 1]  // phi(h1)
                                 + GRDQ(2, 1, k)*phi_add[4*n + 2]  // phi(h2)
                                 + GRDQ(2, 2, k)*phi_add[4*n + 3]; // phi(M)

            grad_nonadd[ih1 + k] += GRDQ(0, 0, k)*phi_nonadd[4*n + 1] // phi(h1)
                                  + GRDQ(0, 1, k)*phi_nonadd[4*n + 2] // phi(h2)
                                  + GRDQ(0, 2, k)*phi_nonadd[4*n + 3];// phi(M)

            grad_nonadd[ih2 + k] += GRDQ(1, 0, k)*phi_nonadd[4*n + 1] // phi(h1)
                                  + GRDQ(1, 1, k)*phi_nonadd[4*n + 2] // phi(h2)
                                  + GRDQ(1, 2, k)*phi_nonadd[4*n + 3];// phi(M)

            grad_nonadd[io + k] += GRDQ(2, 0, k)*phi_nonadd[4*n + 1] // phi(h1)
                                 + GRDQ(2, 1, k)*phi_nonadd[4*n + 2] // phi(h2)
                                 + GRDQ(2, 2, k)*phi_nonadd[4*n + 3];// phi(M)
        }
    }

//    for(size_t a = 0; a < 3*nmol; ++a){
//	std::cout << grad_add[a] << " + " << grad_nonadd[a] << " = " 
//	          << grad_add[a] + grad_nonadd[a] << " ? " << grad[a]
//		  << std::endl;
//    }
//    exit(1);

    return Eint + Evdw + Eelec;
}

//----------------------------------------------------------------------------//

void ttm4f::dip(double* dip, double* q, double* pass_xyz, 
		double* moldipind) const
{
    // setup pointers (keep in sync with operators () above)

    const double* charge = m_memory;
    const double* polfac = charge + 4*m_nmol;
    const double* polar = polfac + 4*m_nmol;
    const double* xyz = polar + 4*m_nmol;

    for (size_t a = 0; a < 3*m_nmol; ++a)
	dip[a] = 0.0;

    const double* induced = m_electrostatics.dipoles();
    for (size_t n = 0; n < m_nmol; ++n) {

	const size_t n3 = 3*n;
        const size_t n4 = 4*n;
        const size_t n12 = 3*n4;

        for (size_t i = 0; i < 4; ++i)
            for (size_t k = 0; k < 3; ++k)
		dip[n3 + k] += charge[n4 + i]*xyz[n12 + 3*i + k] 
		             + induced[n12 + 3*i + k];
    }

    for (size_t a = 0; a < 3*m_nmol; ++a)
	dip[a] *= constants::DEBYE/CHARGECON; //convert to debye

    if(q){

	for (size_t a = 0; a < 12*m_nmol; ++a){
	    pass_xyz[a] = xyz[a];
	}

	for (size_t a = 0; a < 3*m_nmol; ++a)
	    moldipind[a] = 0;

	for (size_t n = 0; n < m_nmol; ++n){
	    const size_t n4 = 4*n;
	    const size_t n12 = 12*n;

	    for (size_t i = 0; i < 4; ++i)
		for (size_t k = 0; k < 3; ++k){
		    moldipind[3*n + k] += induced[n12 + 3*i + k];
//		    moldipind[3*n + k] += charge[n4 + i]*xyz[n12 + 3*i + k]; 
		}
	}

	for (size_t a = 0; a < 4*m_nmol; ++a)
	    q[a] = charge[a]/CHARGECON;
    }
    
}

} // namespace ttm
