#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include <cassert>
#include <cstdlib>

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <cmath>
#include <algorithm>

#include "read-hist.h"
#include "read-bin.h"
#include "pbc-tools.h"
#include "struct-sfg.h"
#include "math.h"

//#define SHIFT_HALF_CELL yes
#define CENTER_CELL yes
//#define BINARY yes

//#define TYPE2 yes

#include "ps.h"
#include "avila.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

////////////////////////////////////////////////////////////////////////////////

static std::vector<opt::hist_frame> frames;

std::vector<sfg::eprop> eprops;
std::vector<sfg::connect> connects;

const double zc1(1.0);
const double zc2(2.0);
const double rOO_max(4);

//----------------------------------------------------------------------------//

inline double g_scale(const double* crd)
{

    double z = std::fabs(crd[2]);
    if(z < zc1)
	return 0.0;
    else if(z > zc2)
	return copysign(1.0, crd[2]);
    else{
	const double c = std::cos(M_PI*(z-zc2)/(2*(zc1-zc2)));
	return copysign(c*c, crd[2]); // multiply switch by sign of z (crd[2])
    }

}

////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
    if (argc < 2) {
        std::cerr << "usage: z_dependent_theta_profiler POSITION_CMD > output.dat" 
	          << std::endl;
        return EXIT_FAILURE;
    }

    for(int i = 1; i < argc; ++i){
	try {

#ifdef BINARY
	    opt::read_bin(argv[i], frames, true);
#else
	    opt::read_cmd(argv[i], frames, true);
#endif

	    std::cerr <<" # Frames : " << frames.size() << std::endl;

	} catch (const std::exception& e) {
	    std::cerr << e.what() << std::endl;
	    return EXIT_FAILURE;
	}
    }

//    Open files for Writing

    // First, go over trajectory and pre-compute \mu_z and \alpha_{xx}

#ifdef CENTER_CELL
    for(size_t n = 0; n < frames.size(); ++n){
#ifdef SHIFT_HALF_CELL
    	tools::center_slab_at_origin(frames[n].natm, frames[n].cell,
				     frames[n].mass, frames[n].xyz, true);
#else
    	tools::center_slab_at_origin(frames[n].natm, frames[n].cell,
				     frames[n].mass, frames[n].xyz, false);
#endif
    }
#endif

    double dz = 0.2;
    double max_z_cm = 35.0;
    double min_z_cm = 00.0;

    int num_z_points = (max_z_cm - min_z_cm) / dz ;

    std::vector<int> N_z_theta;
    int num_z_waters[num_z_points];

    for (int i = 0; i < num_z_points; ++i)
      for (int j = 0; j < 181; ++j)
        N_z_theta.push_back(0);

    for (int i = 0; i < num_z_points; ++i) num_z_waters[i] = 0;

    // N_z_theta is now of dimenstions [z_pnts][180]

    std::cerr << "counting dipoles" << std::endl;

    for(size_t n = 0; n < frames.size(); ++n){

	const size_t nmolec = frames[n].natm/3;

        { // calculate electrostatic properties

	    for(size_t i = 0; i < nmolec; ++i){
                const size_t i3 = 3*i;
		const size_t i9 = 9*i;

		double mon[9];
		tools::image_monomer(frames[n].cell, frames[n].xyz + i9, mon);
               
		// z-component of the dipole
		double tmp_dip[3];
//		ps::dipole(mon, tmp_dip);
                for (size_t j = 0; j < 3; ++j) tmp_dip[j] = frames[n].dipmol[i3 + j];

//              std::cerr << "outside" << tmp_dip[2] << std::endl;
                // check if dipole = 0
                if (abs(tmp_dip[2]) > 0.001){

                   int z_index = int(abs(mon[2] / dz));

                   double mag = 0.0;
                   for (int k = 0; k < 3; ++k) mag += (tmp_dip[k] * tmp_dip[k]);
                   mag = sqrt(mag);
                   for (int k = 0; k < 3; ++k) tmp_dip[k] /= mag;

                   int angle_index = int(acos(tmp_dip[2]) * 57.2958);
                   // need to check if we are on the negative side of the box
                   if (mon[2] < 0) angle_index = int(acos(-1 * tmp_dip[2]) * 57.2958);

//                 std::cerr << z_index << "  " << angle_index << std::endl;
//                 std::cerr << "inside" << tmp_dip[2]; // << "  " << frames[n].xyz[0];
                   ++N_z_theta[z_index * 181 + angle_index];
                   ++num_z_waters[z_index];
                }
                
                // take dot-product with z-axis to find theta
	    }
	}
    }

    std::cout << std::setprecision(10) << std::scientific;

    // loop over z-bins
    //   then loop over 180 degrees to print N(z, theta)
    for (int i = 0; i < num_z_points; ++i){

      double total_wtheta = 0.0;

      for (int j = 0; j < 181; ++j) total_wtheta += (N_z_theta[i * 181 + j] * j);
//      for (int j = 0; j < 181; ++j) 

      if (num_z_waters[i] > 0){
        std::cout << std::setw(20) << min_z_cm + (i * dz)
                  << std::setw(20) << total_wtheta
                  << std::setw(20) << total_wtheta / num_z_waters[i]
                  << std::setw(20) << num_z_waters[i]
                  << std::endl;
       } else {
        std::cout << std::setw(20) << min_z_cm + (i * dz)
                  << std::setw(20) << 0
                  << std::setw(20) << 0
                  << std::setw(20) << 0
                  << std::endl;
       }
    }

    return 0;
}
